import http
import uuid
import logging
import os

from typing import Tuple, List, Dict, Any
from flask import request, g, abort
from flask_restplus import Resource, fields
from werkzeug.datastructures import FileStorage

from app.errors.invalid_usage import InvalidUsage
from app.libs import createGCodeParser
from app.libs.file_upload_handler import FileUploadHandler
from app.models.print_job import PrintJob
from app.models.printer import Printer
from app.libs.flask_rest_plus_additions import Namespace

logger = logging.getLogger(__name__)
ns = Namespace("print_jobs", description="Endpoint for interaction with the print job queue")


material_model = ns.model("Material", {
    "guid": fields.String(required=True, description="Material GUID"),
    "brand": fields.String(required=True, description="Material brand"),
    "material": fields.String(required=True, description="Material kind"),
    "color": fields.String(required=True, description="Material color")
})

print_core_configuration = ns.model("PrintCoreConfiguration", {
    "extruder_index": fields.Integer(required=True, description="Extruder index"),
    "material": fields.Nested(material_model),
    "nozzle_diameter": fields.String(required=True, description="Nozzle diameter (mm)"),
    "print_core_id": fields.String(required=False, description="Print core ID")
})

constraints_model = ns.model("PrintJobConstraint", {
    "require_printer_name": fields.String(required=False, description="Enforce specific printer"),
})

configuration_change_model = ns.model("ConfigurationChange", {
    "type_of_change": fields.String(required=True, description="either 'material' or 'print_core_change' string" +
                                                               " indicating the type of change."),
    "index": fields.Integer(required=False, description="the hotend slot or extruder index to change"),
    "target_id": fields.String(required=True, description="target material guid or hotend id"),
    "origin_id": fields.String(required=True, description="original/current material guid or hotend id"),
    "target_name": fields.String(required=False, description="target material name or hotend id"),
    "origin_name": fields.String(required=False, description="original/current material name or hotend id")
})

match_results_model = ns.model("MatchResult",{
    "translation_key": fields.String(
        required=True,
        description="A string indicating A reason the print cannot be printed, such as 'does_not_fit_in_build_volume'"
    ),
    "severity": fields.String(
        required=True,
        description="A number indicating the severity of the problem"
    )
})

buildPlateModel = ns.model("BuildPlate", {
    "type": fields.String(description="The build plate type required for the job.")
})

api_model = ns.model("PrintJob", {
    "uuid": fields.String(required=True, description="Uniquely identifies the print job"),
    "name": fields.String(required=True, description="The human readable filename"),
    "created_at": fields.DateTime(description="The start time of the print job"),
    "started": fields.Boolean(required=True, description="Whether the job has been started"),
    "status": fields.String(description="The status of the print job"),
    "printer_uuid": fields.String(description="Printer uuid associated to this job"),
    "configuration": fields.List(fields.Nested(print_core_configuration)),
    "machine_variant": fields.String(required=True, description="Target machine variant for the job"),
    "constraints": fields.Nested(constraints_model),
    "time_elapsed": fields.Integer(required=True, description="Current elapsed print time in seconds."),
    "time_total": fields.Integer(required=True, description="Current estimated total print time in seconds."),
    "last_seen": fields.Float(
        required=False,
        description="Seconds since the print job was last checked"),
    "network_error_count": fields.Integer(
        required=False,
        description="Number of failed connection attempts to check the job status"),
    "force": fields.Boolean(
        required=True,
        description="A boolean indicating whether the job ignore matching criteria such as material type, " +
                    "this only works in combination with a constraint to a specific printer"
    ),
    "assigned_to": fields.String(required=False, description="Printer uuid of printer reserved for this job"),
    "owner": fields.String(required=False, description="Name of the person who submitted the job"),
    "build_plate": fields.Nested(buildPlateModel),
    "configuration_changes_required": fields.List(fields.Nested(configuration_change_model)),
    "impediments_to_printing": fields.List(fields.Nested(match_results_model))
})

update_print_job_model = ns.model("UpdatePrintJob", {
    "force": fields.Boolean(
        required=False,
        description="A boolean indicating whether the job ignore matching criteria such as material type, " +
                    "this only works in combination with a constraint to a specific printer"
    )
})

print_job_action_fields = ns.model("PrintJobAction", {
    "action": fields.String(
        required=True,
        description=\
            '"print", "pause" or "abort". Change the current state of the print. Note that only changes to abort ' +
            '/ pause are always allowed and changing to print only when state is paused.'
    )
})

print_job_position_fields = ns.model("PrintJobPosition", {
    "to_position": fields.Integer(
        required=True,
        description="Set position of print job; only position 0 is supported"
    )
})

upload_parser = ns.parser()
upload_parser.add_argument("file", location="files", type=FileStorage, required=True)


@ns.route("/")
class PrintJobList(Resource):
    """
    Shows a list of all print jobs, and lets you POST to add new jobs to the queue
    """

    @ns.doc("list_print_jobs")
    @ns.marshal_with(api_model, as_list=True)
    def get(self):
        """
        Return a list of all current print jobs in the queue
        """
        return g.print_job_service().print_jobs.all()

    @ns.doc("create_print_job")
    @ns.expect(upload_parser)
    @ns.marshal_with(api_model)
    @ns.response(200, "Print job added to queue")
    @ns.response(400, "Bad request, invalid or unsupported g-code file")
    def post(self) -> PrintJob:
        """
        Add a new print job to the queue
        """
        logger.debug("Receiving file...")
        if "file" not in request.files or request.files["file"].filename == "":
            raise InvalidUsage("No file found")

        file = request.files["file"]
        job_uuid = str(uuid.uuid4())
        handler = FileUploadHandler(job_uuid, file)
        handler.save_to_disk()

        gcode_filename = handler.get_filepath()
        gcode_parser = createGCodeParser(gcode_filename)
        if gcode_parser is None or not gcode_parser.is_valid:
            logger.warning("G-code file [%s] is not valid, removing...", gcode_filename)
            os.remove(gcode_filename)
            return abort(400, 'G-code file is not valid')
        if not gcode_parser.is_supported:
            logger.warning("G-code file [%s] is not supported, removing...", gcode_filename)
            os.remove(gcode_filename)
            return abort(400, 'G-code file is not supported')
        if gcode_parser.machine_variant is None:
            logger.warning("G-code file [%s] for unknown machine variant, removing...", gcode_filename)
            os.remove(gcode_filename)
            return abort(400, 'G-code file is not supported')

        constraints = {}
        if "require_printer_name" in request.form:
            constraints["require_printer_name"] = request.form["require_printer_name"]
            logger.debug("Constraints found: %s", constraints)

        force = False
        if "force" in request.form:
            force = bool(request.form["force"])
            logger.debug("'Force' flag set to : %s", force)

        owner = None
        if "owner" in request.form:
            owner = request.form["owner"]
            logger.debug("Owner is: %s", owner)

        g.print_job_service().print_jobs.add(
            PrintJob(
                uuid=job_uuid,
                name=handler.get_filename(),
                filepath=handler.get_filepath(),
                constraints=constraints,
                force=force,
                owner=owner
            )
        )

        print_job = g.print_job_service().get(job_uuid)

        return print_job


@ns.route('/<string:uuid>')
class PrintJobs(Resource):
    @ns.doc('delete_print_job')
    @ns.response(204, 'Print job deleted')
    @ns.response(404, 'Print job not found')
    def delete(self, uuid):
        if not uuid:
            return '', http.client.BAD_REQUEST

        print_job = g.print_job_service().get(uuid)  # type: PrintJob
        if print_job is None:
            return "Print job was not found", http.client.NOT_FOUND

        g.print_job_service().print_jobs.delete(print_job)
        return '', http.client.NO_CONTENT


@ns.route('/<string:uuid>/update')
class PrintJobUpdate(Resource):
    @ns.doc("update_a_setting_on_the_print_job")
    @ns.response(http.client.OK, "Job updated!")
    @ns.response(http.client.NOT_FOUND, "Print job was not found")
    @ns.expect(update_print_job_model)
    def put(self, uuid: str) -> Tuple[str, int]:
        print_job = g.print_job_service().get(uuid)  # type: PrintJob
        if print_job is None:
            return "Print job was not found", http.client.NOT_FOUND

        data = request.get_json(force=True)  # type: Dict[str, Any]

        force_flag = data.get("force")
        if force_flag is not None:
            g.print_job_service().set_force_flag(print_job.uuid, force_flag)

        return "Job updated!", http.client.OK


@ns.route('/<string:uuid>/move_to_position')
class PrintJobUpdatePosition(Resource):
    @ns.doc("update_position_of_print_job")
    @ns.response(http.client.OK, "Job updated!")
    @ns.response(http.client.NOT_FOUND, "Print job was not found")
    @ns.expect(print_job_position_fields)
    def put(self, uuid: str) -> Tuple[str, int]:
        """Set position of print job"""
        data = request.get_json(force=True)
        to_position = data.get("to_position")
        print_job = g.print_job_service().get(uuid)
        if print_job is None:
            return "Print job was not found", http.client.NOT_FOUND
        if to_position is None or to_position != 0:
            return "Cannot perform to_position != 0", http.client.BAD_REQUEST
        g.print_job_service().print_jobs.move(print_job, to_position)

        return "Job updated!", http.client.OK


@ns.route('/<string:uuid>/action')
class PrintJobPerformAction(Resource):
    @ns.doc("print_job_perform_action")
    @ns.response(http.client.NO_CONTENT, "Job action succeeded")
    @ns.response(http.client.NOT_FOUND, "Print job was not found")
    @ns.response(http.client.METHOD_NOT_ALLOWED, "Print job is not printing")
    @ns.response(http.client.BAD_REQUEST, "Cannot perform action on print job")
    @ns.expect(print_job_action_fields)
    def put(self, uuid: str) -> Tuple[str, int]:
        """Perform action on print job

        "pause", "abort" or "print" (=resume)
        """
        print_job = g.print_job_service().get(uuid)
        data = request.get_json(force=True)
        action = data.get("action")
        if action not in {"print", "abort", "pause"}:
            return "Invalid state", http.client.BAD_REQUEST
        if print_job is None:
            return "Print job was not found", http.client.NOT_FOUND
        if not print_job.printing:
            return "Print job is not printing", http.client.METHOD_NOT_ALLOWED
        success = g.print_job_service().request_print_job_state(uuid, action)
        if not success:
            return "Could not set state", http.client.BAD_REQUEST

        return "State changed", http.client.OK


@ns.route("/queued")
class QueuedPrintJobs(Resource):
    """
    Shows a list of all queued print jobs
    """
    @ns.doc("queued_print_jobs")
    @ns.marshal_with(api_model, as_list=True)
    def get(self) -> List[PrintJob]:
        """
        Return a list of all queued print jobs
        """
        return g.print_job_service().print_jobs.queued()


@ns.route("/printing")
class PrintingPrintJobs(Resource):
    """
    Shows a list of all started print jobs
    """

    @ns.doc("printing_print_jobs")
    @ns.marshal_with(api_model, as_list=True)
    def get(self) -> List[PrintJob]:
        """
        Return a list of all started print jobs
        """
        return g.print_job_service().print_jobs.printing()
