import http.client
import socket
import logging

from flask import g, request
from flask_restplus import Resource, fields, Model
from typing import Any, Dict, Union, Tuple, Optional

from app.errors.invalid_usage import InvalidUsage
from app.libs.printer_api import PrinterApi
from app.libs.printer_api_exceptions import PrinterApiException
from app.libs.printer_credentials_store import PrinterCredentialsStore
from app.libs.ultimaker3.system import System
from app.services.printer_service import PrinterService
from app.libs.printer_discovery import PrinterDiscovery
from app.libs.printer_api import PrinterApiHttpException
from app.libs.flask_rest_plus_additions import Namespace
from app.models.printer import Printer as PrinterModel
from um3_commons.thread import Thread

logger = logging.getLogger(__name__)

ns = Namespace("printers", description="Endpoint for interaction with the printers")

# TODO: This model has the exact same signature as the one in print_jobs.py. Might want to abstract this?
material_model = ns.model("Material", {
    "guid": fields.String(required=True, description="Material GUID"),
    "brand": fields.String(required=True, description="Material brand"),
    "material": fields.String(required=True, description="Material kind"),
    "color": fields.String(required=True, description="Material color")
})

print_core_configuration = ns.model("PrintCoreConfiguration", {
    "extruder_index": fields.Integer(required=True, description="Extruder index"),
    "nozzle_diameter": fields.Float(required=True, description="Nozzle diameter (mm)"),
    "print_core_id": fields.String(required=False, description="Print Core ID"),
    "material": fields.Nested(material_model)
})

buildPlateModel = ns.model("BuildPlate", {
    "type": fields.String(description="The build plate type currently loaded in this printer.")
})

printerModel = ns.model("Printer", {
    "uuid": fields.String(required=False, description="The identifier"),
    "status": fields.String(required=False, description="Printer status"),
    "unique_name": fields.String(required=True, description="The printer's unique name"),
    "ip_address": fields.String(required=True, description="The printer's last known IP address"),
    "firmware_version": fields.String(required=False, description="Firmware version"),
    "friendly_name": fields.String(required=True, description="Human readable name"),
    "enabled": fields.Boolean(required=False, description="Whether or not the printer may receive print jobs"),
    "reserved_by": fields.String(required=False, description="Is the printer reserved by any job"),
    "machine_variant": fields.String(description="Printer machine variant"),
    "build_plate": fields.Nested(buildPlateModel),
    "configuration": fields.List(fields.Nested(print_core_configuration))
})

addPrinterModel = ns.model("AddPrinter", {
    "address": fields.String(required=True, description="The printer's IP address or host name")
})

updatePrinterModel = ns.model("UpdatePrinter", {
    "enabled": fields.Boolean(required=True, description="Whether or not to allow print jobs to be sent to the printer")
})

# a couple of models for all the maintenance api uses...
add_maintenance_item_input = ns.model("add_maintenance_item", {
    "action":  fields.String(required=True, description="The maintenance action to be logged"),
    "mechanic": fields.String(required=False, description="The user or mechanic who preformed the maintenance action")
})

get_maintenance_history_output = ns.model("get_maintenance_output", {
    "action": fields.String(required=True, description="The maintenance action to be logged"),
    "mechanic": fields.String(required=True, description="The user or mechanic who preformed the maintenance action"),
    "timestamp": fields.DateTime(
        required=True,
        description="a timestamp string in a YYYY-mm-ddTHH:MM:SS format where T is a divider between time and date"
    )
})


@ns.route("/")
class PrinterList(Resource):
    """
    Shows a list of all printers, and lets you POST to add new printers
    """
    @ns.doc("list_printers")
    @ns.marshal_with(printerModel, as_list=True)
    def get(self):
        """
        Return a list of all the connected printers
        """
        return g.printer_service().printers.all()

    @ns.doc("create_printer")
    @ns.expect(addPrinterModel)
    @ns.marshal_with(printerModel)
    def post(self):
        """
        Add a new printer
        """
        printer_json = request.get_json(force=True)
        logger.info("Adding printer [%s]...", printer_json)

        if printer_json and "address" in printer_json and printer_json["address"]:
            printer_ip_address = self.__ensure_ip_address_resolved(printer_json["address"])

            return self.__check_and_authorize_printer(printer_ip_address)

        raise InvalidUsage("Provide an IP address or hostname")

    @staticmethod
    def __ensure_ip_address_resolved(address: str):
        """
        Check if address is an IP address. If so, return the address. If not, try to resolve it to an IP address.
        :param address: IP address or hostname
        :return: the resolved IP address
        """
        address = address.strip()
        try:
            socket.inet_aton(address)
            return address
        except OSError:
            try:
                return socket.gethostbyname(address + ".local")
            except socket.gaierror:
                return None

    @staticmethod
    def __is_printer_in_group(address: str):
        """
        Deny adding printers that are already part of the group
        :param address:
        :return: Bool
        """
        printer_in_group = g.printer_service().find_by_ip_address(address)

        if printer_in_group and g.printer_service().find_by_unique_name(printer_in_group.unique_name):
            return True

        return False

    @staticmethod
    def __check_and_authorize_printer(printer_ip_address: str) -> Union[Optional[PrinterModel], Tuple[str, int]]:
        printer_service = g.printer_service()  # type: PrinterService
        credentials_store = PrinterCredentialsStore()

        printer_api = PrinterApi(printer_ip_address, credentials_store)

        if printer_service.system_dbus is None:
            machine_name = "ultimaker"
            unique_name = ""
            logger.warning("Could not get this machine's name through dbus, should never happen on machine")
        else:
            machine_name = printer_service.system_dbus.getMachineName()
            unique_name = printer_service.system_dbus.getHostName()

        try:
            system = printer_api.get_system()
        except PrinterApiException:
            return "Unable to reach the printer", http.client.SERVICE_UNAVAILABLE

        printer = printer_service.find_by_unique_name(system.hostname)
        if printer is not None:
            if printer.ip_address != printer_ip_address and printer_ip_address != "127.0.0.1":
                printer.ip_address = printer_ip_address
            return "Can’t add same printer twice. Printer already in this group.", http.client.FORBIDDEN

        if printer_api.is_hosting_multiple_printers():
            return "Printer is a host", http.client.PRECONDITION_FAILED

        t = Thread(
            "handle_wait_for_printer_{}_auth".format(printer_ip_address),
            target=PrinterList.__auth_printer,
            args=(printer_service,
                  printer_ip_address,
                  printer_api,
                  credentials_store,
                  unique_name,
                  machine_name,
                  system)
        )
        t.start()

        return "Waiting for authentication", http.client.OK

    @staticmethod
    def __auth_printer(
            printer_service: PrinterService,
            printer_ip_address: str,
            printer_api: PrinterApi,
            credentials_store: PrinterCredentialsStore,
            unique_name: str,
            machine_name: str,
            system_info: System,
    ) -> None:
        try:
            printer_api.authenticate(exclusion_key="cluster", host_name=unique_name, api_user=machine_name)

            if printer_api.wait_until_authenticated(timeout=5 * 60):

                printer = PrinterModel(
                    ip_address=printer_ip_address,
                    unique_name=system_info.hostname,
                    friendly_name=system_info.name,
                    machine_variant=system_info.machine_variant,
                    digest_auth_id=credentials_store.id,
                    digest_auth_key=credentials_store.key,
                    digest_auth_authorized=True
                )
                printer_service.printers.add(printer)

        except PrinterApiException:
            logger.warning("Unable to reach the printer!")


@ns.route('/<string:uuid>')
class Printer(Resource):
    @ns.doc("update_printer")
    @ns.marshal_with(printerModel)
    @ns.expect(updatePrinterModel, validate=True)
    @ns.response(http.client.OK, "Printer updated")
    @ns.response(http.client.NOT_FOUND, "Printer not found")
    def put(self, uuid):
        """
        Update printer information
        """
        printer_service = g.printer_service()  # type: PrinterService
        printer = printer_service.get(uuid)
        if printer is None:
            return "", http.client.NOT_FOUND

        json = request.get_json(force=True)

        enabled = json.get("enabled")
        if enabled is not None:
            printer.enabled = enabled
            return printer, http.client.OK
        else:
            return "", http.client.NOT_FOUND

    @ns.doc("delete_printer")
    @ns.response(204, 'Printer removed')
    @ns.response(404, 'Printer not found')
    def delete(self, uuid):
        """
        Remove printer from the cluster
        """
        if uuid:
            printer_service = g.printer_service()  # type: PrinterService
            printer = printer_service.get(uuid)

            if printer is not None:
                printer_service.printers.delete(printer)
                return '', http.client.NO_CONTENT
            else:
                return 'Printer not found', http.client.NOT_FOUND

        return 'Printer not found', http.client.NOT_FOUND


@ns.route("/discovered/")
class NetworkPrinterList(Resource):
    @ns.doc("list_network_printers")
    @ns.marshal_with(printerModel, as_list=True)
    def get(self):
        discovered_printers = PrinterDiscovery.getInstance().get_discovered_printers()
        added_printers = g.printer_service().printers.all()
        added_printer_hosts = [printer.ip_address for printer in added_printers]

        new_discovered_printers = filter(
            lambda printer: printer.ip_address not in added_printer_hosts,
            discovered_printers
        )

        return sorted(new_discovered_printers, key=lambda printer: printer.friendly_name.lower())


@ns.route("/<string:uuid>/maintenance/")
class MaintenanceItemInput(Resource):
    @ns.doc("add_maintenance_action_to_the_log")
    @ns.expect(add_maintenance_item_input, validate=True)
    @ns.response(http.client.OK, "Action logged")
    @ns.response(http.client.NOT_FOUND, "Printer not found")
    @ns.response(http.client.SERVICE_UNAVAILABLE, "Failed to log action on printer")
    def post(self, uuid):
        printer_service = g.printer_service()   # type:PrinterService
        json = request.get_json(force=True)

        printer = printer_service.get(uuid)
        if not printer:
            return "Printer not found", http.client.NOT_FOUND

        action = json.get("action")
        if not action:
            return "missing printer uuid or action ({}, {})".format(str(printer), str(action))

        if printer.add_maintenance_history_item(json.get("action"), json.get("mechanic")):
            return "Action logged", http.client.OK
        else:
            return "Failed to log action on printer", http.client.SERVICE_UNAVAILABLE

    @ns.doc("get_list_of_maintenance_actions")
    @ns.marshal_list_with(get_maintenance_history_output)
    @ns.response(http.client.OK, "data", get_maintenance_history_output)
    @ns.response(http.client.NOT_FOUND, "Printer not found")
    @ns.response(http.client.SERVICE_UNAVAILABLE, "Error with remote printer")
    def get(self, uuid):
        printer_service = g.printer_service()  # type:PrinterService

        printer = printer_service.get(uuid)
        if not printer:
            return 'Printer not found', http.client.NOT_FOUND

        try:
            data = printer.get_maintenance_history()
        except PrinterApiHttpException:
            return "Error on remote printer", http.client.SERVICE_UNAVAILABLE
        except Exception:
            return "Unknown error", http.client.SERVICE_UNAVAILABLE

        if data is None:
            return "Error connecting to remote printer", http.client.SERVICE_UNAVAILABLE
        return data
