from http.client import SERVICE_UNAVAILABLE
from flask import Flask, request
from typing import Optional
from typing import Tuple

from app.events.print_job_deleted import PrintJobDeletedEvent
from app.events.print_job_started_event import PrintJobStartedEvent
from app.events.print_job_updated_event import PrintJobUpdatedEvent
from app.events.printer_added_event import PrinterAddedEvent
from app.events.printer_deleted_event import PrinterDeletedEvent
from app.events.printer_status_changed_event import PrinterStatusChangedEvent
from app.events.printer_updated_event import PrinterUpdatedEvent
from app.models.print_job_collection import PrintJobCollection
from app.models.printer_collection import PrinterCollection
from app.storage.print_job_event_storage_manager import PrintJobEventStorageManager
from app.storage.printer_event_storage_manager import PrinterEventStorageManager
from um3_commons.registry.registryFile import RegistryFile
from um3_commons.thread import Thread
from werkzeug.contrib.fixers import ProxyFix
import logging

from app.libs import watchdog, printer_discovery
from app.events.print_job_added_event import PrintJobAddedEvent
from app.events.print_job_moved_event import PrintJobMovedEvent
from app.libs.signal import Signal
from app.loghelper import setup_logging
from app.services.event_types import EventTypes
from app.services.material_manager import MaterialManager
from app.services.scheduler.schedule_runner import ScheduleRunner
from app.services.printer_service import PrinterService
from app.services.print_job_service import PrintJobService
from app.services.printer_monitor_service import PrinterMonitorService
from app.services.scheduler.scheduler import PrintJobScheduler

from app.storage.event_store import EventStore
from.settings import Version
from app.stats.statistics_logger import StatisticsLogger


from app.site import mod as site
from app.apiv1 import mod as api, URL_PREFIX


logger = logging.getLogger(__name__)


class Application:

    def __init__(self, config_object_name: str) -> None:
        self.__printer_service = None           # type: PrinterService
        self.__print_job_service = None         # type: PrintJobService
        self.__material_manager = None         # type: MaterialManager
        self.__flask_app = None                 # type: Flask
        self.__event_store = None               # type: EventStore
        self.__printer_monitor_service = None   # type: PrinterMonitorService
        self.__registry_file = None              # type: Optional[RegistryFile]
        self.__application_started = False
        self.__application_started_signal = Signal[Tuple]()
        self.__flask_app = None                 # type: Flask

        app = Flask(__name__)
        self.__flask_app = app

        # Expose this Application object to requests via Flask's magic 'g'.
        app.app_ctx_globals_class = self.__create_request_context

        app.wsgi_app = ProxyFix(app.wsgi_app)
        app.config.from_object(config_object_name)

        setup_logging(console_level=app.config["LOG_LEVEL"])

        Version()  # construct version object so that we output a nice version string in the log

        RegistryFile.setStorageLocation(base_path=app.config["CONFIGURATION_FILE_BASE_DIR"])
        self.__registry_file = RegistryFile(app.config["CONFIGURATION_FILE_NAME"])

        # register our blueprints
        app.register_blueprint(site)
        app.register_blueprint(api)

        app.before_request(self.__pre_request_check_app_started)
        app.after_request(self.__apply_caching)

        # Create the service instances
        self.__event_store = EventStore(app.config["EVENTSTORE_FILEPATH"])
        self.register_event_types(self.__event_store)

        print_job_collection = PrintJobCollection()
        printer_collection = PrinterCollection()

        PrintJobEventStorageManager(
            event_store=self.__event_store,
            replay_done=self.__application_started_signal,
            collection=print_job_collection
        )
        PrinterEventStorageManager(
            event_store=self.__event_store,
            replay_done=self.__application_started_signal,
            collection=printer_collection
        )

        self.__material_manager = MaterialManager(use_dbus=config_object_name.endswith("ProdConfig"))

        self.__printer_monitor_service = PrinterMonitorService(use_thread=not app.config["TESTING"])
        self.__printer_service = PrinterService(
            collection=printer_collection,
            printer_monitor_service=self.__printer_monitor_service,
            material_manager=self.__material_manager,
            application_started_signal=self.__application_started_signal)

        self.__print_job_service = PrintJobService(
            collection=print_job_collection,
            printer_service=self.__printer_service,
            printer_monitor_service=self.__printer_monitor_service,
            application_started_signal=self.__application_started_signal,
            gcode_spool_directory=app.config["SPOOL_DIR"],
            material_manager=self.__material_manager
        )

        self.__scheduler_runner = ScheduleRunner(
            printer_service=self.__printer_service,  # TODO: pass it only the collections
            print_job_service=self.__print_job_service,
            material_manager=self.__material_manager
        )
        self.__application_started_signal.connect(lambda _: self.__scheduler_runner.start())

        # initialize the analytics system
        if config_object_name.endswith("ProdConfig"):
            StatisticsLogger(self.__registry_file, self.printer_service())

        # make stuff findable via dbus
        if config_object_name.endswith("ProdConfig"):
            from app.dbus.cluster_service import ClusterService
            ClusterService(self.__application_started_signal, self.__event_store, self.__printer_service,
                           self.__print_job_service, self.__scheduler_runner, app, self.__material_manager,
                           self.__printer_monitor_service, self.__registry_file)

        # Initiate the printer discovery, so we always have a list of available printers on the network.
        printer_discovery.PrinterDiscovery.startPrinterDiscovery()

        watchdog.get_instance().application_started()
        if not app.config["TESTING"]:
            start_up_thread = Thread(
                name="StartUp",
                target=self.__startup_phase2)
            start_up_thread.start()
        else:
            self.__startup_phase2()

    def __startup_phase2(self) -> None:
        logger.info("Starting up the application in the background")
        watchdog.get_instance().start_monitoring()

        watchdog.get_instance().tick()

        # Replay the event store
        self.__event_store.replay_all_events()

        # Record that the application has started.
        application_started_event = self.__event_store.create_event(event_type=EventTypes.APPLICATION_STARTED,
                                                                    entity_uuid=None)
        self.__event_store.append_event(application_started_event)

        self.__application_started_signal.emit(())
        self.__application_started = True
        self.__printer_monitor_service.start()

        logger.info("Completed starting up the application in the background")
        watchdog.get_instance().stop_monitoring()

    def __pre_request_check_app_started(self):
        if request.path.startswith(URL_PREFIX) and not self.__application_started:
            return "Application is busy starting up and not ready for requests.", SERVICE_UNAVAILABLE
        return None

    def __apply_caching(self, response) -> None:
        response.headers["cache-control"] = "no-cache"
        return response

    def __create_request_context(self) -> 'Application':
        return self

    def printer_service(self) -> PrinterService:
        return self.__printer_service

    def print_job_service(self) -> PrintJobService:
        return self.__print_job_service

    def material_manager(self) -> MaterialManager:
        return self.__material_manager

    def print_job_scheduler(self) -> PrintJobScheduler:
        return self.__scheduler_runner.scheduler

    def registry_file(self) -> RegistryFile:
        return self.__registry_file

    def get_flask_app(self) -> Flask:
        return self.__flask_app

    @staticmethod
    def register_event_types(event_store: EventStore) -> None:
        event_store.register_event(PrinterAddedEvent, EventTypes.PRINTER_ADDED)
        event_store.register_event(PrinterUpdatedEvent, EventTypes.PRINTER_UPDATED)
        event_store.register_event(PrinterStatusChangedEvent, EventTypes.PRINTER_STATUS_CHANGED)
        event_store.register_event(PrinterDeletedEvent, EventTypes.PRINTER_REMOVED)

        event_store.register_event(PrintJobAddedEvent, EventTypes.PRINT_JOB_ADDED)
        event_store.register_event(PrintJobStartedEvent, EventTypes.PRINT_JOB_STARTED)
        event_store.register_event(PrintJobUpdatedEvent, EventTypes.PRINT_JOB_UPDATED)
        event_store.register_event(PrintJobDeletedEvent, EventTypes.PRINT_JOB_REMOVED)
        event_store.register_event(PrintJobMovedEvent, EventTypes.PRINT_JOB_MOVED)


# Main entry point for starting the app.
def create_app(config_object_name) -> Flask:
    app = Application(config_object_name)
    return app.get_flask_app()
