import logging
from typing import Tuple

import dbus
# included for typing
from flask import Flask

from app.libs import watchdog
from app.libs.file_upload_handler import FileUploadHandler
from app.libs.signal import Signal
from app.services.material_manager import MaterialManager
from app.services.print_job_service import PrintJobService
from app.services.printer_monitor_service import PrinterMonitorService
from app.services.printer_service import PrinterService
from app.services.scheduler.schedule_runner import ScheduleRunner
from app.storage.event_store import EventStore
from um3_commons import dbusif
from um3_commons import thread
from um3_commons.registry.registryFile import RegistryFile

logger = logging.getLogger(__name__)


class ClusterService(dbusif.ServiceObject):
    def __init__(self,
                 application_started_signal: Signal,
                 event_store: EventStore,
                 printer_service: PrinterService,
                 print_job_service: PrintJobService,
                 print_job_scheduler: ScheduleRunner,
                 app: Flask,
                 material_manager: MaterialManager,
                 printer_monitor_service: PrinterMonitorService,
                 registry_file: RegistryFile) -> None:
        super().__init__("cluster")
        # stored as public variables as they are used purely for debug puroposes
        self.application_started_signal = application_started_signal
        self.event_store = event_store
        self.printer_service = printer_service
        self.print_job_service = print_job_service
        self.print_job_scheduler = print_job_scheduler
        self.app = app
        self.material_manager = material_manager
        self.printer_monitor_service = printer_monitor_service
        self.registry_file = registry_file

        # the dbus service events need to be triggered by a dbus thread, this thread is not automatically created.
        self._dbus_thread = thread.Thread("dbus mainloop", target=dbusif.runMainLoop)
        self._dbus_thread.start()

        printer_service.amount_of_printers_changed.connect(self.onAmountOfPrintersInClusterChanged)

        application_started_signal.connect(self.__handle_application_started)

        # Add callback to this service which implements the call to the resetToFactorySettings
        dbusif.RemoteObject("system").addFactoryResetCallback("cluster", "resetSettings")

    @dbusif.method("", "")
    def stop(self):
        dbusif.abortMainLoop()

    def __handle_application_started(self, _: Tuple):
        try:
            avahi = dbusif.RemoteObject("interface/avahi")
            avahi.connectKeyValuePairToSignal("cluster", "onAmountOfPrintersInClusterChanged", "cluster_size")
        except dbus.exceptions.DBusException:
            logger.warning("****** Could not connect to avahi service, this should never happen on the actual printer")

        dog = watchdog.get_instance()
        dog.delay_monitoring_by(60)
        # somehow this can take about a minute
        self.onAmountOfPrintersInClusterChanged(len(self.printer_service.printers))  # relay initial value
        dog.tick()

    # this is a dbus signal, if this python function is called internally , then an external dbus event is fired.
    # the naming here (on... and camelcase) is taken to be similar to all other dbus serveces and events used.
    @dbusif.signal("i")
    def onAmountOfPrintersInClusterChanged(self, nr_of_printers):
        logger.info("onAmountOfPrintersInClusterChanged(%d)", nr_of_printers)
        return True

    # @brief The callback function to register which needs to be called when a factory reset is issued
    # @param reset_type One of the defined values from FactoryReset which would indicate a hard or soft reset
    @dbusif.method("s", "")
    def resetSettings(self, reset_type):
        self.registry_file.erase(restart_after_erase=False)
        self.event_store.clear_event_store()
        FileUploadHandler.clear_all_uploaded_files()
