import json
from http.client import OK  # import http status codes to avoid magic numbers
from functools import wraps

from flask import make_response
from flask_restplus import fields
from flask_restplus.utils import unpack
from typing import Any, List

plain_json_fields = (fields.Integer, fields.Float, fields.Boolean)


def marshal_to_json(schema: fields.Raw, obj: Any) -> str:
    """Format an object as JSON using the Flask fields schema.

    This function is designed to be as fast as possible.
    Note: It doesn't support all of the features on Flask Restplus'
    implementation.
    """
    if isinstance(schema, fields.Nested):
        return marshal_to_json(schema.model, obj)

    result = ['{']  # type: List[str]
    comma = ''
    for k, v in schema.items():

        if isinstance(obj, dict):
            target_value = obj.get(k, None)
        else:
            if hasattr(obj, k):
                target_value = getattr(obj, k)
            else:
                target_value = None

        result.append(comma)
        comma = ','

        result.append('"')
        result.append(k)
        result.append('":')

        if isinstance(v, fields.String):
            if target_value is None:
                result.append('null')
            else:
                result.append(json.dumps(str(target_value)))

        elif isinstance(v, plain_json_fields):
            result.append(json.dumps(target_value))

        elif isinstance(v, fields.DateTime):
            if target_value is None:
                result.append('null')
            else:
                result.append('"')
                result.append(target_value.isoformat())
                result.append('"')

        elif isinstance(v, fields.List):
            if target_value is None:
                result.append('null')
            else:
                container = v.container
                result.append(marshal_as_list_to_json(container, target_value))

        elif isinstance(v, fields.Nested):
            result.append(marshal_to_json(v.model, target_value))

        else:
            raise Exception('Unsupported field type "' + repr(v) + '" for key "' + k + '".')

    result.append('}')

    return ''.join(result)


def marshal_as_list_to_json(schema: fields.Raw, obj: List[Any]) -> str:
    result = ['[']
    list_comma = ""
    for item in obj:
        result.append(list_comma)
        list_comma = ','
        if isinstance(schema, fields.String):
            if item is not None:
                result.append(json.dumps(str(item)))
            else:
                result.append("null")
        elif isinstance(schema, plain_json_fields):
            result.append(json.dumps(item))
        else:
            result.append(marshal_to_json(schema, item))

    result.append(']')
    return ''.join(result)


class fast_marshal_with(object):
    """A decorator that apply fast JSON marshalling to the return values of your methods.
    """
    def __init__(self, fields, as_list):
        self.fields = fields
        self.as_list = as_list

    def __call__(self, f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            resp = f(*args, **kwargs)

            if isinstance(resp, tuple):
                data, code, headers = unpack(resp)
            else:
                data = resp
                code = OK
                headers = {}

            if isinstance(data, str):
                json_string = json.dumps(data)  # add "'s
            elif self.as_list:
                json_string = marshal_as_list_to_json(self.fields, data)
            else:
                json_string = marshal_to_json(self.fields, data)

            response = make_response(json_string, code)

            for k, v in headers:
                response.headers[k] = v
            response.headers["Content-Type"] = "application/json"

            return response

        return wrapper
