import http.client
import logging
from typing import Dict, Any
from flask_restplus.utils import merge
from flask_restplus.reqparse import RequestParser
from flask_restplus import fields, Model, Namespace as FlaskRestPlusNamespace
from flask import request
from app.libs.fast_json_marshal import fast_marshal_with

logger = logging.getLogger(__name__)

plain_json_fields_conversion = {fields.Integer: int, fields.Float: float, fields.Boolean: bool, fields.String: str}


class Namespace(FlaskRestPlusNamespace):
    """Replacement Namespace class which extends Flash restplus' version with
    fast code to serialize objects to JSON.
    """
    def marshal_with(self, fields, as_list=False, code=200, description=None, **kwargs):
        def wrapper(func):
            doc = {
                'responses': {
                    code: (description, [fields]) if as_list else (description, fields)
                },
                '__mask__': kwargs.get('mask', True),  # Mask values can't be determined outside app context
            }
            func.__apidoc__ = merge(getattr(func, '__apidoc__', {}), doc)

            return fast_marshal_with(fields, as_list, **kwargs)(func)
        return wrapper

    def expect(self, *inputs, **kwargs):
        super_func = super().expect(*inputs, **kwargs)

        if isinstance(inputs[0], RequestParser):
            return super_func

        assert isinstance(inputs[0], Model), "First param must be a model and is %s" % type(inputs[0])

        def class_time_decorator(original_func):
            def wrapper(*args, **kwargs):
                try:
                    data = sanitize_json(request.get_json(force=True), inputs[0])
                except TypeError as e:
                    return str(e), http.client.BAD_REQUEST
                request._cached_json = data  # yes this works, not super pretty though.
                return original_func(*args, **kwargs)

            return super_func(wrapper)
        return class_time_decorator


def sanitize_json(data: Dict[str, Any], model: Model):
    output = {key: check_type(m, data.get(key)) for key, m in model.items() if key in data or m.required}

    if output.keys() != data.keys():
        for key in data.keys():
            if key not in output.keys():
                logger.warning("additional api key '%s' will be ignored because it is not in model: '%s'", key, model)
    return output


def check_type(model: fields.Raw, data: Any) -> Any:
    expected_type = plain_json_fields_conversion.get(type(model), None)

    if expected_type is not None and type(data) != expected_type:
        if data is None and not model.required:
            return data
        raise TypeError("Json input type(%s) did not match model (expected type:%s)" % (type(data), expected_type))
    return data
