from enum import IntEnum
import http.client
import logging

import requests
from requests.models import Response
from requests import RequestException
from requests.auth import HTTPDigestAuth
import time
from typing import List, Optional, Any, Dict

from app.libs.ultimaker3.material import Material
from app.libs.ultimaker3.print_job import PrintJob
from app.libs.ultimaker3.status import Status
from app.libs.ultimaker3.system import System
from app.libs.ultimaker3.maintenance import Maintenance
from app.libs import watchdog
from app.models.led_color import LedColor
from .printer_api_exceptions import PrinterApiHttpException, PrinterApiNetworkException, PrinterApiException

AUTH_AUTHORIZED = "authorized"
AUTH_UNAUTHORIZED = "unauthorized"
AUTH_UNKNOWN = "unknown"

CONNECT_TIMEOUT_SECONDS = 5
RESPONSE_TIMEOUT_SECONDS = 5
FILE_UPLOAD_TIMEOUT_SECONDS = 500

logger = logging.getLogger(__name__)


class GcodeType(IntEnum):
    PLAIN_GCODE = 0
    GZIP_GCODE = 1


class PrinterApi:
    def __init__(self, host, credentials_store=None) -> None:
        self.host = host
        self.url = "http://" + host + "/api/v1"
        self.credentials = credentials_store

    def authenticate(self,
                     exclusion_key: str="",
                     host_name: str="",
                     api_appname="Cura Connect",
                     api_user="ultimaker") -> None:
        try:
            data = {
                "application": api_appname,
                "user": api_user
            }
            if host_name is not None and host_name != "":
                data["host_name"] = host_name
            if exclusion_key is not None and exclusion_key != "":
                data["exclusion_key"] = exclusion_key

            auth_request_url = self.url + "/auth/request"
            response = requests.post(auth_request_url, data=data)

            response_data = response.json()

            self.credentials.id = response_data["id"]
            self.credentials.key = response_data["key"]
            self.credentials.write()
        except RequestException as request_exception:
            logger.warning("Failed to authenticate printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def authenticating(self) -> bool:
        return self.credentials.id is not None \
               and self.credentials.key is not None \
               and not self.credentials.authorized

    def authenticated(self) -> bool:
        return self.credentials.id is not None \
               and self.credentials.key is not None \
               and self.credentials.authorized

    def wait_until_authenticated(self, timeout) -> bool:
        try:
            start_time = time.monotonic()
            while time.monotonic() - start_time < timeout:
                message = self.__fetch_authentication_status()

                if message == AUTH_AUTHORIZED:
                    self.credentials.authorized = True
                    self.credentials.write()
                    return True
                elif message == AUTH_UNAUTHORIZED:
                    return False
                elif message == AUTH_UNKNOWN:
                    time.sleep(1)
                else:
                    raise ValueError("Unexpected authorization result: " + message)

        except RequestException:
            logger.warning("Failed to fetch authentication status for printer at %s" % self.url)

        self.credentials.reset()
        return False

    def create_job(self, job_name: str, file_path: str) -> str:
        try:
            data = {"jobname": job_name}
            files = {"file": open(file_path, "rb")}

            dog = watchdog.get_instance()
            dog.delay_monitoring_by(FILE_UPLOAD_TIMEOUT_SECONDS)

            response = self.__post(
                "/print_job",
                data=data,
                files=files,
                response_timeout=FILE_UPLOAD_TIMEOUT_SECONDS)

            dog.tick()

            if response.status_code == http.client.CREATED:
                return response.json()["uuid"]
            raise PrinterApiHttpException(
                response, "URL %s responded with status code: %s" % (str(response.url), response.status_code))

        except RequestException as request_exception:
            logger.warning("Failed to create job on printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def get_status(self) -> Status:
        try:
            response = self.__get("/printer")
            if response.status_code == http.client.OK:
                return Status(response.json())
            raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to fetch status for printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def get_system(self) -> System:
        try:
            response = self.__get("/system")
            if response.status_code == http.client.OK:
                return System(response.json())
            raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to fetch system info for printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def set_system_time(self, new_system_time: float):
        try:
            response = self.__put("/system/time/utc", new_system_time)
            if response.status_code == http.client.OK:
                return System(response.json())
            raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to fetch system info for printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def get_current_print_job(self) -> PrintJob:
        try:
            response = self.__get("/print_job")
            if response.status_code == http.client.OK:
                return PrintJob(response.json())
            raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to fetch status for print job at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def get_print_job_from_history(self, print_job_id: str) -> PrintJob:
        try:
            response = self.__get("/history/print_jobs/" + print_job_id)
            if response.status_code == http.client.OK:
                return PrintJob(response.json())
            raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to fetch status for historical print job %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def download_print_job(self, destination_filename: str) -> GcodeType:
        try:
            response = self.__get("/print_job/gcode", stream=True)
            if response.status_code == http.client.OK:
                content_type = response.headers.get("Content-Type", "")
                if content_type == "text/plain" or content_type.startswith("text/plain;"):
                    gcode_type = GcodeType.PLAIN_GCODE
                elif content_type == "application/gzip":
                    gcode_type = GcodeType.GZIP_GCODE
                else:
                    raise PrinterApiHttpException(response)

                with open(destination_filename, 'wb') as fhandle:
                    for chunk in response.iter_content(chunk_size=64 * 1024):
                        fhandle.write(chunk)
                return gcode_type
            else:
                raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to download print job gcode from %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def get_camera_feed_url(self) -> str:
        try:
            response = self.__get("/camera/feed")
            if response.status_code == http.client.OK:
                return response.json()
        except RequestException:
            logger.warning("Failed to fetch camera feed url %s" % self.url)
        return None

    def set_led_color(self, color: LedColor) -> None:
        try:
            response = self.__put("/printer/led", {
                "hue": color.hue,
                "saturation": color.saturation,
                "brightness": color.value
            })
            if response.status_code != http.client.NO_CONTENT:
                raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to set the LED color for printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def set_display_message(self, message: str, button_text: str) -> None:
        try:
            response = self.__put("/system/display_message", {"message": message, "button_caption": button_text})
            if response.status_code == http.client.OK:
                return
            raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to display message on printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def request_print_job_state(self, target_state: str) -> None:
        try:
            response = self.__put("/print_job/state", {"target": target_state})
            if response.status_code == http.client.NO_CONTENT:
                return
            raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to set print job state on printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def get_maintenance_history(self, offset: Optional[int]=None, count: Optional[int]=None) -> List[Maintenance]:
        try:
            data = {"type_id": Maintenance.type_id}  # add the filter to only get maintenance type events
            if offset:
                data.update({"offset": offset})
            if count:
                data.update({"count": count})
            response = self.__get("/history/events?" + "&".join(["{}={}".format(key, data.get(key)) for key in data]))
            if response.status_code == http.client.OK:
                return [Maintenance(maintenance) for maintenance in response.json()]
            raise PrinterApiHttpException(response)
        except RequestException as request_exception:
            logger.warning("Failed to fetch maintenance log for printer at %s", self.url)
            raise PrinterApiNetworkException(request_exception)

    def add_maintenance_history_item(self, action: str, mechanic: Optional[str]=None) -> bool:
        try:
            data = {"type_id": Maintenance.type_id, "parameters": [action]}  # type:Dict[str,Any]
            if mechanic:
                data["parameters"].append(mechanic)

            response = self.__post("/history/events", data)

            if response.status_code == http.client.OK:
                return True
            raise PrinterApiHttpException(
                response, "URL %s responded with status code: %s" % (str(response.url), response.status_code))

        except RequestException as request_exception:
            logger.warning("Failed to create new maintenance action on printer at %s" % self.url)
            raise PrinterApiNetworkException(request_exception)

    def __get(self,
              path: str,
              data: Any=None,
              connect_timeout: float=CONNECT_TIMEOUT_SECONDS,
              response_timeout: float=RESPONSE_TIMEOUT_SECONDS,
              url_override: Optional[str] = None,
              stream=False) -> Response:
        return self.__request("get",
                              path,
                              data=data,
                              connect_timeout=connect_timeout,
                              response_timeout=response_timeout,
                              url_override=url_override,
                              stream=stream)

    def __put(self, path: str, json: Any) -> Response:
        return self.__request("put", path, json)

    def __post(self, path: str, json: Any=None, **kwargs) -> Response:
        return self.__request("post", path, json, **kwargs)

    def __request(self,
                  method: str,
                  path: str,
                  json: Any=None,
                  connect_timeout: float=CONNECT_TIMEOUT_SECONDS,
                  response_timeout: float=RESPONSE_TIMEOUT_SECONDS,
                  url_override: Optional[str]=None,
                  **kwargs) -> Response:
        auth = None
        if self.credentials:
            auth = HTTPDigestAuth(self.credentials.id, self.credentials.key)
        new_path = url_override + path if url_override is not None else self.url + path
        result = requests.request(method,
                                  new_path,
                                  auth=auth,
                                  json=json,
                                  timeout=(connect_timeout, response_timeout),
                                  **kwargs)
        if result.status_code == 403:
            # Authentication failure, credentials are no longer valid. So we have to drop those.
            logger.warning("Authorization failed for printer at %s%s, disabling credentials", self.url, path)
            self.credentials.reset()
        return result

    def verify_authentication(self) -> bool:
        ret = False
        response = self.__get("/auth/verify", data={})
        if response.status_code == http.client.OK:
            ret = response.json().get("message") == "ok"

        logger.debug("verifying %s!", "succeeded" if ret else "FAILED")
        return ret

    def __fetch_authentication_status(self) -> Optional[str]:
        response = self.__get("/auth/check/%s" % (self.credentials.id))
        if response.status_code == http.client.OK or response.status_code == http.client.UNAUTHORIZED:
            auth_status = response.json()
            return auth_status["message"]
        return None

    def is_hosting_multiple_printers(self) -> bool:
        url = "http://" + self.host + "/cluster-api/v1"
        response = self.__get(path="/printers/", url_override=url)
        if response.status_code == http.client.OK:
            return len(response.json()) > 1
        else:
            raise PrinterApiException
