import threading
import traceback
import sdnotify
import logging
import time
import sys
from typing import Dict, Any, Optional

from um3_commons.thread import Thread

logger = logging.getLogger('app.libs.watchdog')

SLEEP_SECONDS = 10


class MonitoredThread:
    def __init__(self, *, alive: bool, delay: Optional[float] = None) -> None:
        self.alive = alive  # type: bool
        self.delay = delay  # type: Optional[float]


class Watchdog:
    """Application Watchdog which communicates with the system watchdog.

    Threads in the application should register themselves with this object
    by calling `start_monitoring()`, and then regularly call `tick()` to
    signal that they are still running. Threads which have no work to do
    or will be idle for long periods of time, can unregister by calling
    `stop_monitoring()`.
    """

    def __init__(self, use_thread=True) -> None:
        self.__lock = threading.Lock()
        self.__monitored_threads = {}   # type: Dict[threading.Thread, MonitoredThread]
        self.__heartbeat_thread = None  # type: Thread

        self.__use_thread = use_thread

    def _get_current_thread(self) -> threading.Thread:
        return threading.current_thread()

    def start_monitoring(self) -> None:
        """Start monitoring the current thread
        The calling thread will now be required to check in with the system
        by regularly calling the tick() method. Once the calling thread has
        finished its work, it must also call stop_monitoring().
        """
        with self.__lock:
            current_thread = self._get_current_thread()
            if current_thread not in self.__monitored_threads:
                self.__monitored_threads[current_thread] = MonitoredThread(alive=True)
            else:
                logger.error("watchdog.start_monitoring() was called twice without a stop_monitoring()")

    def stop_monitoring(self) -> None:
        """Stop monitoring the current thread.
        """
        with self.__lock:
            current_thread = self._get_current_thread()
            try:
                del self.__monitored_threads[current_thread]
            except KeyError:
                logger.error("watchdog.stop_monitoring() was called before start_monitoring() by %s", str(current_thread))

    def tick(self) -> None:
        """Signal to the watchdog that the current thread is still alive.

        Ticking the thread will cause the delay to be reset.
        """
        with self.__lock:
            current_thread = self._get_current_thread()
            if current_thread in self.__monitored_threads:
                monitored_thread = self.__monitored_threads[current_thread]
                monitored_thread.delay = None
                monitored_thread.alive = True
            else:
                logger.error("watchdog.tick() was called before start_monitoring() by %s", str(current_thread))

    def check_health(self) -> bool:
        with self.__lock:
            result = self.__is_healthy()
            self.__reset_flags()
            return result

    def delay_monitoring_by(self, timeout: float) -> None:
        """ Please note that the thread will be considered alive for 10-20 seconds longer then indicated.
        """

        current_thread = self._get_current_thread()
        if current_thread in self.__monitored_threads:
            delay = time.monotonic() + timeout
            self.__monitored_threads[current_thread].delay = delay

            logger.info("Thread {} is long running and allowed to be unresponsive for {} minutes.".format(
                str(current_thread),
                int(timeout / 60))
            )

    def __is_healthy(self) -> bool:
        healthy = True
        for thread, monitor_object in self.__monitored_threads.items():
            if not monitor_object.alive:
                logger.error("Thread '%s' failed to tick the watchdog! Alive: %s", str(thread), thread.is_alive())
                if thread.is_alive():
                    # Get the stack trace of the blocked thread and log it. This should help in debugging.
                    current_frames = sys._current_frames()
                    if thread.ident in current_frames:
                        for line in traceback.format_stack(current_frames[thread.ident]):
                            logger.error(line.rstrip())
                healthy = False
        return healthy

    def __reset_flags(self) -> None:
        for thread, monitor_object in self.__monitored_threads.items():
            if monitor_object.delay is not None and monitor_object.delay >= time.monotonic():
                continue

            self.__monitored_threads[thread].alive = False

    def application_started(self) -> None:
        logger.info("Signaling application started")
        if self.__use_thread:
            if self.__heartbeat_thread is None:
                self.__start_watchdog_thread()
            else:
                logger.error("application_started() was called more than once")

    def __start_watchdog_thread(self) -> None:
        self.__heartbeat_thread = Thread(name="Watchdog", target=self.__watchdog_thread_entry)
        self.__heartbeat_thread.start()

    def __watchdog_thread_entry(self) -> None:
        notifier = sdnotify.SystemdNotifier()
        notifier.notify("READY=1")

        while True:
            time.sleep(SLEEP_SECONDS)
            if self.check_health():
                notifier.notify("WATCHDOG=1")


watchdog_instance = None


def get_instance(use_thread=True) -> Watchdog:
    global watchdog_instance
    if watchdog_instance is None:
        watchdog_instance = Watchdog(use_thread)
    return watchdog_instance


def stop() -> None:
    global watchdog_instance
    watchdog_instance = None
