from collections import OrderedDict
from threading import RLock
from typing import TypeVar, Generic, Dict, Tuple, Sized

import logging

from app.libs.signal import Signal
from app.models.abstraction.abstract_model import AbstractModel

T = TypeVar('T', bound=AbstractModel)
logger = logging.getLogger(__name__)


class AbstractCollectionModel(Generic[T], Sized):
    def __init__(self) -> None:
        self.onAdd = Signal[T]()
        self.onMove = Signal[Tuple[T, int]]()
        self.onDelete = Signal[T]()
        self.__db = OrderedDict()  # type: OrderedDict[str, T]
        self.__lock = RLock()

    def add(self, model: T) -> None:
        with self.__lock:
            assert model.uuid not in self.__db, "the added model can not already exist!"
            self.__db[model.uuid] = model
        self.onAdd.emit(model)

    def delete(self, model: T) -> None:
        deleted = True
        with self.__lock:
            try:
                del self.__db[model.uuid]
            except KeyError:
                deleted = False
        if deleted:
            model.onChange.clear_connections()
            self.onDelete.emit(model)

    def move(self, model: T, to_position: int) -> None:
        assert to_position == 0, "Not supported: move print job [%s] to position [%s]".format(model.uuid, to_position)
        with self.__lock:
            if model.uuid not in self.__db:
                logger.warning("Ignored mode command of object(%s) not in this collection(%s)", model, self)
                return  # Ignore move of object not in the collection.
            self.__db.move_to_end(model.uuid, last=False)
        self.onMove.emit((model, to_position))

    @property
    def data(self) -> Dict[str, T]:
        return self.__db.copy()

    def get(self, uuid: str) -> T:
        return self.__db.get(uuid)

    def __len__(self):
        return len(self.__db)
