import copy
import inspect
import logging
from collections import OrderedDict
from threading import RLock
from typing import Any, Dict, Type, Tuple

from app.libs.signal import Signal
from app.models.abstraction._atomic_modification_context import _AtomicModificationContext
from app.models.abstraction.model_field import ModelField

logger = logging.getLogger(__name__)


# This class should be used as the base class for any model class that you want to store in a NVM/backing store.
# any Modelfield defined in the subclass is stored along with the uuid for each instance of the Model.
# These resulting Models should be stored in a variation of the abstract collection model.
# A storage manager can then make sure the data is saved and restored automatically.
class AbstractModel:
    uuid = ModelField[str](data_type=str, read_only=True)

    def __init__(self) -> None:
        # Obviously Type Any[AbstractModel] is weird, but if you constrain the type to be abstract model,
        # then expecting a printJob gives errors....
        self.onPreChange = Signal[Tuple[Any, Dict[str, Any]]]()
        self.onChange = Signal[Tuple[Any, Dict[str, Any]]]()

        # This context allows you to set a number of modelfield Type
        # properties in one action, instead of as sepparate set actions.
        self.atomic_modification = _AtomicModificationContext()
        self.atomic_modification.on_release.connect(self.__handle_atomic_context_exit)

        # This context can be used on construction to set the properties of storage backed modelfields
        # to the value they should be initially, so that there are not models with half valid data.
        self.initial_properties = _AtomicModificationContext()
        self.initial_properties.on_release.connect(self.__handle_initial_properties_context_exit)

        self.__properties = OrderedDict()  # type: OrderedDict[str, Any]
        self.__lock = RLock()
        self.__backed_properties = self.__initialize_model_fields()  # type: Dict[str, Type[Any]]

    def _get_via_ModelField(self, property_name: str) -> Any:
        with self.__lock:
            return self.__properties[property_name]

    def _set_via_ModelField(self, property_name: str, value: Any) -> None:
        changed = False
        with self.__lock:
            if self.__properties.get(property_name) != value:
                changed = True
        if changed:
            self.onPreChange.emit((self, {property_name: value}))

            with self.__lock:
                self.__properties[property_name] = value

            self.onChange.emit((self, {property_name: value}))

    def __handle_atomic_context_exit(self, properties: Dict[str, Any]) -> None:
        changed = False
        with self.__lock:
            copy = properties.copy()
            copy.update(properties)
            if copy != self.__properties:
                changed = True
        if changed:
            self.onPreChange.emit((self, properties))

            with self.__lock:
                self.__properties.update(properties)

            self.onChange.emit((self, properties))

    def __handle_initial_properties_context_exit(self, properties: Dict[str, Any]) -> None:
        with self.__lock:
            assert len(self.__properties) == 0, "properties should not exist yet when initializing them!"
            self.__properties.update(properties)
            assert self.__properties.get("uuid") is not None

        with self.__lock:
            for key, propery_type in self.__backed_properties.items():
                value = self.__properties.get(key)
                if value is None:
                    try:
                        value = propery_type()
                        logger.warning("generating %s as %s -> %s", key, type(value), value)
                    except TypeError:
                        # The type can't be constructed.
                        pass
                self.__properties[key] = value

    # this returns an ordered dict because typing does not work for the ordered dict class;
    # So we type it as a dict[str, Any] instead!
    def to_dict(self) -> Dict[str, Any]:
        with self.__lock:
            return copy.deepcopy(self.__properties)

    def __initialize_model_fields(self) -> Dict[str, Type[Any]]:
        backed_properties = {}  # type: Dict[str, Type[Any]]
        for member_name, member_object in inspect.getmembers(self.__class__):
            if inspect.isdatadescriptor(member_object) and isinstance(member_object, ModelField):
                logger.debug("%s: %s", member_name, type(member_object))
                member_object.label = member_name
                backed_properties[member_object.key] = member_object.type

        return backed_properties
