import uuid

import logging
import os
import shutil
import threading
from typing import Any, Dict, List, Optional, Tuple, Set

from app.constants.print_job_status import PrintJobStatus
from app.libs.signal import Signal
from app.libs.ultimaker3.print_job import PrintJob as UM3PrintJob
from app.models.print_job import PrintJob
from app.models.print_job_collection import PrintJobCollection
from app.models.printer import Printer
from app.services.alien_print_job_downloader import AlienPrintJobDownloader
from app.services.material_manager import MaterialManager
from app.services.printer_monitor_service import PrinterMonitorService
from app.services.printer_service import PrinterService
from um3_commons import thread

logger = logging.getLogger(__name__)


class PrintJobService:
    def __init__(
            self,
            *,
            collection: PrintJobCollection,
            printer_monitor_service: PrinterMonitorService,
            printer_service: PrinterService,
            application_started_signal: Signal[Tuple],
            gcode_spool_directory: str,
            material_manager: MaterialManager = None
    ) -> None:
        application_started_signal.connect(self.__application_started)
        self._printer_service = printer_service
        self._printer_service.printer_is_about_to_be_removed.connect(
            self.__handle_printer_is_about_to_be_removed)

        self._print_jobs = collection   # type: PrintJobCollection
        self.__lock = thread.RLock()    # Used to guard write access to _print_jobs

        self.__gcode_spool_directory = gcode_spool_directory

        self.print_job_added = Signal[PrintJob]()
        self.print_job_updated = Signal[PrintJob]()
        self.print_job_removed = Signal[PrintJob]()
        self._material_manager = material_manager

        self._printer_monitor_service = printer_monitor_service
        self._printer_monitor_service.print_job_status_updated.connect(
            self.__handle_print_job_status_updated)
        self._printer_monitor_service.print_job_status_unavailable.connect(
            self.__handle_print_job_status_unavailable)

        self.print_jobs.onAdd.connect(self.__handle_print_job_added)
        self.print_jobs.onDelete.connect(self.__delete_print_job_gcode)
        self._printer_monitor_service.print_job_alien_detected.connect(
            self.__handle_print_job_alien_detected
        )
        self.__alien_print_job_downloaders_lock = threading.RLock()
        self.__alien_print_job_downloaders = set()  # type: Set[str]

    def __application_started(self, payload: Tuple) -> None:
        self.__remove_bad_jobs()
        self.__clean_up_gcode_spool_directory()
        self.__add_materials_info(self.print_jobs.all())
        self.__start_monitoring_active_print_jobs()

    def __remove_bad_jobs(self) -> None:
        for print_job in self.print_jobs.all():
            try:
                gcode_handle = open(print_job.filepath, "rb")
                gcode_handle.close()
            except FileNotFoundError:
                self.print_jobs.delete(print_job)

    def __clean_up_gcode_spool_directory(self) -> None:
        valid_directory_names = set([print_job.uuid for print_job in self.print_jobs.all()])

        for item in os.listdir(self.__gcode_spool_directory):
            if item not in valid_directory_names:
                item_path = os.path.join(self.__gcode_spool_directory, item)
                logger.info("Deleting unneeded directory or file '%s'.", item_path)
                shutil.rmtree(item_path, ignore_errors=True)

    def __start_monitoring_active_print_jobs(self) -> None:
        for print_job in self.print_jobs.printing():
            self.__start_monitoring_print_job(print_job)

    @property
    def print_jobs(self) -> PrintJobCollection:
        return self._print_jobs

    def get(self, uuid: str) -> Optional[PrintJob]:
        return self._print_jobs.get(uuid)

    def __handle_print_job_added(self, print_job: PrintJob) -> None:
        self.__add_material_info_to_print_job(print_job)
        print_job.onChange.connect(self.__handle_print_job_updated)
        print_job.onPreChange.connect(self.__pre_handle_print_job_updated)
        self.print_job_added.emit(print_job)

    def __start_monitoring_print_job(self, print_job: PrintJob) -> None:
        printer = self._printer_service.get(print_job.printer_uuid)

        if printer is None:
            logger.warning("job %s is being printed on an unknown printer(uuid=%s)", print_job, print_job.printer_uuid)
            if print_job.status in [PrintJobStatus.WAIT_CLEANUP, PrintJobStatus.WAIT_USER_ACTION]:
                self.print_jobs.delete(print_job)
        else:
            self._printer_monitor_service.start_monitoring_print_job(
                printer=printer,
                print_job_uuid=print_job.uuid,
                external_uuid=print_job.external_uuid
            )

    def set_force_flag(self, uuid: str, force_flag: bool) -> None:
        logger.debug("Update print job %s with force flag %s", uuid, force_flag)
        print_job = self.get(uuid)
        if print_job is None:
            logger.debug("print job %s not found" % uuid)
            return

        with print_job.atomic_modification:
            if not print_job.constraints.get("require_printer_name"):
                printer = print_job.assigned_to_printer
                if printer is not None:
                    print_job.constraints.update({"require_printer_name": printer.unique_name})
                else:
                    logger.warning("Ignoring this update job: %s, force flag: %s," +
                                   "because force is set without knowing which printer to force to.",
                                   print_job, force_flag)
                    return
            print_job.force = True

    def __handle_print_job_updated(self, signal_payload: Tuple[PrintJob, Dict[str, Any]]) -> None:
        print_job, properties = signal_payload
        if "status" in properties.keys():
            if print_job.status in PrintJobStatus.COMPLETED_STATUSES:
                self.print_jobs.delete(print_job)
            elif print_job.status == PrintJobStatus.QUEUED:
                self._printer_monitor_service.stop_monitoring_print_job(print_job.uuid)
            if print_job.status == PrintJobStatus.FAILED:
                print_job.status = PrintJobStatus.QUEUED
            if print_job.status in PrintJobStatus.SENT_TO_PRINTER:
                self.__start_monitoring_print_job(print_job)
        self.print_job_updated.emit(print_job)

    def __pre_handle_print_job_updated(self, signal_payload: Tuple[PrintJob, Dict[str, Any]]) -> None:
        print_job, properties = signal_payload
        if "status" in properties.keys():
            logger.info("PrintJob [%s] status %s -> %s", print_job, print_job.status, properties.get("status"))

    @staticmethod
    def _compare_print_job_properties_to_dict(print_job: PrintJob, properties: Dict[str, Any]) -> bool:
        for key, value in properties.items():
            existing_value = getattr(print_job, key)
            if existing_value != value:
                return True
        return False

    def request_print_job_state(self, print_job_id: str, new_state: str) -> bool:
        logger.debug("print job [%s], set new state [%s]", print_job_id, new_state)
        if new_state not in {"pause", "print", "abort"}:
            logger.warning("New state [%s] is invalid", new_state)
            return False
        print_job = self._print_jobs.get(print_job_id)
        if print_job is None:
            logger.warning("Print job [%s] does not exist", print_job_id)
            return False
        if print_job.printer_uuid is None:
            logger.warning("Print job does not have a printer. %s", print_job)
            return False
        printer = self._printer_service.get(print_job.printer_uuid)
        if printer is None:
            logger.warning("Printer [%s] does not exist", print_job.printer_uuid)
            return False
        logger.info("requesting PrintJob %s to %s on printer %s", print_job, new_state, printer)
        return printer.request_print_job_state(new_state)

    def __delete_print_job_gcode(self, print_job: PrintJob) -> None:
        if print_job is not None and os.path.exists(print_job.filepath):
            logger.info("removed gcode file: %s", print_job.filepath)
            os.remove(print_job.filepath)
            dir_name = os.path.dirname(print_job.filepath)
            if dir_name.endswith(print_job.uuid):
                logger.info("removed uuid dir: %s", dir_name)
                os.removedirs(dir_name)  # delete uuid directory
            self.print_job_removed.emit(print_job)

    def __handle_printer_is_about_to_be_removed(self, printer: Printer) -> None:
        for job in self._print_jobs.all():
            if job.printer_uuid == printer.uuid:
                logger.debug("printer removed: %s, remove print_job: %s", str(printer), job.uuid)
                self._print_jobs.delete(job)

    def __handle_print_job_status_unavailable(self, signal_payload: Tuple[Printer, str]) -> None:
        logger.debug("Print job status is unavailable")
        printer_address, print_job_uuid = signal_payload
        print_job = self.get(print_job_uuid)
        if print_job is not None:
            print_job.status_check_failed()
        else:
            logger.debug("Print job appears to be deleted.")

    def __handle_print_job_alien_detected(self, signal_payload: Tuple[Printer, UM3PrintJob]) -> None:
        printer, um3_print_job = signal_payload
        logger.debug("Detected alien print job with external uuid [%s]", um3_print_job.uuid)
        logger.debug("um3_print_job.state [%s]",um3_print_job.state)

        with self.__alien_print_job_downloaders_lock:
            external_uuid = um3_print_job.uuid
            # A job download can be in progress
            if external_uuid in self.__alien_print_job_downloaders:
                return
            # A job could have been created during the poll cycle in that case we catch it here.
            if um3_print_job.uuid in [job.external_uuid for job in self.print_jobs.all()]:
                logger.debug("Detected alien print job was sent by us! Don't download job with external uuid [%s]",
                             um3_print_job.uuid)
                return

            job_internal_uuid = str(uuid.uuid4())

            if um3_print_job.reprint_original_uuid != "":
                owner = "reprint"
            else:
                owner = um3_print_job.source_user if um3_print_job.source_user != "" else um3_print_job.source

            def finished_download(gcode_filename: str) -> None:
                print_job = PrintJob(
                    uuid=job_internal_uuid,
                    name=um3_print_job.name,
                    filepath=gcode_filename,
                    constraints={"require_printer_name": printer.unique_name},
                    force=True,
                    owner=owner,
                    status=um3_print_job.state,
                    printer_uuid=printer.uuid,
                    external_uuid=um3_print_job.uuid
                )
                self._print_jobs.add(print_job)

                print_job = self.get(job_internal_uuid)
                self.__start_monitoring_print_job(print_job)

                with self.__alien_print_job_downloaders_lock:
                    self.__alien_print_job_downloaders.discard(external_uuid)

            def failed_download(_) -> None:
                with self.__alien_print_job_downloaders_lock:
                    self.__alien_print_job_downloaders.discard(external_uuid)

            downloader = AlienPrintJobDownloader(printer, self.__gcode_spool_directory,
                                                 job_internal_uuid, um3_print_job.name)
            downloader.finished.connect(finished_download)
            downloader.failed.connect(failed_download)

            self.__alien_print_job_downloaders.add(external_uuid)

    def __handle_print_job_status_updated(self, signal_payload: Tuple[Printer, str, UM3PrintJob]) -> None:
        printer, print_job_uuid, network_print_job_status = signal_payload
        print_job = self.get(print_job_uuid)
        if print_job is None:
            return
        print_job.status_check_successful()

        if print_job.status != network_print_job_status.state:
            print_job.status = network_print_job_status.state

            if network_print_job_status.state in PrintJobStatus.COMPLETED_STATUSES:
                self._printer_monitor_service.stop_monitoring_print_job(print_job_uuid)

        print_job.time_elapsed = network_print_job_status.time_elapsed
        if network_print_job_status.time_total != 0:
            print_job.time_total = network_print_job_status.time_total
            printer.time_remaining = max(0, print_job.time_total - print_job.time_elapsed)

    def __add_materials_info(self, print_jobs: List[PrintJob]) -> None:
        """
        Sets print_job.configuration, which is print_job.extruder_trains in a list, with material info.
        Currently needed by Web UI
        """
        for print_job in print_jobs:
            self.__add_material_info_to_print_job(print_job)

    def __add_material_info_to_print_job(self, print_job: PrintJob) -> None:
        if print_job.configuration.is_empty():
            try:
                for extruder_index, extruder_train in print_job.extruder_trains.items():
                    extruder_train.material = self._material_manager.get(extruder_train.material_guid)
                    print_job.configuration.update(extruder_train)
            except FileNotFoundError:
                logger.exception("Exception trying to read extruder_trains.")
