from typing import List, Dict, Any, Set

from app.models.print_job import PrintJob
from app.models.printer import Printer


# This class is a container for the result of the planning step in the scheduler
class PlannerResultSchedule:
    def __init__(self) -> None:
        self.__print_jobs_per_printer = {}  # type: Dict[Printer, List[PrintJob]]

        # We keep track of all print jobs to ensure jobs are only added once, and so we can quickly check if a job
        # is in the schedule.
        self.__all_print_jobs = set()  # type: Set[PrintJob]

    def get_schedule_count(self, printer: Printer) -> int:
        if printer not in self.__print_jobs_per_printer:
            return 0
        return len(self.__print_jobs_per_printer[printer])

    def add_to_schedule(self, printer: Printer, print_job: PrintJob) -> None:
        assert print_job not in self.__all_print_jobs
        self.__all_print_jobs.add(print_job)
        if printer not in self.__print_jobs_per_printer:
            self.__print_jobs_per_printer[printer] = []
        self.__print_jobs_per_printer[printer].append(print_job)

    def has(self, other: PrintJob) -> bool:
        return other in self.__all_print_jobs

    def jobs_for_printer(self, printer: Printer) -> List[PrintJob]:
        return self.__print_jobs_per_printer.get(printer, [])

    def debug_get_schedule(self) -> List[Dict[str, Any]]:
        # extra typing line needed because typing doesn't understand...
        return [{"printer": printer, "jobs": list} for printer, list in self.__print_jobs_per_printer.items()]

    def __repr__(self) -> str:
        result = ["{"]
        for printer, jobs in self.__print_jobs_per_printer.items():
            result.append("%s: %s" % (printer, jobs))
        result.append("}")
        return "\n".join(result)
