from app.models.extruder_train import ExtruderTrain
from app.models.print_job import PrintJob
from app.models.printer import Printer
from app.libs.ultimaker3.material import Material
from app.services.scheduler.match_rule import MatchRule, Optional
from app.services.scheduler.result_containers.match_results import ConfigurationChangeMatchResult
from app.services.scheduler.result_containers.match_results import IncompatibleMatchResult
from app.services.scheduler.result_containers.match_results import MatchResult


class IsMaterialCompatibleMatchRule(MatchRule):
    def __init__(self, printcore_index: int) -> None:
        self.__printcore_index = printcore_index

    @property
    def printcore_index(self):
        return self.__printcore_index

    def apply(self, printer: Printer, print_job: PrintJob) -> Optional[MatchResult]:
        if printer.configuration.is_empty():
            return IncompatibleMatchResult(
                origin=self,
                debug_message="The printer does no have any configuration yet",
                translation_key="loading_config"
            )
        if print_job.configuration.is_empty():
            return IncompatibleMatchResult(
                origin=self,
                debug_message="The print_job does no have any configuration yet",
                translation_key="loading_config"
            )

        printer_config = printer.configuration.get(self.__printcore_index)  # type: ExtruderTrain
        job_config = print_job.configuration.get(self.__printcore_index)  # type: ExtruderTrain

        if job_config is None:
            return None

        if not self._are_materials_compatible(printer_config.material, job_config.material):
            return ConfigurationChangeMatchResult(
                change=ConfigurationChangeMatchResult.ChangeSet(
                    type_of_change="material_change",
                    index=self.__printcore_index,
                    origin_id=printer_config.material_guid,
                    target_id=job_config.material_guid),
                origin=self,
                debug_message="Material mismatch on slot {}: printer has {}, job expects {}".format(
                    self.__printcore_index, printer_config.material.pretty_string, job_config.material.pretty_string),
                translation_key="material_change"
            )

        return None

    def _are_materials_compatible(self, printer_material: Material, job_material: Material) -> bool:
        if job_material is not None and printer_material is not None:
            if job_material.brand == "Generic" and job_material.color == "Generic":
                return printer_material.material == job_material.material
            return printer_material.guid == job_material.guid
        return False
