import logging
from typing import List

from app.libs import watchdog
from app.models.print_job import PrintJob
from app.models.printer import Printer
from app.services.material_manager import MaterialManager
from app.services.print_job_service import PrintJobService
from app.services.printer_service import PrinterService
from app.services.scheduler.result_containers.planner_result_schedule import PlannerResultSchedule
from .configuration_mismatch_stringifier import ConfigurationMisMatchStringifier
from .print_job_dispatcher import PrintJobDispatcher
from .result_containers.match_results import ConfigurationChangeMatchResult
from .result_containers.match_results import FutureMatchResult
from .result_containers.match_results import MatchResult
from .result_containers.match_results import Severity
from .result_containers.match_results_containers import MatchMatrix
from .result_containers.match_results_containers import MatchResultListPerPrinter
from app.libs.printer_api import FILE_UPLOAD_TIMEOUT_SECONDS

logger = logging.getLogger(__name__)


# this gets a PlannerResultSchedule and a MatchMatrix and executes the plan contained in those.
# This is the scheduler's link back to the system and is the only class that should make changes to the model,
# such as reservations, setting impediments to printing (this means it likely can't be printed at all) or
# dispatching jobs.
class ScheduleExecutor:
    def __init__(self,
                 printer_service: PrinterService,
                 print_job_service: PrintJobService,
                 material_manager: MaterialManager) -> None:
        self.__printer_service = printer_service
        self.__print_job_service = print_job_service
        self.__configuration_mismatch_stringifier = ConfigurationMisMatchStringifier(material_manager)

        self.__dispatched_jobs = []  # type: List[PrintJobDispatcher]
        self.__request_reschedule = False

    def execute_plan_result(self, schedule: PlannerResultSchedule, match_matrix: MatchMatrix) -> None:
        self.__dispatched_jobs.clear()
        self.__request_reschedule = False

        # Mark the jobs schedulable or not depending on if they are in our schedule.
        self.__mark_impeded_jobs(schedule, match_matrix)

        # For each printer, handle the action for the first job in the schedule.
        for printer in self.__printer_service.printers.all():
            jobs = schedule.jobs_for_printer(printer)
            if len(jobs) > 0:
                first_job = jobs[0]
                for job in jobs[1:]:
                    job.configuration_changes_required = []
                match_result = match_matrix.get_match_result(printer, first_job)

                self.__execute_action_for_result(printer, first_job, match_result)
            else:
                if printer.reserved_by_job is not None:
                    printer.reserved_by_job.configuration_changes_required = []
                self.__remove_reservation(printer)

        # Starting print jobs are pushed to different threads, we want to wait until
        # they are all started before we are finished with the execution of the plan.
        if self.__dispatched_jobs:  # avoid log spam.
            dog = watchdog.get_instance()
            dog.delay_monitoring_by(FILE_UPLOAD_TIMEOUT_SECONDS)
            self.__wait_till_all_jobs_started()
            dog.tick()

    def __execute_action_for_result(self,
                                    printer: Printer,
                                    print_job: PrintJob,
                                    match_result: MatchResultListPerPrinter) -> None:

        # From our best match, find out what action will will execute.
        if match_result.has(FutureMatchResult):
            # Cannot start now, as the printer is busy, so we do nothing.
            print_job.configuration_changes_required = []
            # This could be a future result because the printer is still reserved for a different job that has moved
            # to a different printer. So we want to remove this reservation and reschedule.
            if printer.reserved_by_job is not None and printer.reserved_by_job != print_job:
                self.__remove_reservation(printer)
                self.__request_reschedule = True
        elif match_result.has(ConfigurationChangeMatchResult):
            self.__reserve_printer(printer, print_job, match_result)
            print_job.configuration_changes_required = self.__configuration_mismatch_stringifier.combine_and_name_changes(match_result)
        else:
            self.__remove_reservation(printer)
            self.__start_print_job(print_job, printer)
            print_job.configuration_changes_required = []

    def __reserve_printer(self, printer: Printer, print_job: PrintJob, match_result: MatchResultListPerPrinter) -> None:
        if printer.reserved_by is not None and printer.reserved_by == print_job.uuid:
            return
        self.__remove_reservation(printer)
        # Reserve the printer by marking it as active reservation.
        # This should put a message on the printer with the required configuration change
        logger.info("Going to reserve %s for %s", printer, print_job)
        self.__printer_service.set_reservation(
            printer=printer,
            print_job=print_job,
            user_message=self.__configuration_mismatch_stringifier.stringify(match_result))

    def __remove_reservation(self, printer: Printer) -> None:
        # Remove the reservation from the printer if it is set.
        if printer.reserved_by is None:
            return
        logger.info("Removing reservation on %s", printer)
        self.__printer_service.remove_reservation(printer)

    def __mark_impeded_jobs(self, schedule: PlannerResultSchedule, match_matrix: MatchMatrix) -> None:
        for job_match in match_matrix.get_match_result_per_job():
            if not schedule.has(job_match.print_job):
                # Cannot start now, as the printer is busy, so we do not show a configuration change.
                job_match.print_job.configuration_changes_required = []
                # get least impediments
                best_printer_matches = sorted(job_match, key=lambda match_result: match_result.score)  # type: List[MatchResultListPerPrinter]
                if len(best_printer_matches) < 1:
                    job_match.print_job.impediments_to_printing = [
                        MatchResult(
                            origin=None,
                            translation_key="no_printers",
                            severity=Severity.USER_FIXABLE,
                            debug_message="no printers available.")
                    ]
                    continue
                best_printer_match = best_printer_matches[0]
                # stringify the impediments
                impediments = [result for result in best_printer_match.results if not isinstance(result, ConfigurationChangeMatchResult)]
                if not impediments:  # empty list
                    impediments.append(MatchResult(None, "unknown reason why job is not scheduled", translation_key="unknown_error"))
                job_match.print_job.impediments_to_printing = impediments
            else:
                job_match.print_job.impediments_to_printing = []

    def __start_print_job(self, print_job: PrintJob, printer: Printer) -> None:
        # Start a print job dispatcher and add it to the list of dispatched jobs
        logger.info("Going to start %s on %s", print_job, printer)
        self.__dispatched_jobs.append(PrintJobDispatcher(self.__print_job_service, print_job, printer))

    def __wait_till_all_jobs_started(self) -> None:
        for job_dispatcher in self.__dispatched_jobs:
            job_dispatcher.join()

            # If a job was not dispatched successfully, we need to reschedule, as we want to retry
            # or even the match matrix could have been changed due to the failure mark.
            if not job_dispatcher.was_successful:
                job_dispatcher.job.log_start_error_with_printer(job_dispatcher.printer)
                self.__request_reschedule = True

    def requires_reschedule(self) -> bool:
        return self.__request_reschedule

    def has_started_jobs(self) -> bool:
        return len(self.__dispatched_jobs) > 0

    def clear_reservations(self, printers: List[Printer]) -> None:
        for printer in printers:
            self.__remove_reservation(printer)
