import logging
import time
from typing import Optional, Any

from app.libs import watchdog
from app.services.material_manager import MaterialManager
from app.services.print_job_service import PrintJobService
from app.services.printer_service import PrinterService
from app.services.scheduler.scheduler import PrintJobScheduler
from um3_commons.thread import Thread, Event

logger = logging.getLogger(__name__)


# This class is responsible for starting the scheduler and is the only link to changes in the model
# and reasons why the scheduler should be triggered.
class ScheduleRunner:
    SLEEP_SECONDS = 5

    def __init__(self,
                 printer_service: PrinterService,
                 print_job_service: PrintJobService,
                 material_manager: MaterialManager) -> None:
        self.__scheduler = PrintJobScheduler(printer_service, print_job_service, material_manager)
        self.__running = False
        self.__worker_run_event = Event()
        self.__worker_thread = None  # type: Optional[Thread]

        printer_service.printer_status_changed.connect(lambda _: self.__trigger_worker("printer status changed"))
        printer_service.printer_enabled_changed.connect(lambda _: self.__trigger_worker("printer enabled"))
        printer_service.amount_of_printers_changed.connect(lambda _: self.__trigger_worker("printer amount changed"))
        print_job_service.print_job_added.connect(lambda _: self.__trigger_worker("print job added"))
        print_job_service.print_job_updated.connect(lambda _: self.__trigger_worker("print job updated"))
        print_job_service.print_job_removed.connect(lambda _: self.__trigger_worker("print job removed"))
        print_job_service.print_jobs.onMove.connect(self.__handle_on_move)

    @property
    def scheduler(self) -> PrintJobScheduler:
        return self.__scheduler

    def start(self) -> None:
        logger.info("starting scheduleRunner")
        self.__running = True
        self.__worker_thread = Thread("schedule runner", self.__run)  # type: Thread
        self.__worker_thread.start()

    def __handle_on_move(self, _: Any) -> None:
        self.__scheduler.clear_reservations()
        self.__trigger_worker("print job moved")

    def __trigger_worker(self, reason: str) -> None:
        logger.debug("Triggering worker thread [%s]", reason)
        self.__worker_run_event.set()

    def __run(self) -> None:
        watchdog.get_instance().start_monitoring()
        while self.__running:
            watchdog.get_instance().tick()
            self.__worker_run_event.wait(timeout=self.SLEEP_SECONDS)
            if not self.__running:
                return
            watchdog.get_instance().tick()
            if self.__worker_run_event.is_set():
                self.__worker_run_event.clear()
                logger.debug("Worker thread is scheduling print jobs")
                request_reschedule = self.__scheduler.run()
                logger.debug("Worker thread is done scheduling print jobs")

                # please note: the scheduler will trigger itself until all jobs that can be sent to a printer
                # have been sent or have failed enough times to be ignored!
                if request_reschedule:
                    self.__worker_run_event.set()
                    watchdog.get_instance().tick()
                    # allow a bit of time in case the job was denied due to a momentary state.
                    time.sleep(self.SLEEP_SECONDS)

    def stop(self) -> None:
        logger.debug("stopping scheduleRunner")
        self.__running = False
        self.__worker_run_event.set()
        self.__worker_thread.join()
        self.__worker_thread = None
        logger.debug("scheduleRunner has stopped")
