import time
import logging
import uuid
import json
import requests
from typing import Dict, Any, Optional

from app.services.printer_service import PrinterService
from um3_commons.registry.registryFile import RegistryFile
from um3_commons.thread import Thread

logger = logging.getLogger(__name__)

CLUSTER_UUID = "CLUSTER_UUID"
POLL_PERIOD_SECONDS = 60 * 60 * 12     # Every 12 hours


class StatisticsLogger:
    """
    Printcluster heartbeat.
    Gathers data about the cluster and sends the request to the analytics backend.
    """

    def __init__(self, registry_file: RegistryFile, printer_service: PrinterService) -> None:
        self.__registry_file = registry_file
        self.__printer_service = printer_service

        cluster_id = self.__registry_file.getAsString(CLUSTER_UUID)  # type: Optional[str]
        if cluster_id is None:
            cluster_id = str(uuid.uuid4())
            self.__registry_file.setAsString(CLUSTER_UUID, cluster_id)

        logger.info("Stats cluster.id: %s", cluster_id)

        checkin_thread = Thread(name="StatisticsLogger", target=self.__poll_stats)
        checkin_thread.start()

    def __poll_stats(self) -> None:
        while True:
            if self.__registry_file.getAsBoolean("ALLOW_ANONYNMOUS_DATA_COLLECTION", None):
                self.__check_in()
            time.sleep(POLL_PERIOD_SECONDS)

    def __check_in(self) -> None:
        if len(self.__printer_service.printers) == 0:
            return

        checkin_dict = {}   # type: Dict[str, Any]
        checkin_dict["time_stamp"] = int(time.time())
        checkin_dict["hash"] = self.__registry_file.getAsString(CLUSTER_UUID)
        checkin_dict["printers"] = []

        for printer in self.__printer_service.printers.all():
            checkin_dict["printers"].append({
                "firmware_version": str(printer.firmware_version),
                "variant": str(printer.machine_variant),
                "language": str(printer.language),
                "uptime": printer.uptime
            })

        self.send_statistics_request(checkin_dict)

    def send_statistics_request(self, checkin_dict: Dict[str, Any]) -> None:
        try:
            checkin_contents = json.dumps(checkin_dict)
            r = requests.post("https://stats.ultimaker.com/api/printcluster",
                              data=bytes(checkin_contents, "utf-8"),
                              timeout=3)
            logger.debug("Sending analytics checkin request. HTTP status code: %s", r.status_code)
        except requests.ConnectionError:
            pass
        except requests.HTTPError:
            pass
