import logging
import time

log = logging.getLogger(__name__.split(".")[-1])


## Class that implements the low level ISP (In system programming) protocol present on all Atmel AVR chips.
#  This class is build from the documentation found in the ATMega2560 datasheet, which contains the command constants and protocol details.
#  This uses the SPI protocol to communicate with the chip. Note not to confuse SPI with ISP when reading this code.
# See ATMega2560 datasheet, page 351 for details on programming sequence, including timing and commands.
class LowLevelISP():
    PROG_ENABLE_COMMAND = b"\xAC\x53\x00\x00"
    CHIP_ERASE_COMMAND = b"\xAC\x80\x00\x00"
    PROG_POLL_RDY = b"\xF0\x00\x00\x00"

    LOAD_EXT_ADDR = b"\x4D\x00"
    LOAD_FLASH_HIGH = b"\x48\x00"
    LOAD_FLASH_LOW = b"\x40\x00"

    READ_LOCK_BITS = b"\x58\x00\x00\x00"
    READ_SIGNATURE_0 = b"\x30\x00\x00\x00"
    READ_SIGNATURE_1 = b"\x30\x00\x01\x00"
    READ_SIGNATURE_2 = b"\x30\x00\x02\x00"
    READ_LFUSE = b"\x50\x00\x00\x00"
    READ_HFUSE = b"\x58\x08\x00\x00"
    READ_EFUSE = b"\x50\x08\x00\x00"

    WRITE_LFUSE = b"\xAC\xA0\x00"
    WRITE_HFUSE = b"\xAC\xA8\x00"
    WRITE_EFUSE = b"\xAC\xA4\x00"
    WRITE_FLASH_PAGE = b"\x4C"

    READ_HIGH = b"\x28"
    READ_LOW = b"\x20"

    ## Create a new low level ISP programmer, this requires spi hardware, and a reset line.
    #  @param spi_hardware: One of the hardwareSPI or softwareSPI class instances.
    #  @param reset_hardware: string pointing to the gpio device that is used for reset.
    def __init__(self, spi_hardware, reset_hardware):
        assert spi_hardware is not None
        assert reset_hardware is not None
        
        self.__spi_device = spi_hardware
        self.__reset_device = reset_hardware

    ## Enter programming mode.
    #  @return True if entering programming mode was succesful, false if otherwise.
    def enterProgrammingMode(self):
        self.__setReset(False)
        time.sleep(0.001)       # Reset pulse needs to be at least 2.5us (ATMega2560 datasheet, page 372)
        self.__setReset(True)
        time.sleep(0.050)       # Wait for at least 20ms after the reset before sending the Programming Enable sequence.
        res = self.__spi_device.transfer(self.PROG_ENABLE_COMMAND)
        if res[2] != self.PROG_ENABLE_COMMAND[1]:   # The prog enable command will echo the 2nd byte back as 3th byte.
            log.debug("PROG_ENABLE_COMMAND response: " + ' '.join([hex(i) for i in res]))
            return False
        self.__waitWhileBusy()
        return True

    ## Leave programming mode after we entered programming mode previously by enterProgrammingMode()
    def leaveProgrammingMode(self):
        self.__setReset(False)

    ## Read the signature bytes
    #  @return bytes object of the length of 3. The signature bytes indicate which chip is connected to the programmer.
    def readSignature(self):
        return bytes([
            self.__spi_device.transfer(self.READ_SIGNATURE_0)[3],
            self.__spi_device.transfer(self.READ_SIGNATURE_1)[3],
            self.__spi_device.transfer(self.READ_SIGNATURE_2)[3],
        ])
        
    ## Read the lock byte
    #  @return integer containing the lock bits specific for this chip.
    def readLock(self):
        return self.__spi_device.transfer(self.READ_LOCK_BITS)[3]

    ## Read the chip fuse bytes. The fuse bytes configure clock source, as well as bootloader type, and are chip specific.
    #  @return bytes object with the length of 3.
    def readFuse(self):
        return bytes([
            self.__spi_device.transfer(self.READ_LFUSE)[3],
            self.__spi_device.transfer(self.READ_HFUSE)[3],
            self.__spi_device.transfer(self.READ_EFUSE)[3],
        ])

    ## Write the fuse bytes, if required. This first reads the fuse bytes to check if writing is needed.
    #  @param fuses: bytes object containing the fuse data for the fuses.
    #  @return True if the fuses where programmed succesfully.
    def writeFuse(self, fuses):
        assert(len(fuses) == 3)

        wrote_fuse = False
        current_fuses = self.readFuse()
        if current_fuses[0] != fuses[0]:
            log.info("Programming low fuse")
            wrote_fuse = True
            self.__spi_device.transfer(self.WRITE_LFUSE + fuses[0:1])
            self.__waitWhileBusy()
        if current_fuses[1] != fuses[1]:
            log.info("Programming high fuse")
            wrote_fuse = True
            self.__spi_device.transfer(self.WRITE_HFUSE + fuses[1:2])
            self.__waitWhileBusy()
        if current_fuses[2] != fuses[2]:
            log.info("Programming extended fuse")
            wrote_fuse = True
            self.__spi_device.transfer(self.WRITE_EFUSE + fuses[2:3])
            self.__waitWhileBusy()

        if wrote_fuse:
            current_fuses = self.readFuse()
        return current_fuses == fuses

    ## Read the memory between the low and high address, excluding the high address itself.
    #  @param low_addr: Integer containing the low address
    #  @param high_addr: Integer containing the high address
    #  @return bytes object containing the flash data from the low to high address range.
    def readMemory(self, low_addr, high_addr):
        ret = bytearray(high_addr - low_addr)
        last_extended_address = -1
        for address in range(low_addr, high_addr):

            extended_mem_byte = address >> 17
            if last_extended_address != extended_mem_byte:
                self.__spi_device.transfer(self.LOAD_EXT_ADDR + extended_mem_byte.to_bytes(1, "big") + b"\x00")
                last_extended_address = extended_mem_byte

            if address & 1 == 0:
                cmd = self.READ_LOW + ((address >> 1) & 0xFFFF).to_bytes(2, "big") + b"\x00"
            else:
                cmd = self.READ_HIGH + ((address >> 1) & 0xFFFF).to_bytes(2, "big") + b"\x00"
            ret[address - low_addr] = self.__spi_device.transfer(cmd)[3]
        return ret

    ## Write memory page
    #  @param address: integer, containing the absolute flash address of this page.
    #  @param page_data: bytes object containing the data for this page to write. Should be the exact lenght of a flash page for this chip.
    #  @return boolean: True if writing the page was succesful.
    def writePage(self, address, page_data):
        extended_mem_byte = address >> 17
        self.__spi_device.transfer(self.LOAD_EXT_ADDR + extended_mem_byte.to_bytes(1, "big") + b"\x00", do_read=False)

        for word_nr in range(len(page_data) // 2):
            index = word_nr * 2
            self.__spi_device.transfer(self.LOAD_FLASH_LOW + (word_nr).to_bytes(1, "big") + page_data[index: index + 1], do_read=False)
            self.__spi_device.transfer(self.LOAD_FLASH_HIGH + (word_nr).to_bytes(1, "big") + page_data[index + 1: index + 2], do_read=False)

        self.__spi_device.transfer(self.WRITE_FLASH_PAGE + ((address >> 1) & 0xFFFF).to_bytes(2, "big") + b"\x00", do_read=False)
        self.__waitWhileBusy()
        return True

    ## Get a checksum of an address range.
    #  @param low_addr: Integer containing the low address
    #  @param high_addr: Integer containing the high address
    #  @return None as checksumming is not supported by ISP.
    def getChecksum(self, low_addr, high_addr):
        return None

    ## Erase the chip
    #  After this function the chip will contain no data.
    #  Lock bits will be reset.
    #  Fuse bits are unmodified.
    def eraseChip(self):
        self.__spi_device.transfer(self.CHIP_ERASE_COMMAND)
        # IMPORTANT: Waiting till the erase is done is especially important. Reading the signature while erasing will erase the signature bytes without a way to restore them!
        self.__waitWhileBusy()

    ## Set the reset line.
    #  @param enable: boolean, True: The chip will be in reset, False, the chip will not be in reset and run the main program.
    def __setReset(self, enable):
        with open(self.__reset_device, "w") as f:
            if enable:
                f.write("0")
            else:
                f.write("1")

    ## Check if the programmer is busy.
    #  @return True if the programmer is busy and no flash commands can be executed.
    def __isBusy(self):
        ret = self.__spi_device.transfer(self.PROG_POLL_RDY)
        return (ret[3] & 0x01) == 0x01

    ## Wait while the programmer is busy.
    #  Used after we execute flash commands.
    def __waitWhileBusy(self):
        while self.__isBusy():
            pass
