#!/bin/sh

ARCHIVE_TYPE=''
TAR_STATUS=''

# Function that works like echo, but prints bold. Used to distinct between our output and output from tools we call.
info() {
	printf "$(tput bold)${*}$(tput sgr0)\n"
}

# Function that works like echo, but prints in orange. Used to make the user notice a warning.
warn() {
	printf "$(tput setaf 3)${*}$(tput sgr0)\n"
}

# Function that works like echo, but prints in red. Used to make the user notice an error.
error() {
	printf "$(tput bold)$(tput setaf 1)${*}$(tput sgr0)\n"
}

echo "==========================================="
echo "Jedi system update"
echo "==========================================="
echo ""

if [ -f /tmp/rootfs.tar.gz ]; then
	ARCHIVE_TYPE="gz"
elif [ -f /tmp/rootfs.tar.xz ]; then
	ARCHIVE_TYPE="xz"
fi
if [ -z ${ARCHIVE_TYPE} ]; then
	error "/tmp/rootfs.tar.[gz|xz] is missing"
	exit 1
fi

info "Verifying update archive integrity ..."
if [ ${ARCHIVE_TYPE} = "gz" ]; then
	tar -tzf /tmp/rootfs.tar.gz > /dev/null 2> /dev/null
	TAR_STATUS=$?
elif [ ${ARCHIVE_TYPE} = "xz" ]; then
	tar -tJf /tmp/rootfs.tar.xz > /dev/null 2> /dev/null
	TAR_STATUS=$?
fi
if [ ${TAR_STATUS} -ne 0 ]; then
	error "/tmp/rootfs.tar.${ARCHIVE_TYPE} corrupt"
	exit 1
fi

info "Making sure /tmp has enough space."
find /tmp -type f ! -name "rootfs.tar.${ARCHIVE_TYPE}" -exec rm -f {} +
mount -o remount,size=75% /tmp

info "Starting update process."
UPDATE_ROOT_DIR=$(mktemp -t -d um-updater_tmproot.XXXXXX)

info "Extracting new rootfs to ${UPDATE_ROOT_DIR}"
if [ ${ARCHIVE_TYPE} = "gz" ]; then
	tar -xzf /tmp/rootfs.tar.gz -C ${UPDATE_ROOT_DIR} > /dev/null 2>&1
	TAR_STATUS=$?
elif [ ${ARCHIVE_TYPE} = "xz" ]; then
	tar -xJf /tmp/rootfs.tar.xz -C ${UPDATE_ROOT_DIR} > /dev/null 2>&1
	TAR_STATUS=$?
fi
if [ ${TAR_STATUS} -ne 0 ]; then
	error "Unable to extract /tmp/rootfs.tar.${ARCHIVE_TYPE}"
	exit 1
fi

run_update_scripts() {
	update_scripts_todo=${*}
	if [ -z "${update_scripts_todo}" ]; then
		return 0
	fi

	for update_script in ${update_scripts_todo}; do
		if [ -x "${update_script}" ]; then
			echo "Running update script '${update_script}' ... "
			"${update_script}" ${UPDATE_ROOT_DIR}
			retval=$?
			if [ ${retval} -eq 0 ]; then
				info "[OK]"
				updates_from_tar=$((updates_from_tar + 1))
			else
				error "[FAILED]"
			fi
		else
		    warn "Cannot execute script '${update_script}' ... "
		fi
	done
}

# Run update scripts from archive
updates_from_tar=0
run_update_scripts "${UPDATE_ROOT_DIR}/usr/lib/systemupdate/updates/*.preup"
run_update_scripts "${UPDATE_ROOT_DIR}/usr/lib/systemupdate/updates/*.sh"
run_update_scripts "${UPDATE_ROOT_DIR}/usr/lib/systemupdate/updates/*.postup"
echo "Successfully ran ${updates_from_tar} updates from within the update the archive."

# If there where no updates in the update, we are downgrading.
# Downgrading is not really supported, but can happen!
# Lets do our best however.
if [ ${updates_from_tar} -eq 0 ]; then
	warn "Detected downgrade. Operation cannot always be guaranteed. Doing our best!"
	# Prevent issues such as recursively calling itself
	if [ ! -f "${UPDATE_ROOT_DIR}/usr/lib/systemupdate/initiate_update.sh" ]; then
		error "No update downgrade script available."
	else
		UPDATESCRIPT_FILE=$(mktemp -t systemupdate.XXXXXX.sh)
		cp "${UPDATE_ROOT_DIR}/usr/lib/systemupdate/initiate_update.sh" ${UPDATESCRIPT_FILE}
		CP_STATUS=$?
		chmod a+x ${UPDATESCRIPT_FILE}
		CHMOD_STATUS=$?
		if [ ${CP_STATUS} -eq 0 ] && [ ${CHMOD_STATUS} -eq 0 ]; then
			run_update_scripts "${UPDATESCRIPT_FILE}"
		else
			error "Could not copy or execute initiate_update.sh script in tmp location."
		fi
	fi
fi

info "Rebooting system"
echo "==========================================="
systemctl reboot
sleep 20
exit 0
