import logging
import re

from griffin import dbusif

log = logging.getLogger(__name__.split(".")[-1])


## Interface to the camera attached to the system
#  This service detects which cameras are present in the system, and provides a uniform API to access information about them.
#  TODO: Somewhere in the future this should also allow the cameras to be configured for brightness and stuff.
class CameraService(dbusif.ServiceObject):
    def __init__(self):
        super().__init__("camera")

        self.__cameras = {}

        self.__systemd_manager = dbusif.RemoteObject("org.freedesktop.systemd1", "/org/freedesktop/systemd1", "org.freedesktop.systemd1.Manager")
        self.__systemd_manager.connectSignal("UnitNew", self.__newSystemdUnit)
        self.__systemd_manager.connectSignal("UnitRemoved", self.__removeSystemdUnit)
        self.__systemd_manager.Subscribe()  # We need to call Subscribe to be sure that we get signals
        for unit_id, description, load_state, state, sub_state, follow_unit, path, job_id, job_type, job_path in self.__systemd_manager.ListUnits():
            self.__newSystemdUnit(unit_id, path)

    @dbusif.method("", "a{ia{ss}}")
    def getCameras(self):
        return self.__cameras

    ## Event from systemd when a new unit is created. This happens when hardware is inserted and mjpg-streamer is started.
    #  @param unit_id: Name of the unit that is created by systemd.
    #  @param path: dbus path at which you can access information about this unit in systemd (not used by this service)
    def __newSystemdUnit(self, unit_id, path):
        index = self.__getCameraIndexFromUnit(unit_id)
        if index is None:
            return
        port_nr = index + 8080
        log.info("New camera: %d port: %d", index, port_nr)

        self.__cameras[index] = {
            "stream": "http://{host}:%d/?action=stream" % (port_nr),
            "snapshot": "http://{host}:%d/?action=snapshot" % (port_nr),
        }

    ## Event from systemd when a unit was removed. This happens for example if hardware is removed and the mjpg-streamer is killed
    #  @param unit_id: Name of the unit that is created by systemd.
    #  @param path: dbus path at which you can access information about this unit in systemd (not used by this service)
    def __removeSystemdUnit(self, unit_id, path):
        index = self.__getCameraIndexFromUnit(unit_id)
        if index is None:
            return
        log.info("Removed camera:%d", index)
        del self.__cameras[index]

    ## Get the camera index from a unit_id, if this is a camera service.
    #  This function assumes a maximum of 10 cameras. Any more and our current systemd service implementation fails.
    #  @param unit_id: string which is the name of the systemd unit.
    #  @return None if this isn't a camera server that was created.
    #  @return int: index of the camera service. Currently 0 to 9.
    def __getCameraIndexFromUnit(self, unit_id):
        result = re.match("mjpg-streamer@([0-9]).service", unit_id)
        if result is None:
            return None
        return int(result.group(1))
