import collections
from griffin import dbusif

### EventType module.
# This module contains all the possible events in the system with their respective message.
# Events are created in ranges to indicate the general source of the event. This allows for filtering on specific related events.
# Note: These event numberings and parameters are part of the external API. So only additions should be made.
# Note: When deprecating event numbers, keep them in this list, as they could still remain in the database. So we do need the message for this.


## A named tuple class to hold an specific event type.
EventType = collections.namedtuple("EventType", ["type_id", "message"])

#0x0000XXXX range, system related events
SYSTEM_STARTUP = EventType(0x00000001, "System started")
CRITICAL_ERROR = EventType(0x00000002, "Critical error {0}")
SYSTEM_RESET = EventType(0x00000003, "Cleared all settings and history")
SYSTEM_MAINTENANCE = EventType(0x00000004, "The maintenance action {0} was performed by {1}.")

## defaults for the maintenance events
SYSTEM_MAINTENANCE_DEFAULT_MECHANIC = "unknown"
## These are the actions that can be used to post a maintenance event.
SYSTEM_MAINTENANCE_ACTIONS = [
    # every month
    "clean_printer",
    "lubricate_axles",
    # every 3 months
    "check_for_play_on_axles",
    "check_tension_of_short_belts",
    "check_for_residue_in_front_fan_of_print_head",
    "check_quality_of_silicone_nozzle_cover",
    "clean_print_cores",
    "lubricate_lead_screw_z_motor",
    # Every year
    "clean_feeders_and_replace_bowden_tubes"
]

#0x0001XXXX range, hotend related events
HOTEND_CARTRIDGE_CHANGE = EventType(0x00010000, "Hotend {0} changed to {1} with serial {2}")
HOTEND_MATERIAL_CHANGE = EventType(0x00010001, "Hotend {0} material changed to {1} by {2}")
HOTEND_CARTRIDGE_REMOVE = EventType(0x00010002, "Hotend {0} removed")


#0x0002XXXX range, print related events
PRINT_STARTED = EventType(0x00020000, "Print {0} started with name {1}")
PRINT_PAUSED = EventType(0x00020001, "Print {0} paused")
PRINT_RESUMED = EventType(0x00020002, "Print {0} resumed")
PRINT_ABORTED = EventType(0x00020003, "Print {0} aborted")
PRINT_FINISHED = EventType(0x00020004, "Print {0} finished")
PRINT_CLEARED = EventType(0x00020005, "Print {0} cleared")


#0x0010XXXX range, authentication related events
AUTHENTICATION_KEYS_ADDED = EventType(0x00100000, "API Authentication added for application: {1} user: {2} with id: {0}")
AUTHENTICATION_KEYS_REMOVED = EventType(0x00100001, "API Authentication removed for application: {1} user: {2} with id: {0}")


# events that the user can post to the event store API
USER_PROVIDED_EVENTS = [SYSTEM_MAINTENANCE.type_id]


## Create the "event_by_id" mapping. This allows us to get the EventType object from an event_id by using "eventType.event_by_id[type_id]"
event_by_id = {}
for event_name, event_type in list(locals().items()):
    if isinstance(event_type, EventType):
        assert event_type.type_id not in event_by_id
        event_by_id[event_type.type_id] = event_type


## Post an event to the event history.
#  This calls dbus and handles proper parameter conversion.
#  @param event_type: A EventType object from the list in this module.
#  @param parameters: None or a list. If a list is passed all entries in the list are converted to string and stored as parameters with this event.
def postEvent(event_type, parameters=None):
    assert isinstance(event_type, EventType)
    assert event_type.type_id in event_by_id

    if parameters:
        dbusif.RemoteObject("event_history").addEvent(event_type.type_id, list(map(str, parameters)))
    else:
        dbusif.RemoteObject("event_history").addEventSimple(event_type.type_id)
