import logging

from griffin import dbusif
from griffin.datatypes.gcodeHeaderData import GCodeHeaderData
from griffin.math.unitConverter import UnitConverter
from griffin.preferences.registry import Registry

log = logging.getLogger(__name__.split(".")[-1])


# @brief Contains the GCode's metadata (header) and provides access functions to retrieve it.
class GCodeMetaDataContainer(Registry):

    def __init__(self, metadata):
        super().__init__(metadata)
        assert isinstance(metadata, Registry)
        self.__cached_material_types = {}
        self.__cached_material_guids = {}

    # @brief Implements the validator for this data container
    # @return Creates and returns the validator for this container
    def getValidator(self):
        # Prevent circular referencing
        from .gcodeMetaDataContainerValidator import GCodeMetaDataContainerValidator
        return GCodeMetaDataContainerValidator(self)

    # @brief Returns the header version.
    # @return A floating point header version.
    def getVersion(self):
        return self.get("HEADER_VERSION")

    # @brief Returns the Flavor.
    # @return A string representing the flavor.
    def getFlavor(self):
        return self.get("FLAVOR")

    # @brief Returns whether the header is valid by checking it's version!
    # @return A boolean indicating header validity.
    def isValid(self):
        return self.getVersion() == GCodeHeaderData.HEADER_VERSION

    # @return the (uncompressed) filesize in bytes.
    def getFileSize(self):
        return self.getAsInt("FILE_SIZE")

    # @return A boolean indicating whether an old style time estimate is available
    def hasOldStyleTimeEstimate(self):
        return self.getAsBoolean("HAS_LAYER_COUNT") and self.getAsBoolean("HAS_OLD_PRINT_TIME")

    # @brief Returns the initial temperature of the hotend specified.
    # @return temperature in degrees Celsius, or None.
    def getInitialTemperature(self, hotend_nr):
        hotend = self.getAsRegistry("EXTRUDER_TRAIN").getAsRegistry(str(hotend_nr), None)
        if hotend is None:
            return None
        return hotend.get("INITIAL_TEMPERATURE")

    # @brief returns the initial temperature of the build-plate.
    # @return temperature in degrees Celsius, or None.
    def getInitialBuildplateTemperature(self):
        return self.get("BUILD_PLATE").get("INITIAL_TEMPERATURE")

    # @brief Returns the time estimate found in the header.
    # @return time in seconds or -1 if no time was found in the header.
    def getTimeEstimate(self):
        return self.getAsRegistry("PRINT").getAsInt("TIME")

    # @return The old style layer count that does give some vague indication of progress in the file.
    def getLayerCount(self):
        return self.getAsInt("LAYER_COUNT", None)

    # @return A list of extruder trains actively used in this print.
    def getExtruderTrains(self):
        extruder_train = self.get("EXTRUDER_TRAIN")
        return [extruder_key for extruder_key in extruder_train.keys() if extruder_train.get(extruder_key).get("MATERIAL").get("VOLUME_USED") is not None]

    # @brief Returns the material estimated to be used by this print of the hotend specified.
    # @return amount of material in mm3 or None.
    def getMaterialToBeUsed(self, hotend_nr):
        hotend = self.getAsRegistry("EXTRUDER_TRAIN").getAsRegistry(str(hotend_nr), None)
        if hotend is None:
            return None
        return hotend.getAsRegistry("MATERIAL").get("VOLUME_USED")

    # @brief get the material type.
    # @param hotend_nr the hotend nr to get the material type for.
    # @return the material type name of the material GUID in the header is any, None if no material is listed.
    def getMaterialType(self, hotend_nr):
        if hotend_nr not in self.__cached_material_types:
            extruder = self.getAsRegistry("EXTRUDER_TRAIN").getAsRegistry(str(hotend_nr), None)
            if extruder is None:
                return None
            guid = extruder.getAsRegistry("MATERIAL").get("GUID") # should be of type uuid.UUID
            if guid is None:
                return None
            material_type = dbusif.RemoteObject("material").getMaterialNameFields(str(guid))[1]
            if material_type == "?":
                return None
            self.__cached_material_types[hotend_nr] = material_type

        return self.__cached_material_types[hotend_nr]

    # @brief get the material GUID.
    # @param hotend_nr the hotend nr to get the material guid for.
    # @return the material guid in the header is any, None if no material is listed.
    def getMaterialGuid(self, hotend_nr):
        if hotend_nr not in self.__cached_material_guids:
            extruder = self.getAsRegistry("EXTRUDER_TRAIN").getAsRegistry(str(hotend_nr))
            if extruder is None:
                return None
            guid = extruder.getAsRegistry("MATERIAL").get("GUID") # should be of type uuid.UUID
            if guid is None:
                return None
            self.__cached_material_guids[hotend_nr] = guid

        return self.__cached_material_guids[hotend_nr]

    # @return compensated material usage and time estimations in a dictionary.
    def getDurationsAndEstimationMetadata(self):
        metadata = {}
        if self.isValid():
            metadata.update({"time": float(self.getTimeEstimate())})

            for extruder_train_nr in self.getExtruderTrains():
                material_used =  self.getMaterialToBeUsed(extruder_train_nr) # Returns a volume in mm^3
                if material_used is not None:
                    material_guid = self.getMaterialGuid(extruder_train_nr)
                    if material_guid:
                        # Convert volume in mm^3 to length in millimeters, for this the filament diameter is needed, obtained from the material service
                        material_properties = dbusif.RemoteObject("material").getMaterialProperties(str(material_guid))
                        material_used = UnitConverter(**material_properties).convertFromVolumeToLength(material_used)
                    else:
                        material_used = GCodeHeaderData.DATA_UNKNOWN
                else:
                    material_used = GCodeHeaderData.DATA_UNKNOWN
                metadata.update({"material_" + extruder_train_nr: material_used})

        elif self.hasOldStyleTimeEstimate():
            metadata.update({"time": float(self.getTimeEstimate())})

        else:
            metadata.update({"error": True})

        return metadata

    # @return time estimation
    def getEstimationMetadata(self):
        metadata = {}
        if self.isValid() or self.hasOldStyleTimeEstimate():
            metadata.update({"time": float(self.getTimeEstimate())})
        else:
            metadata.update({"time": 0.0})

        return metadata
