from collections import namedtuple
import logging
import re
import uuid

from griffin import dbusif
from griffin.datatypes.gcodeHeaderData import GCodeHeaderData
from griffin.preferences.registry import Registry
from griffin.printer.faultHandler import FaultCode, FaultLevel
from griffin.printer.controller import Controller


log = logging.getLogger(__name__.split(".")[-1])

# @brief Handles validating the GCodeMeta Data (header) for the correctness of the settings and compatibility with the printer
class GCodeMetaDataContainerValidator:

    # @brief Define validation tuple type
    # @param fault_code The fault code for the error message to be displayed
    # @param fault_type The fault type, either FaultLevel.ERROR  or FaultLevel.WARNING
    ValidationResult = namedtuple('ValidationResult', 'fault_code, fault_level, message, data')

    def __init__(self, header_container):
        from griffin.datatypes.gcodeMetaDataContainer import GCodeMetaDataContainer
        assert isinstance(header_container, GCodeMetaDataContainer)
        self.__header_container = header_container
        self.__validation_results = []
        self.__controller = Controller.getInstance()


    # @brief Validates the header data and returns an empty list if no errors are found, otherwise the list of validation errors
    # @return Returns a list of ValidationResult tuples in case of errors, otherwise an empty list
    def validate(self):
        self.__validation_results = []

        # If we don't have any (valid) header, discontinue validation
        if self.__validateHeaderCheck():
            self.__validatePrintSizeCheck()
            self.__validateGeneratorNameCheck()
            self.__validateGeneratorDateCheck()
            self.__validateGeneratorVersionCheck()
            self.__validateTimeEstimateCheck()
            self.__validateNrOfExtrudersCheck()
            self.__validateBuildplateTemperatureCheck()
            self.__validateBuildplateTemperature()

            for extruder_nr, extruder in self.__iterateExtruders():
                if self.__validateExtruderNrCheck(extruder_nr) and self.__controller.getPrintHeadController().getHotendCartridgeManager().isPresent(extruder_nr) and self.__header_container.getInitialTemperature(extruder_nr) is not None:
                    self.__validateNozzleDiameterCheck(extruder_nr, extruder)
                    self.__validateGuidCheck(extruder_nr, extruder)
                    self.__validateVolumeUsedCheck(extruder_nr, extruder)

        return self.__validation_results

    # @brief Creates the named ValidationResult tuple given a fault_code, fault_level, message and optional a data container
    # @param fault_code The fault code
    # @param fault_level The fault level
    # @param data The data (dictionary) or None if no addition information is required
    # @return Returns a ValidationResult tuple
    def __createValidationResult(self, fault_code, fault_level, message, data=None):
        log.debug("Creating result: %d %d %s %r", fault_code, fault_level, message, data)
        return GCodeMetaDataContainerValidator.ValidationResult(int(fault_code), int(fault_level), message, data)

    # @return A list of (extruder_nr, Registry) tuples.
    def __iterateExtruders(self):
        extruder_dictionary = self.__header_container.get("EXTRUDER_TRAIN")
        extruder_train = Registry(extruder_dictionary)
        extruder_keys = extruder_dictionary.keys()
        return [(int(extruder), extruder_train.getAsRegistry(extruder)) for extruder in extruder_keys]

    # @brief Checks the header and sets it to defaults if no header is found
    # @return Returns true if there is a valid header
    def __validateHeaderCheck(self):
        msg = None
        if not self.__header_container.isValid():
            msg = "Correct header was not present"
        elif self.__header_container.getFlavor() != GCodeHeaderData.HEADER_FLAVOR:
            msg = "Flavor is missing or incorrect (gcode header '%s'; '%s' expected by machine)" % (self.__header_container.getFlavor(), GCodeHeaderData.HEADER_FLAVOR)
        if msg:
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_NOT_PRESENT, FaultLevel.ERROR, msg))
            return False

        return True

    # @brief Checks the expected print volume size and if it would fit within the build volume (taking into account the movement volume)
    def __validatePrintSizeCheck(self):
        movement_volume = self.__controller.getPropertyValue("movement_volume")

        for side in [self.__header_container.getAsRegistry("PRINT").getAsRegistry("SIZE").get(key) for key in self.__header_container.getAsRegistry("PRINT").get("SIZE").keys()]:
            for axis in side:
                the_axis = axis.lower()
                min_value = movement_volume["min"][the_axis]
                max_value = movement_volume["max"][the_axis]
                if float(side[axis]) < min_value or float(side[axis]) > max_value:
                    msg = "Gcode expected a machine that can reach %.1fmm in the %s axis. Machine limits are (%d - %d) for this axis." % (side[axis], axis, min_value, max_value)
                    self.__validation_results.append(self.__createValidationResult(FaultCode.MACHINE_TOO_SMALL_FOR_GCODE, FaultLevel.WARNING, msg))

    # @brief Checks the generator name.
    def __validateGeneratorNameCheck(self):
        if self.__header_container.getAsRegistry("GENERATOR").getAsString("NAME") is None:
            msg = "Generator name was missing from header"
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_MISSING_ITEM, FaultLevel.ERROR, msg))
            # Future enhancement: check which generator?

    # @brief Checks the generator version.
    def __validateGeneratorVersionCheck(self):
        if self.__header_container.getAsRegistry("GENERATOR").getAsString("VERSION") is None:
            msg = "GENERATOR.VERSION missing from header"
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_MISSING_ITEM, FaultLevel.WARNING, msg))
            # Future enhancement: check which version?

    # @brief Checks the generator build-date.
    def __validateGeneratorDateCheck(self):
        msg = None
        if self.__header_container.getAsRegistry("GENERATOR").get("BUILD_DATE") is not None:
            # matching the date property of the header, for definition of the header see https://docs.google.com/document/d/1ex3VDTSN0PipNgon1rXXXeWti1jPYhrnAjOCanxJKAg
            match = re.match(r"^(\d{4})-(\d\d)-(\d\d)$", self.__header_container.getAsRegistry("GENERATOR").get("BUILD_DATE"))
            if match:
                year, month, day = match.groups()
                log.debug("GENERATOR BUILD_DATE: %4d-%2d-%2d", year, month, day)
                # perhaps date could be checked in the future
            else:
                msg = "GENERATOR.BUILD_DATE in the header was malformed"
        else:
            msg = "GENERATOR.BUILD_DATE was missing from header"

        if msg:
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_MISSING_ITEM, FaultLevel.WARNING, msg))

    # @brief Checks the time estimate.
    def __validateTimeEstimateCheck(self):
        msg = None
        try:
            if self.__header_container.getAsRegistry("PRINT").getAsInt("TIME") < 0:
                msg = "Time estimate is missing from header"
        except Exception:
                msg = "Time estimate was malformed"
        if msg:
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_MISSING_ITEM, FaultLevel.WARNING, msg))

    # @brief Checks that the number of extruder referred to exist.
    def __validateNrOfExtrudersCheck(self):
        nr_of_extruders = len(
            self.__header_container.get("EXTRUDER_TRAIN").keys())  # get as dictionary so there is a .keys() function...

        if nr_of_extruders > self.__controller.getPropertyValue("hotend_count"):
            msg = "Gcode expected %d nozzles, yet the machine has %d nozzles" % (
            nr_of_extruders, self.__controller.getPropertyValue("hotend_count"))
            self.__validation_results.append(self.__createValidationResult(FaultCode.NOZZLE_AMOUNT_MISMATCH, FaultLevel.ERROR, msg))

    # @brief Checks the build-plate temperature.
    def __validateBuildplateTemperatureCheck(self):
        if self.__header_container.getAsRegistry("BUILD_PLATE").get("INITIAL_TEMPERATURE") is None:
            msg = "BUILD_PLATE.INITIAL_TEMPERATURE missing from header"
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_MISSING_ITEM, FaultLevel.ERROR, msg))

    # @brief Checks if a build plate temperature has been set
    def __validateBuildplateTemperature(self):
        msg = None
        bed_temp = self.__header_container.getAsRegistry("BUILD_PLATE").get("INITIAL_TEMPERATURE", None)
        if bed_temp is None:
            msg = "Initial buildplate temperature must be set"
        else:
            max_bed_temp = float(self.__controller.getPropertyValue("max_bed_temperature"))
            if float(bed_temp) > max_bed_temp:
                msg = "Initial buildplate temperature %0.1f is higher then the maximum temperature %0.1f" % (bed_temp, max_bed_temp)
        if msg:
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_MISSING_ITEM, FaultLevel.WARNING, msg))

    # @brief Checks that the extruder number referred to exists.
    # @param hotend_index The hotend index
    # @return Returns true if the hotend_index is a valid one
    def __validateExtruderNrCheck(self, hotend_index):
        max_hotends = self.__controller.getPropertyValue("hotend_count")
        is_valid_index = hotend_index >= 0 and hotend_index < max_hotends
        if not is_valid_index:
            msg = "GCode using a nozzle cartridge (%d) which is not present in machine (1 t/m %d)" % (hotend_index + 1, max_hotends)
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_MISSING_ITEM, FaultLevel.ERROR, msg, {"display_hotend_nr": hotend_index + 1}))
        return is_valid_index

    # @brief Checks the print's nozzle diameter against the machine's nozzle diameter.
    # @param hotend_index The hotend index
    # @param extruder The extruder of which to check the diameter
    def __validateNozzleDiameterCheck(self, hotend_index, extruder):
        hotend_cartridge_manager = self.__controller.getPrintHeadController().getHotendCartridgeManager()
        data = {"display_hotend_nr": hotend_index + 1}
        msg = None
        try:
            nozzle_diam = extruder.getAsRegistry("NOZZLE").get("DIAMETER")
            if nozzle_diam is None:
                msg = "Nozzle diameter for extruder %d not set" % (hotend_index + 1)
            elif hotend_cartridge_manager.getValue(hotend_index, "nozzle_size") != nozzle_diam:
                msg = "Nozzle diameter mismatch. Gcode expected %.2f, machine has %.2f for extruder %d" % \
                      (nozzle_diam, hotend_cartridge_manager.getValue(hotend_index, "nozzle_size"), hotend_index + 1)
        except Exception as e:
            msg = "%s nozzle diameter malformed for extruder %d" % (str(e), hotend_index + 1)
        if msg:
            self.__validation_results.append(self.__createValidationResult(FaultCode.NOZZLE_MISMATCH, FaultLevel.WARNING, msg, data))

    # @brief Checks the volume used.
    # @param hotend_index The hotend index
    # @param extruder The extruder of which to check the volume
    def __validateVolumeUsedCheck(self, hotend_index, extruder):
        data = {"display_hotend_nr": hotend_index + 1}
        msg = None
        try:
            vol = extruder.getAsRegistry("MATERIAL").get("VOLUME_USED")
            if vol is None:
                msg = "VOLUME_USED is missing (for nozzle cartridge %d)" % (hotend_index + 1)
            elif vol < 0.0:
                msg = "VOLUME_USED can not be negative (for nozzle cartridge %d)" % (hotend_index + 1)
        except Exception:
            msg = "VOLUME_USED malformed (for extruder %d)" % (hotend_index + 1)
        if msg:
            self.__validation_results.append(self.__createValidationResult(FaultCode.HEADER_MISSING_ITEM, FaultLevel.WARNING, msg, data))

    # @brief Checks the GUID.
    # @param hotend_index The hotend index
    # @param extruder The extruder of which to check the material guid
    def __validateGuidCheck(self, hotend_index, extruder):
        tmp_material = self.__controller.getPrintHeadController().getHotendSlot(hotend_index).getPropertyValue("material_guid")
        if tmp_material == "":
            msg = "Gcode file uses slot %d where no material is loaded in the machine" % (hotend_index + 1)
            self.__validation_results.append(self.__createValidationResult(FaultCode.MATERIAL_NOT_LOADED, FaultLevel.WARNING, msg, {"display_hotend_nr": hotend_index + 1}))
            return

        guid_loaded = uuid.UUID(tmp_material)

        guid_in_gcode = extruder.getAsRegistry("MATERIAL").get("GUID")  # should be of type uuid.UUID
        if guid_in_gcode is not None and isinstance(guid_in_gcode, uuid.UUID):
            if guid_in_gcode != guid_loaded:  # also avoids unknown guids

                ms = dbusif.RemoteObject("material")
                type_in_gcode = ms.getMaterialNameFields(str(guid_in_gcode))[1]
                type_loaded = ms.getMaterialNameFields(str(guid_loaded))[1]
                if type_in_gcode != type_loaded and type_in_gcode != "?":  # if the material type is unknown or no material is loaded ? is returned.
                    msg = "GCode file was sliced for material %s in extruder %d where machine had %s loaded" % (str(guid_in_gcode), hotend_index + 1, str(guid_loaded))
                    data = {"display_hotend_nr": hotend_index + 1, "material_type_selected": str(type_loaded), "material_type_print": str(type_in_gcode)}
                    self.__validation_results.append(self.__createValidationResult(FaultCode.GUID_MISMATCH, FaultLevel.WARNING, msg, data))
