from uuid import UUID
from datetime import datetime, timedelta

import dbus
import logging

log = logging.getLogger(__name__.split(".")[-1])

## @brief a PrintJob has all the information to print a job
class PrintJob:
    JOB_RESULT_FAILED = "Failed"
    JOB_RESULT_ABORTED = "Aborted"
    JOB_RESULT_FINISHED = "Finished"

    ## @brief construct a new PrintJob instance
    # @param uuid UUID of the job
    # @param name str name of the job
    # @param state the current state of the job
    # @param source str where did the job come from, who/what originated it
    # @param time_elapsed timedelta how long has the job been printing already
    # @param time_total timedelta, how long remaining until the job is expected to be finished
    # @param datetime_started datetime, when was the job started?  Defaults to None
    # @param datetime_finished datetime, when was the job finished?  Defaults to None
    # @param datetime_cleaned datetime, when was the print resulting from this bed removed from the bed? Defaults to None
    # @param time_estimated timedelta how long does the whole job take ?
    # @param result str outcome of the job (was the job finished successfully or not?)
    def __init__(self, uuid, name, state, source, reprint_original_uuid, time_elapsed, time_total, datetime_started=None, datetime_finished=None, datetime_cleaned=None, result=None, time_estimated = None):
        assert isinstance(uuid, UUID)
        assert isinstance(time_elapsed, timedelta)
        assert reprint_original_uuid is None or isinstance(reprint_original_uuid, UUID)
        assert time_total is None or isinstance(time_total, timedelta)
        assert datetime_started is None or isinstance(datetime_started, datetime)
        assert datetime_finished is None or isinstance(datetime_finished, datetime)
        assert datetime_cleaned is None or isinstance(datetime_cleaned, datetime)
        assert time_estimated is None or isinstance(time_estimated, timedelta)

        self.uuid = uuid
        self.name = name
        self.state = state
        self.source = source
        self.reprint_original_uuid = reprint_original_uuid
        self.time_elapsed = time_elapsed
        self.time_total = time_total
        self.datetime_started = datetime_started
        self.datetime_finished = datetime_finished
        self.datetime_cleaned = datetime_cleaned
        self.result = result
        self.time_estimated = time_estimated

    ## @brief convert a PrintJob to a dictionary to be used as a JSON object for API response.
    # It converts timedelta's to seconds of duration and datetimes to seconds since epoch
    def toJson(self):
        return {
            "uuid": str(self.uuid),
            "name": self.name,
            "source": self.source,
            "reprint_original_uuid": str(self.reprint_original_uuid) if self.reprint_original_uuid else None,
            "time_elapsed": self.time_elapsed.total_seconds(),
            "time_total": self.time_total.total_seconds() if self.time_total is not None else 0,
            "datetime_started": self.datetime_started.replace(microsecond=0).isoformat() if self.datetime_started else None,
            "datetime_finished": self.datetime_finished.replace(microsecond=0).isoformat() if self.datetime_finished else None,
            "datetime_cleaned": self.datetime_cleaned.replace(microsecond=0).isoformat() if self.datetime_cleaned else None,
            "result": self.result if self.result else "",
            "time_estimated": self.time_estimated.total_seconds() if self.time_estimated is not None else 0,
        }

    ## @brief convert a PrintJob to a dictionary that can be transmitted over DBus
    # Represents anything time related to integer seconds
    # @param job, PrintJob
    # @returns a DBus.Dictionary of Strings mapping to Variants
    def toDbus(self):
        return dbus.Dictionary(self._toDict(), signature="sv")

    ## @brief Convert a PrintJob to a plain Python dictionary
    # @param job, PrintJob the job to be converted
    # @returns dict
    def _toDict(self):
        return {
            "uuid": str(self.uuid),
            "name": self.name,
            "state": self.state or "",
            "source": self.source,
            "reprint_original_uuid": str(self.reprint_original_uuid) if self.reprint_original_uuid else "",
            "time_elapsed": self.time_elapsed.total_seconds(),
            "time_total": self.time_total.total_seconds() if self.time_total is not None else -1,
            "datetime_started": int(self.datetime_started.timestamp()) if self.datetime_started is not None else 0,
            "datetime_finished": int(self.datetime_finished.timestamp()) if self.datetime_finished is not None else 0,
            "datetime_cleaned": int(self.datetime_cleaned.timestamp()) if self.datetime_cleaned is not None else 0,
            "result": self.result if self.result else "",
            "time_estimated": self.time_estimated.total_seconds() if self.time_estimated is not None else -1,
        }

    ## @brief Construct a PrintJob from a python dict
    # It requires all the fields to be filled or raise an exception if any field is missing
    # @param data dict with the required fields, one for each member of a PrintJob
    # @returns PrintJob
    @staticmethod
    def fromDict(data):
        try:
            converted = dict(data)
            converted["uuid"] = UUID(converted['uuid'])
            converted["name"] = str(converted['name'])

            converted["time_elapsed"] = timedelta(seconds=converted.get("time_elapsed", 0))
            converted["time_total"] = timedelta(seconds=converted.get("time_total", 0)) \
                if converted["time_total"] != -1 else None

            converted["reprint_original_uuid"] = UUID(converted["reprint_original_uuid"]) \
                if "reprint_original_uuid" in converted and converted["reprint_original_uuid"] else None

            converted["datetime_started"] = datetime.fromtimestamp(converted["datetime_started"]) \
                if "datetime_started" in converted and converted["datetime_started"] else None

            converted["datetime_finished"] = datetime.fromtimestamp(converted["datetime_finished"]) \
                if "datetime_finished" in converted and converted["datetime_finished"] else None

            converted["datetime_cleaned"] = datetime.fromtimestamp(converted["datetime_cleaned"]) \
                if "datetime_cleaned" in converted and converted["datetime_cleaned"] else None

            converted["result"] = None if not converted["result"] else str(converted["result"])  # None if there is no value, else just the value

            converted["time_estimated"] = timedelta(seconds=converted.get("time_estimated", 0)) \
                if converted["time_estimated"] != -1 else None

            return PrintJob(**converted)

        except TypeError as type_err:
            log.exception("Can not instantiate PrintJob from dict {d}".format(d=data))
            raise  # re-raise exception and have the caller deal with it, he gave us wrong stuff. Deal with it (⌐■_■)