from sqlalchemy.types import TypeDecorator, CHAR
from sqlalchemy.dialects.postgresql import UUID
import uuid

## @brief Database-platform independent UUID type
#  Uses PostgreSQL's UUID type, otherwise uses CHAR(32), storing as stringified hex values.
#
# See http://docs.sqlalchemy.org/en/latest/core/custom_types.html#backend-agnostic-guid-type
class UuidType(TypeDecorator):
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    ## @brief convert native uuid to database uuid
    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    ## @brief convert database uuid to native uuid
    def process_result_value(self, value, dialect):
        if value is not None:
            return uuid.UUID(value)
        else:
            return value
