from griffin.datatypes import eventType
from griffin.history.alchemy import session_scope
from griffin.history.alchemy import EventRecord
from griffin.history.alchemy import EventParameterRecord
from sqlalchemy.orm import joinedload
from griffin import thread
import queue
import datetime


## @brief Owner of the database that stored all the events. It acts as the only entry point to it.
class EventHistoryBackend():
    def __init__(self):
        self.__event_store_queue = queue.Queue()
        thread.Thread("EventHistoryBackend store", self.__eventStoreThread).start()

    ## Thread that handles storing of events into the database.
    #  Events are posted on a queue, as storing new data in the database can take quite some time and we do not want to block the addEvent call on this.
    def __eventStoreThread(self):
        # Clean up lingering EventParameterRecords. Factory reset in the 3.6 firmware did not properly clean the database and caused these records to go orphan.
        # See EM-1587, this code needs to stay in place to properly handle updates from 3.6
        with session_scope() as session:
            session.query(EventParameterRecord)\
                .filter(~EventParameterRecord.event_id.in_(session.query(EventRecord.id)))\
                .delete(synchronize_session=False)

        ## Start handling the insertion of events.
        while True:
            type_id, parameters = self.__event_store_queue.get()
            with session_scope() as session:
                event = EventRecord(time=datetime.datetime.utcnow(), event_type_id=type_id)
                for index in range(len(parameters)):
                    event.parameters.append(EventParameterRecord(key=index, value=parameters[index]))
                session.add(event)
            self.__event_store_queue.task_done()

    ## @brief Add a event to the database.
    # @param type_id: Integer value that matches an event in the datatypes.eventType list.
    # @param parameters: A list of parameters to be added to this event. Can have a length of zero.
    def addEvent(self, type_id, parameters):
        # Inserting records in the database can take over 100ms in some cases.
        # So we post this to a queue to be processed, so we can return instantly from this function.
        self.__event_store_queue.put((type_id, parameters))

    ## Remove all the events from the database.
    def removeAllEvents(self):
        with session_scope() as session:
            session.query(EventRecord).delete(synchronize_session=False)
            session.query(EventParameterRecord).delete(synchronize_session=False)
        self.addEvent(eventType.SYSTEM_RESET.type_id, [])
        self.__event_store_queue.join()

    ## Get all events within a certain range.
    #  @param offset: The offset from the latest event in the system, 0 based.
    #  @param count: The amount of events returned by this function.
    #  @return a list of tuples, where each tuple contains:
    #           timestamp of the event.
    #           type_id of the event.
    #           list of string parameters to the event.
    def getEvents(self, offset, count):
        with session_scope() as session:
            results = session.query(EventRecord)\
                .options(joinedload(EventRecord.parameters))\
                .order_by(EventRecord.time.desc())\
                .offset(offset)\
                .limit(count)
            return [(event.time.timestamp(), event.event_type_id, [parameter.value for parameter in event.parameters]) for event in results]

    ## Get all events of a type within a certain range.
    #  @param type_id: The type of events to get.
    #  @param offset: The offset from the latest event in the system, 0 based.
    #  @param count: The amount of events returned by this function.
    #  @return a list of tuples, where each tuple contains:
    #           timestamp of the event.
    #           type_id of the event.
    #           list of string parameters to the event.
    def getEventsByType(self, type_id, offset, count):
        with session_scope() as session:
            results = session.query(EventRecord)\
                .options(joinedload(EventRecord.parameters))\
                .filter(EventRecord.event_type_id==type_id)\
                .order_by(EventRecord.time.desc())\
                .offset(offset)\
                .limit(count)
            return [(event.time.timestamp(), event.event_type_id, [parameter.value for parameter in event.parameters]) for event in results]
