import logging
from griffin import dbusif

from griffin.datatypes import eventType
from griffin.history.eventHistoryBackend import EventHistoryBackend

log = logging.getLogger(__name__.split(".")[-1])

## Event object signature. An integer for the ID, and an array of strings as parameters.
EVENT_OBJECT_SIGNATURE = "ias"

## @brief Interface to the history of Events on this machine
class EventHistoryService(dbusif.ServiceObject):
    def __init__(self):
        super().__init__("event_history")

        self.__backend = EventHistoryBackend()

        ## Add callback to this event_history service which implements the call to the resetToFactorySettings
        dbusif.RemoteObject("system").addFactoryResetCallback("event_history", "resetSettings")

    ## @brief The callback function to register which needs to be called when a factory reset is issued
    #  @param reset_type One of the defined values from FactoryReset which would indicate a hard or soft reset
    @dbusif.method("s", "")
    def resetSettings(self, reset_type):
        log.debug("Removing all events, for reset_type {reset_type}".format(reset_type=reset_type))
        self.__backend.removeAllEvents()

    ## @brief Add an event to the event history, without parameters.
    # This is a convenience method next to addEvent, as passing an empty list on dbus from python can be complex.
    # @param type_id, integer: An type_id from the objects in dataTypes.eventType
    @dbusif.method("i", "")
    def addEventSimple(self, type_id):
        assert int(type_id) in eventType.event_by_id, "type_id should be in dataTypes.eventType list"
        self.__backend.addEvent(int(type_id), [])

    ## @brief Add an event to the event history.
    # @param type_id, integer: An type_id from the objects in dataTypes.eventType
    # @param parameters, list of strings: Parameters given to this event.
    @dbusif.method(EVENT_OBJECT_SIGNATURE, "")
    def addEvent(self, type_id, parameters):
        assert int(type_id) in eventType.event_by_id, "type_id should be in dataTypes.eventType list"
        self.__backend.addEvent(int(type_id), parameters)

    ## @brief get all events
    #  @param offset: The offset from the latest event in the system, 0 based.
    #  @param count: The amount of events returned by this function.
    #  @return a list of tuples, where each tuple contains:
    #           timestamp of the event.
    #           type_id of the event.
    #           list of string parameters to the event.
    @dbusif.method("ii", "a(d%s)" % (EVENT_OBJECT_SIGNATURE))
    def getEvents(self, offset, count):
        return self.__backend.getEvents(offset, count)

    ## Get all events of a type within a certain range.
    #  @param type_id: The type of events to get.
    #  @param count: The amount of events returned by this function.
    #  @return a list of tuples, where each tuple contains:
    #           timestamp of the event.
    #           type_id of the event.
    #           list of string parameters to the event.
    @dbusif.method("iii", "a(d%s)" % (EVENT_OBJECT_SIGNATURE))
    def getEventsByType(self, type_id, offset, count):
        return self.__backend.getEventsByType(type_id, offset, count)