import flask
import os
import binascii

from griffin.interface.http.exposedItems.httpExposedObject import HttpExposedObject
from griffin.interface.http.flaskResponses import FlaskResponses
from griffin.led.controllers.mainLightingController import MainLightingController
from griffin import signal
from griffin import dbusif


## Class which defines the HTTP API for authentication
#  This class handles the calls that are related to getting a new authentication id/key combination as well as checking if this new combo is accepted by the end user.
#  It also has a call to allow for a quick check if authentication is working without making a POST/PUT/DELETE request.
class AuthenticationAPI(HttpExposedObject):
    # Construct a new Authentication API. This requires the same parameters as the HttpExposedObject
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__led_service = dbusif.RemoteObject("led")
        self.__generated_id = None
        self.__generated_key = None
        self.__authorized_by_user = False
        self.__user_data = {}


        ## Signal which is emitted when a new API user is trying to register.
        #  3 parameters:
        #  @param string: application: name of the application that wants to register.
        #  @param string: user: name of the end user that wants to register.
        #  @param string: auth_id: internal id that is used to track this authentication.
        self.onNewAuthorizationRequest = signal.Signal()

    ## Register this API on the server object.
    #  This does intentionally not call super().register, as we do not want the recursive functionality of the HttpExposedObject.
    #  It adds the calls that we expose to the API user.
    def register(self):
        self._server.add_url_rule(self.getAbsolutePath() + "/request", view_func=self._requestNewAuth, methods=["POST"], authenticated_methods=[])
        self._server.add_url_rule(self.getAbsolutePath() + "/check/<auth_id>", view_func=self._checkId, methods=["GET"], authenticated_methods=[])
        self._server.add_url_rule(self.getAbsolutePath() + "/verify", view_func=self._verifyAuth, methods=["GET"], authenticated_methods=["GET"])

    ## Call from the DBus service when the user decides not to allow the application access to the printer.
    #  @param auth_id: The authorization id that was given from the onNewAuthorizationRequest callback.
    def denyAuthorizationRequest(self, auth_id):
        self.__led_service.setMainLightingRuntimeFlag(MainLightingController.RUNTIME_FLAG_AUTHENTICATING, False)
        if self.__generated_id != auth_id:
            return
        self.__generated_id = None

    ## Call from the DBus service when the user decides to allow the application access to the printer.
    #  @param auth_id: The authorization id that was given from the onNewAuthorizationRequest callback.
    def allowAuthorizationRequest(self, auth_id):
        self.__led_service.setMainLightingRuntimeFlag(MainLightingController.RUNTIME_FLAG_AUTHENTICATING, False)
        if self.__generated_id != auth_id:
            return
        if self._server.getAuthenticationController().addUser(self.__generated_id, self.__generated_key, self.__user_data):
            self.__authorized_by_user = True

    ## API call to request a new authentication token.
    #  generates a new id/key combination and reports this back as json.
    #  Also emits the onNewAuthorizationRequest signal to let the system know a new authentication request is happening.
    #  The system should respond with a "allowAuthorizationRequest" or "denyAuthorizationRequest" to allow or deny this new API user.
    #  Until this is done, the new id/key combination is not yet valid.
    def _requestNewAuth(self):
        json_data = flask.request.json
        application = flask.request.form.get("application")
        user = flask.request.form.get("user")
        self.__led_service.setMainLightingRuntimeFlag(MainLightingController.RUNTIME_FLAG_AUTHENTICATING, True)
        if json_data is not None:
            application = json_data.get("application")
            user = json_data.get("user")
        if not application or not user:
            return FlaskResponses.createFailed("application or user not supplied.")

        self.__authorized_by_user = False
        self.__generated_id, self.__generated_key = self.__generateAuthenticationIdAndKey()
        self.__user_data = {
            "application": application,
            "user": user
        }

        # Add exclusive info to the user data if the authentication came from a print-cluster
        # (this information is optional and should not be provided by user applications like Cura)
        exclusion_key = flask.request.form.get("exclusion_key")
        host_name = flask.request.form.get("host_name")
        self.__addOptionalInfoToUserData(self.__user_data, exclusion_key, host_name)

        self.onNewAuthorizationRequest.emit(application, user, self.__generated_id)

        return FlaskResponses.createSuccess(json={"id": self.__generated_id, "key": self.__generated_key})

    ## Generate a new authorization id and key combination.
    #  This set is directly already authorized and does not require user authorization.
    #  This function can be used to give local applications an authorization key/id.
    #  @param application: string containing the application name that wants to use this authorization
    #  @param host_name The host name of the application that is adding it, this may be used in the future to redirect websites.
    #  @param exclusion_key makes this authorisation marked as being exclusive, no more than one authentication can exist for one exclusion_key.
    #  @return a tuple containing the authorization id as first value and the authorization key as second value.
    def createNewAuthorizedKeyIdCombination(self, application, exclusion_key="", host_name=""):
        user_data = {
            "application": application,
            "local": True
        }
        self.__addOptionalInfoToUserData(user_data, exclusion_key, host_name)

        while True:
            # generate random id+key until we no longer have a id collision
            new_id, new_key = self.__generateAuthenticationIdAndKey()
            if self._server.getAuthenticationController().addUser(new_id, new_key, user_data):
                return new_id, new_key

    ## @brief internal function used to fill out the user data with the cluster info if it exists.
    def __addOptionalInfoToUserData(self, user_data, exclusion_key, host_name):
        # the "" is for dbus compatibility as None can't be passed by dbus
        if host_name is not None and host_name != "":
            user_data["host_name"] = str(host_name)
        if exclusion_key is not None and exclusion_key != "":
            user_data["exclusive_for"] = str(exclusion_key)

    ## Generate a random authentication id and key.
    #  @return a tuple containing the authorization id as first value and the authorization key as second value.
    def __generateAuthenticationIdAndKey(self):
        id = binascii.hexlify(os.urandom(16)).decode('ascii')
        key = binascii.hexlify(os.urandom(32)).decode('ascii')
        return id, key

    ## API call to check if a newly requested auth_id is already authorized or not.
    #  This should be called after requesting a new auth id/key combo to check if the end user is allowing this application access or not.
    #  @param: auth_id: string containing the id gained from the auth/request API.
    def _checkId(self, auth_id):
        # If we are not checking the last generated id, then report failure.
        if auth_id != self.__generated_id:
            return FlaskResponses.createSuccess("unauthorized")
        if self.__authorized_by_user:
            return FlaskResponses.createSuccess("authorized")
        return FlaskResponses.createSuccess("unknown")

    ## API call to verify authentication.
    #  As authentication is already handled before this function is called, we only need to return our default "ok" response here.
    def _verifyAuth(self):
        return FlaskResponses.createSuccess("ok")

    ## Call from the DBus service when there is an incoming factory reset
    #  @param reset_type One of the defined values from FactoryReset which would indicate a hard or soft reset
    def resetToFactorySettings(self, reset_type):
        self._server.getAuthenticationController().resetToFactorySettings(reset_type)

    ## @brief This function is used to detect if there is a exclusive authentication known on the local printer.
    # This is used during cluster start-up to know if it should automatically add the local printer to itself.
    # @param exclusion_key: str The application key to search for.
    # @return A boolean indicating whether there is a exclusive application already in control of the printer.
    def hasExclusiveAuth(self, exclusion_key):
        return self._server.getAuthenticationController().hasExclusiveAuth(exclusion_key)
