import flask
import os
import struct
import hashlib
import time
import logging
from ..flaskResponses import FlaskResponses
from griffin import thread
from griffin.preferences import registryFile
from griffin.datatypes import eventType

log = logging.getLogger(__name__.split(".")[-1])


# Create a hash function that takes a list of parameters, joins them as a string with ':' separators and returns the hash as a hex string.
# We need this multiple times for different hashing functions to fully support HTTP digest with RFC 5843
def _makeHashFunction(f):
    return lambda *args: f((":".join(args)).encode('utf-8')).hexdigest()


## This class implements HTTP Digest authentication. According to "RFC 2617"
#  https://en.wikipedia.org/wiki/Digest_access_authentication
#  It does differ from the standard RFC 2617 implementation with the following requirements:
#  * "auth" quality of protection is required instead of optional. This prevents basic authentication as fallback in MITM attacks. And does not allow a fallback to RFC 2069.
#  * The "nc" property, called "nonce-count" in RFC 2617, MUST increase every request. Most example implementations do not check this. This prevents replay attacks.
#  * The "nonce" MUST be generated by the server. As this is used to control the authentication session. Unknown "nonce" values will be seen as invalid.
#
# Implementation details:
#  * Passwords are directly stored. While this is normally regarded as bad practice. We generate keys in place of passwords. There is no extra security gained in storing the A1 hash instead of the password, as this still allows for the same attack.
#  * Both MD5 as SHA-512 are supported hash functions. Clients are recommended to use SHA-512 as hash function, however, this isn't widely supported (browsers cannot handle this yet), even tough RFC 5843 is written to support this.
#    SHA-512 greatly enhances security, because MD5 has a few known attacks, and is not considered 100% safe anymore.
class AuthenticationController:
    # Timeout of the nonce sessions. A nonce token is only valid for an hour before a new one needs to be requested.
    __SESSION_TIMEOUT_SECONDS = 60 * 60
    # Filename where the users of the API are stored in.
    __USER_STORAGE_FILENAME = "http_api_users.json"

    # Define which hashing function to use depended on the length of the hash replied by the API user.
    # Most implementations only use md5, but RFC 5843 defines other hash functions as well. Security wise, you want to use better hash functions then MD5.
    __HASH_FUNCTION_LENGTH = {
        32: _makeHashFunction(hashlib.md5),
        40: _makeHashFunction(hashlib.sha1),
        64: _makeHashFunction(hashlib.sha256),
        128: _makeHashFunction(hashlib.sha512),
    }

    ## Create a new authentication controller.
    #  The main entry point for this class are the addUser and wrapViewFunction calls.
    def __init__(self):
        self.__realm = "Jedi-API"
        self.__sessions = {}
        self.__users = registryFile.RegistryFile(self.__USER_STORAGE_FILENAME)
        self.__enabled = True
        # Generate a random token, we use this to generate the nonce values for sessions.
        self.__random_token = os.urandom(8)

        thread.Thread("http session cleanup thread", self.__sessionCleanupThread)

    ## Add a new user for authentication.
    #  @param new_username: string containing the username that is allowed authentication to the API. (called id in the rest of the authentication usage. But username for digest authentication)
    #  @param new_password: string containing the password that is allowed authentication to the API. (called key in the rest of the authentication usage. But password for digest authentication)
    #  @param new_user_data: dictionary containing additional data about the user.
    def addUser(self, new_username, new_password, new_user_data):
        if self.__users.has(new_username):
            return False
        new_user_data.update({
            "password": new_password
        })

        # Remove authentication of other clusters when adding a cluster authentication
        exclusion_key = new_user_data.get("exclusive_for")
        if exclusion_key:
            for username, user_data in self.__users.getAll().items():
                if isinstance(user_data, dict) and user_data.get("exclusive_for") == exclusion_key:
                    self.__users.delete(username)
                    eventType.postEvent(eventType.AUTHENTICATION_KEYS_REMOVED, [username, user_data.get("application", ""), user_data.get("user", "")])

        self.__users.set(new_username, new_user_data)
        eventType.postEvent(eventType.AUTHENTICATION_KEYS_ADDED, [new_username, new_user_data.get("application", ""), new_user_data.get("user", "")])
        self.__users.forceSave()
        return True

    ## Wrap a api view function with a check for authentication and return this new function.
    #  @param view_function: A function that handles flask calls.
    #  @param check_authentication_on_methods: A list of HTTP methods that require authentication.
    #  @return a new function that will first do an authentication check, and if this is successful will call the view_function and return that result.
    def wrapViewFunction(self, view_function, check_authentication_on_methods):
        return lambda *args, **kwargs: self.__checkAuthenticationWrapper(*args, view_function=view_function, check_authentication_on_methods=check_authentication_on_methods, **kwargs)

    ## Enable global authentication, used for the WiFi setup to allow the web interface free API access without authentication.
    def enableAuthentication(self):
        self.__enabled = True

    ## Disable global authentication, used for the WiFi setup to allow the web interface free API access without authentication.
    def disableAuthentication(self):
        self.__enabled = False

    ## Get the information associated with the authenticated user.
    #  During the authorization pairing, user and application information is stored in the authenticated user list.
    #  Here we retrieve this data for the currently authenticated user.
    #  @return None: user data is not found. Authentication is disabled, or this call never had proper authentication to begin with.
    #  @return dictionary: containing the keys "application" and "user", which were filled by the AuthenticationAPI object.
    def getAuthenticatedUserData(self):
        # If authentication is disabled, we do no know if the credentials send are correct.
        # And thus we do not know the user. To prevent impersonation, return None.
        if not self.__enabled:
            return None

        authorization = flask.request.authorization
        # Get the current user from our internal API user data store.
        user_data = self.__users.get(authorization.username, None)
        if user_data is None:
            return None
        # Make a copy of our user data, and remove the password. This so we do not expose the password to the rest of the code and leak it by mistake.
        user_data = user_data.copy()
        del user_data["password"]
        return user_data

    ## Function which is called when a url is viewed on the webserver that possibly requires authentication.
    #  @param *args: arguments passed to the view_function.
    #  @param view_function: Function which is called for page generation when authentication is not needed or successful.
    #  @param check_authentication_on_methods: list of HTTP methods that require authentication.
    #  @param **kwargs: keyword arguments that are passed to the view_function.
    #  @return result of authentication failure, or result of the view_function is no authentication is needed or authentication is successful.
    def __checkAuthenticationWrapper(self, *args, view_function, check_authentication_on_methods, **kwargs):
        if flask.request.method in check_authentication_on_methods and self.__enabled:
            result = self.__checkAuthentication()
            if result is not None:
                return result
        # Everything checks out, return the normal response that the user actually tried to get.
        return view_function(*args, **kwargs)

    ## Check if the current request has a valid authentication header.
    #  @return None: If authentication is valid.
    #  @return flask.Response instance: If authentication is failed, containing the proper status code and error message.
    def __checkAuthentication(self):
        # Check if we have authorization headers, if not, report that we need those.
        authorization = flask.request.authorization
        if authorization is None or authorization.response is None:
            return self.__deny("Authorization required.")
        # Check if our realm matches the requested realm.
        if self.__realm != authorization.realm:
            return self.__deny("Proper realm required.")
        # Check if we are "auth" authorization in the quality of protection, else, we need this.
        if authorization.qop is None or "auth" not in authorization.qop:
            return self.__deny("Auth qop response required.")

        ## Basic checks are done, we now know the request sort of makes sense.

        # Get our session if we have one.
        session = self.__sessions.get(authorization.nonce, None)
        if session is None:  # nonce not found in sessions. Means this is an old or unknown nonce. Report back that we need to use a new one.
            return self.__deny("Authorization required.", stale=True)   # "stale" reports that this once is no longer valid. And the application should try again without informing the user.

        # If the user is not known, report "Authorization required", from this point on we only report "Authorization required" for the simple reason of not allowing brute force username guessing.
        if not self.__users.has(authorization.username):
            return self.__deny("Authorization required.", forbidden=True)

        # Check of this session matches this user. If none was assigned to this session, assign this one.
        if session["username"] is None:
            session["username"] = authorization.username
        if session["username"] != authorization.username:   # Trying to steal someone else nonce is bad.
            return self.__deny("Authorization required.", forbidden=True)

        # nonce-count checks
        if len(authorization.nc) != 8:  # nonce-count (nc) needs to be 8 hex digits
            return self.__deny("Authorization required.", forbidden=True)
        if session["nc"] is None:   # If we don't know the nonce-count yet, accept it and store it in our session.
            session["nc"] = authorization.nc
        elif authorization.nc <= session["nc"]: # If we do know the previous nonce-count, and the new value is lower or the same, reject this request as well.
            return self.__deny("Authorization required.", forbidden=True)

        # Check if the http digest response matches what we would expect. By the length of the response, we know what hashing function to use.
        # Depending on the response length, decide which hash function to use.
        if len(authorization.response) in self.__HASH_FUNCTION_LENGTH:
            response_hash = self.__calculateResponse(self.__HASH_FUNCTION_LENGTH[len(authorization.response)])
            if authorization.response.lower() != response_hash.lower():
                return self.__deny("Authorization required.", forbidden=True)
        else:
            return self.__deny("Authorization required.", forbidden=True)

        return None

    ## Calculate an authentication response. This depends on the hash function used, and different parts of the authentication header.
    #  This function assumes there is a valid authentication object, and the authentication headers are filled (checked by the __checkAuthentication function)
    #  @param: hash_function created with _makeHashFunction
    #  @return: string containing the resulting hash from http digest according to RFC 2617
    def __calculateResponse(self, hash_function):
        authorization = flask.request.authorization
        HA1 = hash_function(authorization.username, authorization.realm, self.__users.get(authorization.username).get("password"))
        HA2 = hash_function(flask.request.method, authorization.uri)
        return hash_function(HA1, authorization.nonce, authorization.nc, authorization.cnonce, "".join(authorization.qop), HA2)

    ## Generated a flask response that this request is denied.
    #  @param message: Message reported back to the API user to indicate the type of failure.
    #  @param stale: bool, indicates that the nonce token was expired, and the new one should be used.
    #  @param forbidden: authentication was tried, but failed. No new nonce is generated, as the API user did something that wasn't allowed.
    #  @return: flask.Response object containing the message and proper status code. When forbidden is false, then this also generates a new nonce session.
    def __deny(self, message, *, stale=False, forbidden=False):
        if forbidden:
            return FlaskResponses.createForbidden(message)
        else:
            response = FlaskResponses.createUnauthorized(message)
            nonce = hashlib.md5(self.__random_token + struct.pack("=f", time.monotonic())).hexdigest()
            self.__sessions[nonce] = {
                "username": None,
                "nc": None,
                "timeout": time.monotonic() + self.__SESSION_TIMEOUT_SECONDS
            }
            response.www_authenticate.set_digest(self.__realm, nonce, stale=stale)
        return response

    ## Thread which cleans up old sessions every minute.
    #  nonce sessions live for a certain amount of time, after this time the API user needs to use a new nonce.
    #  This is optional HTTP digest behaviour, but improved security against replay and sessions stealing attacks.
    def __sessionCleanupThread(self):
        while True:
            time.sleep(60)
            # Clean up all the sessions of which the timeout is expired.
            self.__sessions = {nonce: data for nonce, data in self.__sessions if data["timeout"] < time.monotonic()}

    ## @brief Implementation of the factory reset callback
    #  @param reset_type Specified which type of reset is requested
    def resetToFactorySettings(self, reset_type):
        if reset_type.startswith("DESTROY_EXCLUSION_KEY:"):
            exclusion_key = reset_type.split(":", maxsplit=1)[-1]
            for key, user_data in self.__users.getAll().items():
                if isinstance(user_data, dict) and user_data.get("exclusive_for") == exclusion_key:
                    self.__users.delete(key)
        else:
            self.__users.erase()

    ## @brief This function is used to detect if there is a exclusive authentication known on the local printer.
    # This is used during cluster start-up to know if it should automatically add the local printer to itself.
    # @param exclusion_key: str The application key to search for.
    # @return A boolean indicating whether there is a exclusive application already in control of the printer.
    def hasExclusiveAuth(self, exclusion_key):
        for user_data in self.__users.getAll().values():
            if isinstance(user_data, dict) and user_data.get("exclusive_for") == exclusion_key:
                return True
        return False
