import logging
import os
import pickle
from typing import Optional, Callable

log = logging.getLogger(__name__.split(".")[-1])


# Provides IO and serialization functionality
class FileHelper:

    ## Write object to disk.
    # @param filename File (including absolute path) to write to.
    # @param obj Object to serialize and write to disk
    @staticmethod
    def serializeObjectToDisk(filename: str, obj: object) -> None:
        try:
            with open(filename, "wb") as file:
                pickle.dump(obj, file)
                # Following is taken from: https://docs.python.org/3.5/library/os.html#os.fsync
                file.flush()
                os.fsync(file.fileno())
                log.info("Created file: %s", filename)
        except (OSError, pickle.PickleError) as ex:
            log.exception("Error writing file '%s'\n%s" % (filename, ex))


    ## Write plain text to disk.
    # @param filename File (including absolute path) to write to.
    # @param plain_text Plain text to write into the file.
    @staticmethod
    def writePlainTextToDisk(filename: str, plain_text: str) -> None:
        try:
            with open(filename, "w") as file:
                file.write(plain_text)
                # Following is taken from: https://docs.python.org/3.5/library/os.html#os.fsync
                file.flush()
                os.fsync(file.fileno())
                log.info("Created file: %s", filename)
        except OSError as ex:
            log.exception("Error writing file '%s'\n%s" % (filename, ex))

    ## Fully reads a single plain text file.
    # @param filename Absolute filepath
    # @return contents of the file or None if the file could not be read/found
    @staticmethod
    def readPlainTextFromDisk(filename: str) -> Optional[str]:
        plain_text = None

        try:
            with open(filename, "r") as file:
                plain_text = file.read()
        except OSError as exception:
            log.exception("Error reading probing report file '%s'\n%s" % (filename, exception))

        return plain_text

    ## Deletes files with a prefix from a directory.
    # @param prefix The prefix for files to be deleted.
    # @param directory The directory to delete files from.
    # @return True if all files could be deleted, False otherwise.
    @staticmethod
    def deleteFilesWithPrefix(prefix: str, directory: str) -> bool:
        all_files_deleted = True

        for filename in os.listdir(directory):
            if filename.startswith(prefix):
                file_to_delete = os.path.abspath(os.path.join(directory, filename))
                # Somewhat unexpected: &= has type int.
                all_files_deleted = all_files_deleted and FileHelper.deleteFile(file_to_delete)

        return all_files_deleted


    ## Deletes a single file.
    # OSerrors are caught and logged, but not re-raised.
    # @param filename Full path to the file to delete.
    # @return True if the file could be deleted, False otherwise.
    @staticmethod
    def deleteFile(filename: str) -> bool:
        try:
            log.info("Deleting '%s'" % (filename))
            os.remove(filename)
            return True
        except OSError as exception:
            log.exception("Cannot delete '%s': %s" % (filename, exception))

        return False

    ## Searches a directory for the first directory matching a predicate
    # @param searchRoot The root directory from where to search
    # @param predicate A function that accepts a pathname relative to
    #   searchRoot which returns wether it is the directory we're looking for
    # @return None if not found, else the first directory for which the predicate was true, relative to searchRoot
    @staticmethod
    def searchDirectory(searchRoot: str, predicate: Callable[[str], bool]) -> Optional[str]:
        if predicate(searchRoot):
            return "."

        for root, _, _ in os.walk(searchRoot):
            if predicate(root):
                return root

        return None