from . import abstractController
from ..effects import staticEffect
from ..effects import glowEffect
from griffin import dbusif
from griffin.math import hsvColor
from .. import colorTheme


## The HeadSlot controller, controls one of the LEDs in the print head.
#  This LED controller has the following features:
#  * Set the color to white when printing.
#  * Set the color to ultimaker blue when idle.
#  * Set the color to "HOT" when the print core is hot, but not in a printing state (to warn the user)
#  * When the print core is heating up or cooling down, glow the LEDs.
#  * When interaction is required, glow the LEDs.
class HeadSlotController(abstractController.AbstractController):
    __COLOR_OFF = colorTheme.BLACK
    __COLOR_PRINTING = colorTheme.WHITE
    __COLOR_HOT = hsvColor.HsvColor(15, 100.0, 100.0)

    # Specific frequency. The head LEDs use hardware glow, which is implemented by Marlin.
    # This frequency matches the old hardcoded value that was there.
    __GLOW_FREQUENCY = 0.7

    def __init__(self, slot_index, hardware):
        super().__init__(hardware)

        self.__current_color = colorTheme.BLACK
        self.__current_glow = False

        printer = dbusif.RemoteObject("printer")
        self.__printer_state = printer.getProperty("state")
        self.__job_state = printer.getProperty("job_state")
        printer.connectSignal("propertyChanged", self.__onPrinterPropertyChanged)

        printer_hotend_service = dbusif.RemoteObject("printer", "printer/head/0/slot/%d" % (slot_index))
        self.__hotend_slot_properties = printer_hotend_service.getProperties()
        printer_hotend_service.connectSignal("propertyChanged", self.__onHotendPropertyChanged)

    # Function called by the onPropertyChanged signal of the printer dbus remoteobject.
    # @param property_key, key of the changed property
    # @param value, new value of the property
    def __onPrinterPropertyChanged(self, property_key, value):
        # Check if one of our monitored properties change and
        if property_key == "state":
            self.__printer_state = value
            self.__update()
        if property_key == "job_state":
            self.__job_state = value
            self.__update()

    ## Called when hotend properties are changed.
    #  @param property_key: string containing the property key.
    #  @param value: New value for this property, variant.
    def __onHotendPropertyChanged(self, property_key, value):
        self.__hotend_slot_properties[property_key] = value
        self.__update()

    ## Update the LED state depending on the system and hotend state.
    #  This uses the private printer state/job state and hotend properties to calculate what effect and color to apply.
    def __update(self):
        if self.__job_state == "printing" or self.__job_state == "pausing":
            color = self.__COLOR_PRINTING
        elif self.__hotend_slot_properties.get("is_enabled"):
            color = self.__COLOR_HOT
        elif self.__hotend_slot_properties.get("is_present"):
            color = colorTheme.ULTIMAKER
        elif self.__hotend_slot_properties.get("interaction_required"):
            color = colorTheme.ULTIMAKER
        else:
            color = self.__COLOR_OFF

        if self.__hotend_slot_properties.get("interaction_required"):
            enable_glow_effect = True
        elif self.__hotend_slot_properties.get("is_enabled") and not self.__atTargetTemperature():
            enable_glow_effect = True
        else:
            enable_glow_effect = False

        programming_state = self.__hotend_slot_properties.get("programming_state")
        if programming_state == "removed":
            color = self.__COLOR_OFF
        elif programming_state == "no_profile_selected":
            color = colorTheme.PURPLE
            enable_glow_effect = True
        elif programming_state == "idle":
            color = colorTheme.ORANGE
        elif programming_state == "writing":
            color = colorTheme.CYAN
        elif programming_state == "error":
            color = colorTheme.RED
        elif programming_state == "programmed":
            color = colorTheme.GREEN

        # Only update if something changed, else we are queuing changes on every temperature update.
        if self.__current_color != color or self.__current_glow != enable_glow_effect:
            if enable_glow_effect:
                self._queueEffect(glowEffect.GlowEffect(colorTheme.BLACK, color, self.__GLOW_FREQUENCY))
            else:
                self._queueEffect(staticEffect.StaticEffect(color))
            self.__current_color = color
            self.__current_glow = enable_glow_effect

    ## @return: bool, true if we are at the target temperature, false if not.
    def __atTargetTemperature(self):
        return abs(self.__hotend_slot_properties.get("current_temperature") - self.__hotend_slot_properties.get("target_temperature")) < 10
