from griffin import thread
from griffin.math import hsvColor
from ..hardware import abstractHardware
import logging
import time

log = logging.getLogger(__name__.split(".")[-1])


## Base class for lighting effects.
#  A lighting effect is a output of light in a sequence.
#  For example, blinking is an effect. Fading from one color to the next is also an effect.
#  Subclasses of this class need to implement the _run function.
#
#  Effect objects can be started, aborted when they need to stop, and be joined to wait for them to finish.
#  This is used by the abstractController to properly handle switching between effects.
class AbstractEffect:
    # Create a new effect object.
    def __init__(self):
        self._abort_event = thread.Event()
        self._abort_event.clear()
        self.__finished_event = thread.Event()
        self.__finished_event.clear()
        self.__thread = None
        self.__hardware = None

    ## Start this effect.
    #  This is called from the abstractController when it is this effects turn to run.
    def start(self):
        assert self.__thread is None
        self.__thread = thread.Thread("LedEffect(%s)" % (self), self.__run)
        self.__thread.start()

    ## Set the hardware that needs to be controlled by this effect.
    #  This is called from the abstractController before the start function is called.
    #  @param hardware: A abstract.Abstract subclass, which will be used by this instance _run function to generate update the output.
    def setHardware(self, hardware):
        assert isinstance(hardware, abstractHardware.AbstractHardware)
        self.__hardware = hardware

    ## Private run function, called when start is called, but on a separate thread.
    #  Handles calling the _run function, and guarantees that the finish event is set when the _run function ends.
    def __run(self):
        try:
            self._run()
        except:
            log.exception("Exception in effect run function of effect: %s", self)
        self.__finished_event.set()

    ## Set the output to a color
    #  @param color: a hsvColor.HsvColor object that will be supplied to the linked hardware object.
    def _setOutput(self, color):
        assert isinstance(color, hsvColor.HsvColor)
        self.__hardware.setColor(color)
        time.sleep(0.05) # Tiny delay to prevent looping without sleeping.

    ## Get the latest output color that was send to the hardware
    #  @param color: hsvColor.HsvColor object.
    def _getOutput(self):
        return self.__hardware.getColor()

    ## Try running a hardware effect on the hardware.
    #  Hardware effects are effects that are 100% controlled by hardware, and thus require no CPU load.
    #  This function will wait till we are aborted and then will stop the hardware effect. Making using this function very easy.
    #  Its intended use is to be called from the _run function. And directly return if this function returns true.
    #  @param effect_name: string containing the name of this effect.
    #  @param parameters: a dictionary of parameters to this hardware effect.
    #  @return bool: True if the current hardware can and has handled this effect. False if we need to handle this in software.
    def _runHardwareEffect(self, effect_name, parameters):
        if self.__hardware.startHardwareEffect(effect_name, parameters):
            self._abort_event.wait()
            self.__hardware.endHardwareEffect(effect_name)
            return True
        return False

    ## Abort the current effect.
    #  This is called from the abstractController when a new effect wants to control the lighting
    def abort(self):
        self._abort_event.set()

    ## Check if this effect is finished with what it was doing.
    #  @return bool: True if this effect is finished running and can be discarded
    def isFinished(self):
        return self.__finished_event.is_set()

    ## Wait this this effect is done running and can be discarded.
    #  Be sure to call abort before calling join, else you can be waiting forever, as some effects will run forever if not aborted.
    def join(self):
        self.__finished_event.wait()

    ## This function needs to be implemented in subclasses.
    def _run(self):
        raise NotImplementedError
