import logging
import os

log = logging.getLogger(__name__.split(".")[-1])


## This class implements the linux userspace led class's brightness parameter.
#  We accomplish this by creating and writing to a tmpfs based file in /run.
#  We use /run as this is the most likely to be a tmpfs mount point.
class _LinuxLEDClassFakeBrightness:
    MAX_BRIGHTNESS = int(255)

    def __init__(self, device, color, led_function):
        self.__base_path = "/run/%s-%s-%s_brightness" % (device, color, led_function)

    def write(self, data):
        try:
            with open(self.__base_path, "w") as brightness:
                brightness.write("%s\n" % data.decode("ascii"))
        except Exception as e:
            log.warning("Unable to set fake led %s (%s)" % (self.__base_path, e))


## This class implements the linux userspace led class via sysfs.
#  By convention LED device naming is of the form: "devicename:color:function"
#  See https://www.kernel.org/doc/Documentation/leds/leds-class.txt
class LinuxLEDClass:
    def __init__(self, device, color, led_function):

        self.__base_path = "/sys/class/leds/"

        if device is not None:
            self.__base_path += "%s" % device

        if color is not None:
            self.__base_path += ":%s" % color

        if led_function is not None:
            self.__base_path += ":%s" % led_function

        self._max_brightness = int(0)
        try:
            with open(os.path.join(self.__base_path, "max_brightness"), "r") as max_brightness:
                self._max_brightness = int(max_brightness.read())
        except Exception as e:
            log.warning("Unable to determine maximum brightness for led %s/max_brightness (%s)" % (self.__base_path, e))

        try:
            self.__brightness = open(os.path.join(self.__base_path, "brightness"), "wb", 0)
        except Exception as e:
            log.warning("Unable to open led %s/brightness (%s)" % (self.__base_path, e))
            log.warning("Falling back to fake hardware")
            self.__brightness = _LinuxLEDClassFakeBrightness(device, color, led_function)
            self._max_brightness = self.__brightness.MAX_BRIGHTNESS

    def __del__(self):
        try:
            self.__brightness.close()
        except Exception as e:
            log.warning("Unable to close led %s (%s)" % (self.__base_path, e))

    ## Disable hardware blinking mode.
    #  Sets the hardware back to normal 'brightness controlled' operation
    def disableBlink(self):
        try:
            with open(os.path.join(self.__base_path, "trigger"), "w") as trigger:
                trigger.write("none")
        except Exception as e:
            log.warn("Unable to open led %s/trigger (%s)" % (self.__base_path, e))

    ## Set the leds to a fixed color.
    # @param The value of the LED output
    def setBrightness(self, value):
        brightness = str(int(max(1.0, value * float(self._max_brightness)))).encode("ascii")
        self.__brightness.write(brightness)
