import math


## Class to represent a color in the HSV spectrum.
#  We use HSV colors instead of RGB as they are more natural to work with.
#  The Hue is normalized between 0 and 360 degrees.
#  The Saturation is clamped from 0 to 100%
#  The Value is clamped from 0 to 100% (value is sometimes called "brightness")
#
#  HsvColor objects are mutable.
class HsvColor:
    def __init__(self, hue=0.0, saturation=0.0, value=0.0):
        self.__hue = float(hue % 360.0)
        self.__saturation = max(0.0, min(float(saturation), 100.0))
        self.__value = max(0.0, min(float(value), 100.0))

    @property
    def hue(self):
        return self.__hue

    @hue.setter
    def hue(self, new_value):
        new_value %= 360.0
        self.__hue = float(new_value)

    @property
    def saturation(self):
        return self.__saturation

    @saturation.setter
    def saturation(self, new_value):
        self.__saturation = max(0.0, min(float(new_value), 100.0))

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, new_value):
        self.__value = max(0.0, min(float(new_value), 100.0))

    def __repr__(self):
        return "HSV(%f, %f, %f)" % (self.__hue, self.__saturation, self.__value)

    # @return a new HsvColor object that has the same values as this object.
    def copy(self):
        return HsvColor(self.__hue, self.__saturation, self.__value)

    ## Interpolate a HSV color, this allows to get an "in between" color of two other colors.
    #  This interpolates linear in a cylinder. Where the hue is the angle from the center, saturation is the radius in the cylinder and the value is the height.
    #  This means if you interpolate from a red color (hue:0) to a Cyan color (hue:180) you do not see other colors but go trough a saturation of zero.
    #  @param other: HsvColor class to interpolate to.
    #  @param f: interpolation value from 0.0 to 1.0:
    #            0.0 means totally this color.
    #            1.0 means totally the other color.
    def interpolate(self, other, f):
        assert isinstance(other, HsvColor)

        # Convert the hue and saturation to a position on a circle. So we can interpolate as a line on this circle.
        # This means if we go from 0 to 180 hue, we do not cross the other colors, but we go trough grey.
        x0 = math.cos(math.radians(self.hue)) * self.__saturation / 100.0
        y0 = math.sin(math.radians(self.hue)) * self.__saturation / 100.0

        x1 = math.cos(math.radians(other.hue)) * other.__saturation / 100.0
        y1 = math.sin(math.radians(other.hue)) * other.__saturation / 100.0

        x = x0 + (x1 - x0) * f
        y = y0 + (y1 - y0) * f

        h = math.degrees(math.atan2(y, x))
        s = math.sqrt(x * x + y * y) * 100.0
        v = self.value + (other.value - self.value) * f
        return HsvColor(h, s, v)

    ## Check if two HSV colors are equal.
    #  There are two degenerate cases:
    #  * if the saturation is of both is zero, then the hue does not matter as there is no color.
    #  * if the value is zero, then the color is black, and nothing really matters.
    #  @param other: A HsvColor instance
    #  @return bool: True if the colors match.
    def __eq__(self, other):
        assert isinstance(other, HsvColor)
        if (self.__saturation == 0.0 and other.saturation == 0.0) or self.__value == 0.0 or other.value == 0.0:
            return self.__value == other.__value
        return self.__hue == other.hue and self.__saturation == other.saturation and self.__value == other.value
