import logging
import time
import os
import subprocess
from serial import Serial, STOPBITS_ONE, STOPBITS_TWO, PARITY_NONE, PARITY_EVEN, serial_for_url

from .networkPacket import NetworkPacket

# Enable to test the data communication stability with an introduced error rate.
#from .test.testSerialWrapper import TestSerialWrapper as Serial

log = logging.getLogger(__name__.split(".")[-1])


## Lowest level of communication to the Marlin board.
#  Handles the opening and setting up of serial ports and the error checking related to that.
#  This basically handles the physical layer of sending and receiving data
class MarlinLowLevelCommunication:
    __BAUDRATE = 250000     # Baudrate which is used to communicate with Marlin.

    # Create a new MarlinLowLevelCommunication object.
    # @port_name: The serial port name to communicate with, as a unicode string. Usually "/dev/ttyS1". For desktop simulation "socket:[port]" can be used to connect to a socket on your local machine.
    def __init__(self, port_name):
        self.__port_name = port_name
        self.__serial = None

    # Try to open the serial port. This function blocks till the opening of the port was successful.
    def open(self):
        log.info("Opening serial port: %s", self.__port_name)
        while self.__serial is None:
            try:
                if self.__port_name.startswith("socket:"):
                    port_number = int(self.__port_name.split(":", 1)[-1])
                    # If we have the simulator executable, run it and wait a second for it to start.
                    if os.path.isfile("/usr/bin/jedi-marlin-simulator"):
                        subprocess.Popen(["/usr/bin/jedi-marlin-simulator", "-p", str(port_number)])
                        time.sleep(1)
                    self.__serial = serial_for_url("socket://127.0.0.1:%d" % (port_number), timeout=0.01)
                else:
                    self.__serial = Serial(self.__port_name, self.__BAUDRATE, timeout=0.01)
                    # For some reason some linux versions have problems with serial ports not configuring correctly.
                    # Setting and unsetting the parity bit works around this issue.
                    self.__serial.parity = PARITY_EVEN
                    self.__serial.parity = PARITY_NONE

            except TypeError as err:
                log.exception("Failed to open serial port: %s. Reason: %s", self.__port_name, err)
                log.error("Most likely there you see a division here somewhere in pyserial. Which means you need to update your pyserial package.")
                log.error("This is already done for the latest full image update. So you could do an [update firmware] to get this package.")
                time.sleep(30.0)
            except Exception as err:
                log.exception("Failed to open serial port: %s. Reason: %s", self.__port_name, err)
                time.sleep(30.0)

        # After opening, wait a short while and flush the input buffers. We could still have old data from the previous "run" of Marlin in buffers.
        # So we want to discard that.
        # With 250.000 baud we receive 31.250 bytes / second. A delay 0f 0.1s == 3125 bytes. This should be long enough.
        time.sleep(0.1)
        self.__serial.flushInput()
        self.__serial.flush()

    # Force a refresh of the serial configuration, by changing the configuration twice.
    # This forces the kernel to update the serial driver. As changes are not always applied properly.
    def refreshPortConfiguration(self):
        if self.__serial is not None:
            self.__serial.bytesize = 7
            self.__serial.baudrate = 9600
            self.__serial.stopbits = STOPBITS_TWO
            self.__serial.parity = PARITY_EVEN
            time.sleep(0.1)
            self.__serial.bytesize = 8
            self.__serial.baudrate = self.__BAUDRATE
            self.__serial.stopbits = STOPBITS_ONE
            self.__serial.parity = PARITY_NONE
            time.sleep(0.1)
            self.__serial.flushInput()
            self.__serial.flush()

    # Close the currently open serial port.
    def close(self):
        log.warning("Closing serial port")
        if self.__serial is not None:
            self.__serial.close()
        self.__serial = None

    # @return: bool, indicate if the serial port is opened or not.
    def isOpen(self):
        return self.__serial is not None

    # Write raw data to the serial port.
    # @param: data, a unicode string to send through the serial port. Note that this is converted to 7 bit ascii before sending.
    def write(self, data):
        assert isinstance(data, NetworkPacket)
        try:
            log.debug("Sending: %s, raw: %r", data, data)
            self.__serial.write(data.getBytes())
        except:
            log.exception("Failure during serial write.")


    # Read a full line from the serial port, this function does not block.
    # @return: A full line read back from Marlin, or an empty string in case no data is available.
    def readline(self):
        result = self.__serial.readline().decode("ascii", "ignore")
        if result != "":
            log.debug("Received_From_Marlin: %s", result)
        return result
