from .heatableObject import HeatableObject
from griffin.printer.properties.basicProperty import BasicProperty
from griffin.printer.properties.systemProvidedProperty import SystemProvidedProperty
from griffin.printer.properties.nonVolatileProperty import NonVolatileProperty
from griffin.printer.properties.constrainedProperty import ConstrainedProperty
import logging

log = logging.getLogger(__name__.split(".")[-1])


## A class representing the print bed.
#  This class is a subclass of the HeatableObject, which means is has all the temperature properties.
#  It also contains any other information that needs to be know about the buildplate.
class HeatablePrintbed(HeatableObject):
    # Range in which we decide that the target temperature is reached.
    # Our top boundary is larger than our bottom boundary as we cool down quite slow and do not want to wait too long for cooling down.
    __TEMPERATURE_TARGET_REACHED_LOW = 3.0
    __TEMPERATURE_TARGET_REACHED_HIGH = 5.0

    # Default type glass should always exist as a buildplate type
    __DEFAULT_BUILDPLATE_TYPE = "glass"

    ## Create a new heatable bed object.
    # @param controller: Reference to the printer service controller.
    # @param bed_registry: Reference to the registry where the bed should store its preferences.
    def __init__(self, controller, bed_registry):
        super().__init__("bed", controller.getPropertyValue("default_bed_pid"))

        self.__controller = controller

        # Bind ourselves to intercept M140 and send a M140 to the driver when we need to update the target temperature.
        controller.getGCodeProcessor().addGCodeMapping("M140", self.__handleTemperatureSetCommand)

        # Bind ourselves to temperature updates from the driver.
        controller.getDriver().onBedTemperatureUpdate.connect(self.__onTemperatureUpdateFromDriver)

        # Set the maximum temperature of the bed.
        self.getPropertyAsObject("max_target_temperature").setBySystem(controller.getPropertyValue("max_bed_temperature"))

        self.__skew_x = NonVolatileProperty(bed_registry, "SKEW_X", 0)
        self.__skew_y = NonVolatileProperty(bed_registry, "SKEW_Y", 0)
        self._addProperty(self.__skew_x)
        self._addProperty(self.__skew_y)

        # Build the property for the buildplate type selection.
        # We also fetch the build plate options from the main controller, as that is the only part that can be filled from the json.
        self.__buildplate_gain_compensators = controller.getPropertyValue("temperature_compensation_build_plate")
        self.__buildplate_type = NonVolatileProperty(bed_registry, "buildplate_type", self.__DEFAULT_BUILDPLATE_TYPE)
        self.__buildplate_type.addConstraint(lambda new_type: new_type in self.__buildplate_gain_compensators)
        self.__buildplate_type.onChange.connect(self.__updateTemperatureMeasurementGain)
        self.__updateTemperatureMeasurementGain()

        self._addProperty(self.__buildplate_type)
        # Add a property to list which build plates are available for the user.
        self._addProperty(SystemProvidedProperty("buildplate_types", list(self.__buildplate_gain_compensators.keys())))

        # Add the properties required for the power budget management for the Ultimaker 3
        self._addProperty(ConstrainedProperty("nominal_resistance", controller.getPropertyValue("default_power_budget_management")["nominal_bed_resistance"], lambda value: 0 < value <= 1000)) # should not be 0 and more than 1000 ohm seems unlikely/illogical
        self._addProperty(ConstrainedProperty("resistance_per_degree", controller.getPropertyValue("default_power_budget_management")["bed_resistance_per_degree"], lambda value: -1 <= value <= 1)) # since this value tends to be less than 0.005 ohm/degC this seems like it should be fine.
        self._addProperty(ConstrainedProperty("voltage", controller.getPropertyValue("default_power_budget_management")["bed_voltage"], lambda value: 1 <= value <= 1000))

        # Set the temperature range for waiting for temperature (members of the base class, used by the wait-for-temperature step)
        self._allowed_temperature_difference_low = HeatablePrintbed.__TEMPERATURE_TARGET_REACHED_LOW
        self._allowed_temperature_difference_high = HeatablePrintbed.__TEMPERATURE_TARGET_REACHED_HIGH

    ## Callback from the gcode processor when an M140 is encountered. This sets the pre_tune_target temperature.
    #  The heatable object class will set the propert target temperature property.
    def __handleTemperatureSetCommand(self, gcode_line):
        s_value = gcode_line.getValue("S", default=None)
        if s_value is not None:
            self.setPropertyValue("pre_tune_target_temperature", s_value)

    ## Callback from the driver when the current temperate of the bed is updated.
    #  Update our internal property with this new temperature.
    def __onTemperatureUpdateFromDriver(self, temperature):
        self.getPropertyAsObject("current_temperature").setBySystem(temperature)

    ## This method derives the temperature from the arguments passed to the function and the material profile..
    # @param kwargs A dictionary, the arguments passed to the setup function of a procedure.
    # @return None when no temperature is derived, else a float value of the temperature.
    def deriveTemperature(self, target=None, **kwargs):
        result_target = None
        if target is not None:
            if target == "initial_temperature":
                printProcedure = self.__controller.getProcedure("PRINT")
                if printProcedure.isActive():
                    metadata = printProcedure.getGcodeMetaData() # getGcodeMetaData() raises exception when None
                    result_target = metadata.getInitialBuildplateTemperature()
            elif target == "material_bed_temperature":
                result_target = self.__controller.getMaterialManager().getRequiredBedTemperature()
            else:
                result_target = float(target)

        return result_target

    ## Update the temperature offset and gain factors depending on the selected build plate.
    #  @param args: The property system passes two arguments to callbacks, we ignore them here.
    def __updateTemperatureMeasurementGain(self, *args):
        buildplate_type = self.__buildplate_type.get()
        gain_compensation = self.__buildplate_gain_compensators[buildplate_type]
        self.setPropertyValue("temperature_measurement_gain_factor", gain_compensation.get("factor", 1.0))
        self.setPropertyValue("temperature_measurement_gain_offset", gain_compensation.get("offset", 0.0))
        return True
