import logging

from griffin.printer.buildPlateLocations import BuildPlateLocations
from griffin.printer.controller import Controller
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.steps.gotoPositionStep import GotoPositionStep
from griffin.printer.procedures.steps.homeBedStep import HomeBedStep
from griffin.printer.procedures.steps.homeHeadStep import HomeHeadStep
from griffin.printer.procedures.steps.powerDownSteppersStep import PowerDownSteppersStep
from griffin.printer.procedures.steps.setPropertyStep import SetPropertyStep
from griffin.printer.procedures.diagnostics.compareHomingDistances import CompareHomingDistances

log = logging.getLogger(__name__.split(".")[-1])


## @brief This procedure moves the head around very fast and checks whether the motor or even belts have skipped any positions
class TestMotionSkipsProcedure(Procedure):

    ## @brief accelerations (in mm/s^2) to test with
    __ACCELERATIONS = [3000, 6000, 9000]

    ## @brief Distance (in mm) to move in a test
    __DISTANCE = 40

    ## @brief Number of times to move left *and* right
    __REPEATS = 20

    ## @brief The maximum deviations to still be qualified as OK
    __MAX_DEVIATION = 0.130

    ## @brief a list of axes for easier iteration and less code
    __AXES = ["X", "Y"]

    ## @brief Initializes the procedure
    #  @param key The procedure key
    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_MAINTENANCE)
        self.__meta_data = {}
        self.__steps = []
        self.__controller = Controller.getInstance()
        ## @brief The speed (in mm/s) at which to move during the test (maximum possible speed)
        self.__test_velocity = self.__controller.getPropertyValue("maximum_speed_limit")
        if self.__test_velocity is None:
            # Assume default value = 300mm/s
            self.__test_velocity = { "x": 300, "y": 300 }

    def setup(self, args):
        self.__steps = []  # Otherwise, we just keep adding more and more steps with each setup() call
        self.__meta_data = {"accelerations":",".join([str(a) for a in self.__ACCELERATIONS])}

        center = BuildPlateLocations.getInstance().getPosition("CENTER_HEAD")
        movement_volume = Controller.getInstance().getPropertyValue("movement_volume")
        build_volume = Controller.getInstance().getPropertyValue("build_volume")
        comparing_x = center.x - movement_volume["min"]["x"]
        comparing_y = center.y + movement_volume["max"]["y"] - build_volume["y"]

        self.__steps += [HomeBedStep("HOME_BED_BEFORE_TEST")]
        self.__steps += [HomeHeadStep("HOME_HEAD_BEFORE_TEST")]

        for acceleration in self.__ACCELERATIONS:
            self.__steps += [GotoPositionStep("GOTO_CENTER_{acc}MM/S^2".format(acc=acceleration), x=center.x, y=center.y)]
            self.__steps += [GotoPositionStep("LEFT_{acc}MM".format(acc=acceleration), x=-self.__DISTANCE / 2, relative_positions=True)]
            self.__steps += [SetPropertyStep("SET_ACCELERATION_{acc}MM/S^2".format(acc=acceleration), "printer", "acceleration_xyz", acceleration)]
            for i in range(self.__REPEATS):
                self.__steps += [GotoPositionStep("RIGHT_{acc}MM/S^2_{i}".format(acc=acceleration, i=i), x=self.__DISTANCE, wait_for_queue_empty=False, relative_positions=True, speed=self.__test_velocity["x"])]
                self.__steps += [GotoPositionStep("LEFT_{acc}MM/S^2_{i}".format(acc=acceleration, i=i), x=-self.__DISTANCE, wait_for_queue_empty=False, relative_positions=True, speed=self.__test_velocity["x"])]
            for i in range(self.__REPEATS):
                self.__steps += [GotoPositionStep("FRONT_{acc}MM/S^2_{i}".format(acc=acceleration, i=i), y=self.__DISTANCE, wait_for_queue_empty=False, relative_positions=True, speed=self.__test_velocity["y"])]
                self.__steps += [GotoPositionStep("BACK_{acc}MM/S^2_{i}".format(acc=acceleration, i=i), y=-self.__DISTANCE, wait_for_queue_empty=False, relative_positions=True, speed=self.__test_velocity["y"])]

            self.__steps += [GotoPositionStep("GOTO_CENTER_{acc}MM/S^2".format(acc=acceleration), x=center.x, y=center.y)]

            compare_x = CompareHomingDistances("COMPARE_X_HOMING_FOR_{acc}_MM/S^2".format(acc=acceleration), expected_value=comparing_x, axis="X")
            compare_y = CompareHomingDistances("COMPARE_Y_HOMING_FOR_{acc}_MM/S^2".format(acc=acceleration), expected_value=comparing_y, axis="Y")

            compare_x.stepDataReady.connect(self.check_diff)
            compare_y.stepDataReady.connect(self.check_diff)

            self.__steps += [compare_x]
            self.__steps += [compare_y]

        self.__steps += [
            HomeHeadStep("HOME_HEAD_AFTER_TEST"),
            PowerDownSteppersStep("POWER_DOWN_MOTORS"),
        ]
        self._addStepSequence(self.__steps)
        # Also power down the steppers in case of aborting
        self.setAbortstepForMultiple(self.__steps[:-2], self.__steps[-2])
        return True

    ## @brief Gather metadata of this step
    # Each Compare-step records the measured difference in its metadata.
    # This difference is checked here and the result is put in the metadata of this procedure.
    def getMetaData(self):
        for step in self.__steps:
            for key, value in step.getMetaData().items():
                self.__meta_data[step.getKey() + "/" + key] = value

        ok_per_acceleration = {}
        for acc in self.__ACCELERATIONS:
            for axis  in self.__AXES:
                diff_key = "COMPARE_{axis}_HOMING_FOR_{acc}_MM/S^2/diff".format(axis=axis, acc=acc)
                result_name = "{axis}_{acc}".format(axis=axis, acc=acc)
                if diff_key in self.__meta_data:
                    ok_per_acceleration[result_name + "_OK"] = 1 if abs(self.__meta_data[diff_key]) < self.__MAX_DEVIATION else 0
                else:
                    ok_per_acceleration[result_name + "_OK"] = -1

        self.__meta_data.update(ok_per_acceleration)
        return self.__meta_data

    ## @brief check the observed difference and abort the test procedure if machine fails
    def check_diff(self, diff):
        diff_too_large = abs(diff) >= self.__MAX_DEVIATION
        if diff_too_large:
            log.warning("Homing difference = {d}mm. Too large, aborting test".format(d=diff))
            self.abort()
        else:
            log.info("Homing difference = {d}mm. This is OK, continuing test".format(d=diff))
