from griffin.printer.controller import Controller
from griffin.printer.procedures.otherProcedureStep import OtherProcedureStep
from griffin.printer.procedures.pre_and_post_print.deprimeProcedure import DeprimeProcedure
from griffin.printer.procedures.pre_and_post_print.wipeNozzleStep import WipeNozzleStep
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.procedureStep import OutcomeBase
from griffin.printer.procedures.steps.abortProcedureAndWaitStep import AbortProcedureAndWaitStep
from griffin.printer.procedures.steps.gotoPositionStep import GotoPositionStep
from griffin.printer.procedures.steps.homeBedStep import HomeBedStep
from griffin.printer.procedures.steps.homeHeadStep import HomeHeadStep
from griffin.printer.procedures.steps.powerDownSteppersStep import PowerDownSteppersStep
from griffin.printer.procedures.steps.resetPropertyToDefaultStep import ResetPropertyToDefaultStep
from griffin.printer.procedures.steps.setPropertyStep import SetPropertyStep
from griffin.printer.procedures.queueRawGCodeWaitStep import QueueRawGCodeWaitStep
from griffin.printer.procedures.waitForQueueEmptyStep import WaitForQueueToBeEmptyStep

import logging

log = logging.getLogger(__name__.split(".")[-1])


## @brief This procedure does all the steps to secure the printer after the print.
class PostPrintProcedure(Procedure):
    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_UNKNOWN)
        self.setAbortDisabled(True)

    ## brief Resets printer to a print-ready state.
    def setup(self, args):
        controller = Controller.getInstance()
        steps = [
            AbortProcedureAndWaitStep("ABORT_PAUSE", "PAUSE_PRINT"),
            AbortProcedureAndWaitStep("ABORT_RESUME", "RESUME_PRINT"),
            # Reset a bunch of things at the end of a print as we don't want these having an effect
            # on the next print or on maintenance procedures.
            ResetPropertyToDefaultStep("RESET_PRINT_SPEED", "printer", "movement_speed_modifier"),
            ResetPropertyToDefaultStep("RESET_MATERIAL_FLOW", "printer", "extrusion_amount_modifier"),
            ResetPropertyToDefaultStep("RESET_ACCELERATION_XYZ", "printer", "acceleration_xyz"),
            ResetPropertyToDefaultStep("RESET_ACCELERATION_E", "printer", "acceleration_e"),
            ResetPropertyToDefaultStep("RESET_JERK_XY", "printer", "jerk_xy"),
            ResetPropertyToDefaultStep("RESET_JERK_Z", "printer", "jerk_z"),
            ResetPropertyToDefaultStep("RESET_JERK_E", "printer", "jerk_e"),
            ResetPropertyToDefaultStep("RESET_MAX_SPEED_X", "printer", "max_speed_x"),
            ResetPropertyToDefaultStep("RESET_MAX_SPEED_Y", "printer", "max_speed_y"),
            ResetPropertyToDefaultStep("RESET_MAX_SPEED_Z", "printer", "max_speed_z"),
            ResetPropertyToDefaultStep("RESET_MAX_SPEED_E", "printer", "max_speed_e"),
            ResetPropertyToDefaultStep("RESET_CURRENT_XY", "printer", "motor_current_xy"),
            ResetPropertyToDefaultStep("RESET_CURRENT_Z", "printer", "motor_current_z"),
            ResetPropertyToDefaultStep("RESET_CURRENT_E", "printer", "motor_current_e"),
            SetPropertyStep("COOLDOWN_BED", "printer.bed", "pre_tune_target_temperature", 0.0),
            ResetPropertyToDefaultStep("RESET_TUNE_OFFSET_TEMPERATURE_BED", "printer.bed", "tune_offset_temperature"),

            # Implement a quick retract first (deprime procedure will know) on the current active hotend only
            GotoPositionStep("QUICK_RETRACT", e=-DeprimeProcedure.QUICK_RETRACT_DIST, speed=controller.getPropertyValue("deprime_retraction_speed"), relative_positions=True),
            # Then lower the bed 3mm (z-hop) to prevent colliding with the printed object
            GotoPositionStep("LOWER_BED", z=3, relative_positions=True),
            # Perform the wipe before Z now (knowing a) no material is flowing and b) the head will not collide with the object
            WipeNozzleStep("WIPE_BEFORE_Z"),
            # To be completely safe, move the head to the home position already
            HomeHeadStep("HOME_HEAD"),
            # Lower the bed before we deprime, depriming takes a while, and not moving the bed down looks strange towards the user at this point.
            HomeBedStep("HOME_BED"),
            # The DEPRIME procedure requires the extrusion_amount_modifier to be 100%
            OtherProcedureStep("DEPRIME", "DEPRIME"),
            # And finally stop the fans
            SetPropertyStep("STOP_FANS", "printer.head.0", "cooling_fan_speed", 0.0)
        ]

        for hotend_nr in range(controller.getPropertyValue("hotend_count")):
            steps.append(QueueRawGCodeWaitStep("RESET_OFFSET_{hotend_nr}".format(hotend_nr=hotend_nr), "M218 T{hotend_nr} X0 Y0 Z0".format(hotend_nr=hotend_nr)))
            steps.append(ResetPropertyToDefaultStep("RESET_TUNE_OFFSET_TEMPERATURE_{hotend_nr}".format(hotend_nr=hotend_nr), "printer.head.0.hotend.{hotend_nr}".format(hotend_nr=hotend_nr), "tune_offset_temperature"))

        steps += [
            WaitForQueueToBeEmptyStep("WAITING"),
            OtherProcedureStep("SWITCH_TO_HOTEND", "SWITCH_ACTIVE_HOTEND", {"target": 1}),
            HomeHeadStep("HOME_HEAD"),
            WaitForQueueToBeEmptyStep("WAITING"),
            PowerDownSteppersStep("POWER_DOWN_STEPPERS"),
            WaitForQueueToBeEmptyStep("WAITING")
        ]

        self._addStepSequence(steps)

        # For each step, set the next step as abort step. This means that if a step is aborted, we will continue
        # with the next step. This will handle the abort of the DEPRIME, which will then not bubble up but will
        # just continue with the next step.
        for n in range(0, len(steps) - 1):
            steps[n].setNextConditionalStep(OutcomeBase.Aborted, steps[n+1])

        return True
