import datetime
import os
import urllib.parse
import uuid
import logging

from griffin import dbusif
from griffin.datatypes import eventType
from griffin.datatypes.printJob import PrintJob
from griffin.printer.controller import Controller
from griffin.printer.procedures.otherProcedureStep import OtherProcedureStep
from griffin.printer.procedures.print.parseHeaderStep import ParseHeaderStep
from griffin.printer.procedures.print.printFileStep import PrintFileStep
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.procedureStep import OutcomeBase
from griffin.printer.procedures.startProcedureStep import StartProcedureStep
from griffin.printer.procedures.wizard.checkCartridgeLoaded import CheckCartridgeLoaded
from griffin.printer.procedures.wizard.waitMessageStep import WaitMessageStep
from griffin.printer.faultHandler import FaultHandler, FaultLevel


log = logging.getLogger(__name__.split(".")[-1])


# The main printing procedure.
# Has the following steps:
# RUN_PRE_PRINT_SETUP - Runs the PRE_PRINT_SETUP procedure and waits for it to finish.
#       Sets up the printer for printing, does homing/leveling/etc if needed.
# WAIT_SCHEME - Waits for the scheme handler to finish. When a file needs to be downloaded from the internet, this
#       step might take a while.
# PRINT - Actual printing step, streams gcode to the printer.
# POST_PRINT - End of print handling, cooling down, homing head, retracting filament.
# WAIT_FOR_CLEANUP - Wait till the printed is emptied.
class PrintProcedure(Procedure):
    # List of origins to be excluded from the reprint
    __EXCLUDED_ORIGINS_FOR_REPRINT=["CALIBRATION_MENU"]

    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_PRINT)
        self.__job_name = ""
        self.__origin = ""
        self.__origin_metadata = {}

        self.job_history = dbusif.RemoteObject("job_history")

        self._parse_header_step = None
        self._print_step = None
        self._post_step = None
        self._post_without_wait_step = None

        self._header = None
        self.__gcode_metadata = None

        self.__current_job = None

        self.__previous_job_args = None

        # Register with self.onNextStep and onFinished to post the current job to the JobHistoryService
        self.onNextStep.connect(self.__handleNextStepStarted)
        self.onFinished.connect(self.__handleProcedureFinished)

    def setup(self, args):
        #re-set header from last print
        self._header = None

        # Flag to indicate the option to reprint after abort or being shown in the menu
        allow_reprint = "origin" in args and args["origin"] not in self.__EXCLUDED_ORIGINS_FOR_REPRINT

        FaultHandler.clear(FaultLevel.WARNING) # clear lingering warning screens as we are starting a print,
        #  warnings are not critical and they block the start of a print.

        # resume a previous printjob
        reprint = args.get("reprint", False)
        if reprint:
            args = self.__previous_job_args

        recovered_job = args.get("recovered_job")
        if recovered_job is not None:
            # we failed a previous print after a powerloss reflect that a print was running!
            self.__current_job = recovered_job
            self.__origin = "Reboot"
            self._addStepSequence([WaitMessageStep("WAIT_FOR_CLEANUP", "PRINTER_CLEANED")])
            if recovered_job.result == PrintJob.JOB_RESULT_FAILED:
                self.abort() # start in aborted state (hack suggested by david), the procedure actually does start the first step...
            return True

        if "url" not in args:
            log.warning("No URL given to print procedure")
            return False

        # Do we need to have a cleanup step or not.
        # If the print is done in a wizard (For example X/Y offset calibration) then the wait for cleanup step needs to be skipped.
        cleanup_step = args.get("cleanup_step", "True") == "True"

        # Post step not allowed to abort. On abort in setup step, post step is
        # run. The abort message in the print step is caught, the file read
        # will be aborted, but no "real" abort of procedures is handled. The
        # post step will run as normal.
        if not reprint:
            if self._parse_header_step is not None:
                self._parse_header_step.getSchemeHandler().cleanup()
                self._parse_header_step.getFileHandler().cleanup()
            self._parse_header_step = ParseHeaderStep("PARSE_HEADER")
        self._print_step = PrintFileStep("PRINTING", self._parse_header_step)
        self._post_step = OtherProcedureStep("RUN_POST_PRINT", "POST_PRINT")

        self._post_without_wait_step = OtherProcedureStep("RUN_POST_PRINT_WITHOUT_WAIT_FOR_CLEANUP", "POST_PRINT")

        log.info("Setting up step chain, cleanup: %s", cleanup_step)
        # Setup the step chain. As this can change when the cleanup step is enabled or not.
        pre_print_steps = []
        for hotend_index in range(Controller.getInstance().getPropertyValue("hotend_count")):
            pre_print_steps.append(CheckCartridgeLoaded("CHECK_CARTRIDGE_LOADED_%d" % hotend_index, hotend_index))

        pre_print_steps.append(self._parse_header_step)
        pre_print_steps.append(OtherProcedureStep("RUN_PRE_PRINT_SETUP", "PRE_PRINT_SETUP"))

        print_steps = [self._print_step]
        post_print_steps = [self._post_step]

        if cleanup_step:
            wait_step = WaitMessageStep("WAIT_FOR_CLEANUP", "PRINTER_CLEANED")
            post_print_steps += [wait_step]
            # Even when we're aborted, Material has (likely) been extruded so show the wait for cleanup screen!
            #
            # The _post_step itself *succeeds* in doing whatever it needs to do to handle an abort, so it returns that it succeeds.
            # That gets overruled by the aborted procedure containing the _post_step, so we need to explicitly set the next step
            # of _post_step to handle the abort.
            self._post_step.setNextConditionalStep(OutcomeBase.Aborted, wait_step)

        self._addStepSequence(pre_print_steps + print_steps + post_print_steps)

        # After parse and setup, no extrusion happened yet. So, we do not have to wait for cleanup but can continue straight to a POSTPRINT step.
        self.setAbortstepForMultiple(pre_print_steps, self._post_without_wait_step)  # At this point, no extrusion happened yet.

        # Once we're printing, there is material on the buildplate and thus it needs to be cleaned, so we go to the normal post_step.
        self.setAbortstepForMultiple(print_steps, self._post_step)

        # When we allow reprinting, we want to start the RETRY_AFTER_PRINT_ABORT procedure when we aborted.
        # We conditionally add this step to our leafs of the step tree
        if allow_reprint:
            start_reprint_step = StartProcedureStep("RETRY_AFTER_PRINT_ABORT", "REPRINT_AFTER_ABORT", args)
            # We depend on the procedure outcome here, instead of the step outcome. As we cannot depend
            # on a single step outcome, as the step gets aborted, but the post-print step is run after that.
            # And the post-print is not aborted after that. However, the procedure itself is aborted so we can check on that.
            post_print_steps[-1].setNextConditionalStep(
                lambda _: self.getProcedureOutcome() == OutcomeBase.Aborted,
                start_reprint_step
            )
            self._post_without_wait_step.setNextConditionalStep(
                lambda _: self.getProcedureOutcome() == OutcomeBase.Aborted,
                start_reprint_step
            )

        try:
            url_data = urllib.parse.urlparse(args["url"], allow_fragments=False)
        except Exception:
            log.exception("Failed to parse URL data")
            return False

        # Get the job name, this could be given through the parameters, if not, it is created from the filename.
        # The jobname does not have to equal the filename, as the file could be a temp file.
        if "jobname" in args:
            job_name = args["jobname"]
        else:
            basename = os.path.splitext(os.path.basename(url_data.path))
            if basename[1].upper() == ".GZ" and basename[0].upper().endswith(".GCODE"):
                job_name = os.path.splitext(basename[0])[0]
            else:
                job_name = basename[0]

        # Store the origin of the job, if given through dbus. We give this back through the metadata to allow other services to know where the job came from.
        self.__origin = args.get("origin", "Unknown")

        # Fill the origin metadata with parameters given as origin_, we strip off the prefix and store it in origin_metadata.
        # This origin metadata is data about the specific origin specified as "origin".
        # Example of this, when the origin is "WEB_API", you can have "user" and "application" as origin metadata.
        # (We use this method, instead of a single "origin_metadata" parameter as a dictionary, as encoding a dictionary in a variant on dbus has many restrictions and problems, on both the python and C++ side)
        self.__origin_metadata = {}
        for key, value in args.items():
            if key.startswith("origin_"):
                self.__origin_metadata[key[7:]] = value

        # The parse header needs to know the url of the print job, so it can properly setup the scheme handler and file handlers.
        if not reprint:
            self._parse_header_step.setUrlData(url_data)


        if "origin" in args and args["origin"] == "WEB_API":
            source = "{api}/{user}/{application}".format(api=args.get("origin","Unknown"),
                                                         user=args.get("origin_user", "Unknown"),
                                                         application=args.get("origin_application", "Unknown"))
        else:
            source = args.get("origin", "")  # Could also be just "USB"

        try:
            controller = Controller.getInstance()
            state = controller.getPropertyValue("job_state")
        except Exception:
            log.exception("Could not get job_state")
            state = "unknown"

        # When this is a re-print, get our original uuid for the print.
        # This allows us to trace back what the original print job was, which is important for tracability of jobs.
        reprint_original_uuid = args.get("reprint_original_uuid", None)

        job_uuid = uuid.uuid4()
        self.__current_job = PrintJob(uuid=job_uuid,
                                      name=job_name,
                                      datetime_started=datetime.datetime.utcnow(), # Remember when the print really started
                                      source=source,
                                      reprint_original_uuid=reprint_original_uuid,
                                      state=state,
                                      time_elapsed=datetime.timedelta(0),
                                      time_total=datetime.timedelta(0),
                                      time_estimated=datetime.timedelta(0))
        eventType.postEvent(eventType.PRINT_STARTED, [self.__current_job.uuid, self.__current_job.name])

        if allow_reprint:
            self.__previous_job_args = args.copy()
            self.__previous_job_args["jobname"] = job_name
            if reprint_original_uuid is None:
                self.__previous_job_args["reprint_original_uuid"] = job_uuid
            else:
                self.__previous_job_args["reprint_original_uuid"] = reprint_original_uuid
        else:
            self.__previous_job_args = None

        return True

    def getMetaData(self):
        if self.__current_job is not None:
            meta_data = {
                "jobname": self.__current_job.name,
                "uuid": str(self.__current_job.uuid),
                "reprint_original_uuid": str(self.__current_job.reprint_original_uuid) if self.__current_job.reprint_original_uuid else "",
                "datetime_started": self.__current_job.datetime_started.timestamp() if self.__current_job.datetime_started else 0,
                "datetime_finished": self.__current_job.datetime_finished.timestamp() if self.__current_job.datetime_finished else 0,
                "datetime_cleaned": self.__current_job.datetime_cleaned.timestamp() if self.__current_job.datetime_cleaned else 0,
                "origin" : self.__origin,
                "active_step": str(self.getActiveStep().getKey() if self.getActiveStep() is not None else ""),
                "aborted": self.isAborted(),
                "result": "" if self.__current_job.result is None else self.__current_job.result,
                "paused": Controller.getInstance().getPaused(),
                "file_path": self.file_path,
            }
        else:
            has_previous_job = self.__previous_job_args is not None
            meta_data = {
                "has_previous_job": has_previous_job,
                "previous_job_name": self.__previous_job_args.get("jobname") if has_previous_job else ""
            }
            if self.__gcode_metadata is not None:
                meta_data.update(self.__gcode_metadata.getDurationsAndEstimationMetadata())
        # Add the origin metadata. We do not add this as a single dictionary to the metadata, as encoding and decoding dictionaries as variants,
        # is very problematic on DBUS. So we add the origin_ prefix to each of this data to indicate that it is data about the origin.
        for key, value in self.__origin_metadata.items():
            meta_data["origin_%s" % key] = value

        # Add the data from the _parse_header, which contains time estimates and progress information.
        # This lives for the whole lifecycle of this procedure, unlike the one from self._print_step
        filehandler = None
        if self._parse_header_step is not None:
            filehandler = self._parse_header_step.getFileHandler()
        if filehandler is not None:
            meta_data.update(filehandler.getMetaData())
        # Add the data from the print step, this adds print time estimates, and progress information.
        if self._print_step is not None:
            meta_data.update(self._print_step.getMetaData())

        if self.__gcode_metadata is not None:
            for index, material_type_name in [ (index, str(self.__gcode_metadata.getMaterialType(index))) for index in range(Controller.getInstance().getPropertyValue("hotend_count"))]:
                if material_type_name is not None and material_type_name is not str(None):
                    meta_data["gcode_material_type_"+ str(index)] = material_type_name
        return meta_data

    def setGcodeMetaData(self, metadata):
        self.__gcode_metadata = metadata

    def getGcodeMetaData(self):
        if self.__gcode_metadata is None:
            raise ValueError("metadata is None") # header should never be requested when None!
        return self.__gcode_metadata

    ## @brief When the print-step is finished, log the time at which that happens.
    def __handleNextStepStarted(self, procedure, next_step):
        if next_step == self._post_without_wait_step:
            # if we come here either something failed during the pre-print or the user aborted at that point.
            if self.isAborted():
                eventType.postEvent(eventType.PRINT_ABORTED, [self.__current_job.uuid])
                self.__current_job.result = PrintJob.JOB_RESULT_ABORTED
            else:
                self.__current_job.result = PrintJob.JOB_RESULT_FAILED
            self.__current_job.datetime_finished = datetime.datetime.utcnow()

        if next_step == self._post_step:
            if self.isAborted():
                eventType.postEvent(eventType.PRINT_ABORTED, [self.__current_job.uuid])
                self.__current_job.result = PrintJob.JOB_RESULT_ABORTED
            else:
                eventType.postEvent(eventType.PRINT_FINISHED, [self.__current_job.uuid])
            self.__current_job.datetime_finished = datetime.datetime.utcnow()

    ## @brief When the complete procedure is done, log the time and add the job to history.
    def __handleProcedureFinished(self, procedure):
        log.debug("__handleProcedureFinished({proc})".format(proc=procedure))
        self.__current_job.datetime_cleaned = datetime.datetime.utcnow()

        self.__addJobToHistory()
        eventType.postEvent(eventType.PRINT_CLEARED, [self.__current_job.uuid])

    ## @brief Store the current job data in the JobHistory Service.
    def __addJobToHistory(self):
        metadata = self.getMetaData()

        if self.job_history.getJobByUuid(str(self.__current_job.uuid)):
            log.error("Print job about to be logged(uuid=%s), already exists!\n%s",
                      self.__current_job.uuid, self.__current_job.toJson())
            return

        self.__current_job.time_elapsed = datetime.timedelta(seconds=metadata.get("time_elapsed", 0))
        self.__current_job.time_total = datetime.timedelta(seconds=metadata.get("time_total", 0))
        if self.__current_job.result is None:
            self.__current_job.result = PrintJob.JOB_RESULT_ABORTED if self.isAborted() else PrintJob.JOB_RESULT_FINISHED

        if self.__gcode_metadata is not None:
            g_code_estimated_time = self.__gcode_metadata.getEstimationMetadata()
            self.__current_job.time_estimated = datetime.timedelta(seconds=g_code_estimated_time.get("time", 0))

        self.job_history.addPrintJob(self.__current_job.toDbus())

    @property
    def file_path(self):
        if self._parse_header_step is not None and self._parse_header_step.getFileHandler() is not None:
            return self._parse_header_step.getFileHandler().file_path
        return ""
