import logging

from griffin.printer.controller import Controller
from griffin.printer.procedures.otherProcedureStep import OtherProcedureStep
from griffin.printer.procedures.print.resumeStep import ResumeStep
from griffin.printer.procedures.print.testPausedStep import TestPausedStep
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.steps.gotoPositionStep import GotoPositionStep
from griffin.printer.procedures.steps.setPropertyStep import SetPropertyStep
from griffin.printer.procedures.waitForQueueEmptyStep import WaitForQueueToBeEmptyStep


log = logging.getLogger(__name__.split(".")[-1])


# This procedure resumes the print. The pause procedure gives the position, speed and temperature to this procedure
# for storage and usage upon resume.
class ResumePrintProcedure(Procedure):
    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_UNKNOWN)
        self.__steps =[]
        self.__saved_state = {}

    def setup(self, args):
        c = Controller.getInstance()
        if not self.__saved_state or not c.getPaused():
            return False

        # Simple check: if the calibration print is paused, which is a maintenance procedure, the printing part
        # can be paused and should be able to be resumed.
        is_calibration_print_running = False
        for p in c.getActiveProcedures():
            if p[0] == "CALIBRATE_XY_AXIS_WIZARD":
                is_calibration_print_running = True
                break

        # don't resume while a change material or other procedure is running during the paused state.
        if not args.get("allow_resume_even_if_maintenance_procedures_active", is_calibration_print_running) and [ p for p in c.getAllProcedures() if p.isActive() and p.isMaintenanceProcedure() ]:
            return False

        self.__steps = [
            TestPausedStep("TEST_IF_NOT_PAUSED", True, ["PAUSE_PRINT"]),
            WaitForQueueToBeEmptyStep("WAITING_1"),
        ]

        for hotend_nr in range(c.getPropertyValue("hotend_count")):
            self.__steps.append(SetPropertyStep("START_HEATUP_HOTEND_{hotend_nr}".format(hotend_nr=hotend_nr),
                "printer.head.0.hotend.{hotend_nr}".format(hotend_nr=hotend_nr), "pre_tune_target_temperature",
                self.__saved_state.get("T{hotend_nr}".format(hotend_nr=hotend_nr))))
        self.__steps.append(OtherProcedureStep("RETURN_TO_ACTIVE_HOTEND","SWITCH_ACTIVE_HOTEND", {"target": self.__saved_state.get("active_hotend")}))
        # new loop because we want the temperature to be set simultaneously so we don't have to wait on both nozzles heating up one after the other.
        for hotend_nr in range(c.getPropertyValue("hotend_count")):
            self.__steps.append(OtherProcedureStep("HEATUP_HOTEND_WAIT_{hotend_nr}".format(hotend_nr=hotend_nr),
                "SET_HOTEND_TEMPERATURE_WAIT", {"hotend_nr": hotend_nr, "target": self.__saved_state.get("T{hotend_nr}".format(hotend_nr=hotend_nr)), "heatup_only": True}))

        x = {"X":"x","Y":"y"}
        self.__steps.append(GotoPositionStep("RETUNING_NO_Z", **{x[key]: self.__saved_state[key] for key in x}))

        self.__steps.append(ResumeStep("RESUMING", self))

        self._addStepSequence(self.__steps)

        return True

    ## This function takes the state saved by the PausePrintProcedure and stores it here for when the print is resumed.
    # This function is only called by the pauseStep and the resumeStep to set and reset the printer state respectively.
    # @param values A dictionary containing the state of the printer at the time of pausing, or an empty dictionary.
    def setSavedState(self, values):
        self.__saved_state = values

    ## This function returns the printer state saved by the pause print procedure for use in the ResumeStep.
    # @return A dictionary containing the state of the printer at the time of pausing, or an empty dictionary.
    def getSavedState(self):
        return self.__saved_state
