from enum import Enum
from griffin import signal
from threading import Timer
from collections import OrderedDict

import logging

log = logging.getLogger(__name__.split(".")[-1])


## @brief Basic outcomes for a ProcedureStep
class OutcomeBase(Enum):
    Success = "Success"
    Aborted = "Aborted"
    Error = "Error"


class ProcedureStep:
    """ Runs a step in its own thread.

    Expected flows that are handled:

        Setup --> Running on start()
                  +--> Finished on normal run() return
                  +--> Aborted if run() returns with _abort flag set.
                       +--> Setup on finishStep()
                  +--> Error if run() returns due to an exception.
                       +--> Setup on finishStep()
    """

    TIME_OUT_WAIT_FINISHED_EVENT = 0.1      # A time-out value used in thread wait events. Use low values like 0.1s to improve quick system responses.

    ## @brief Initialize a new ProcedureStep
    # @param step_timeout int amount of seconds that this step should maximally take. After this time, self._onStepTimeout is called, to handle custom handling in subclasses
    # @param outcomes is the additional outcomes the step can have additional to the default ones in OutcomeBase. E.g. if a Step can have 'exploded' as an outcome, list it here if e.g. Error is not a good match
    def __init__(self, key, step_timeout=None, outcomes=None):
        super().__init__()

        self.stepFinished = signal.Signal()

        self.__key = key
        self._next_step = None
        self._step_state = None
        self._is_active = False

        self._step_timeout = step_timeout
        self._timeout_timer = None
        self._step_timed_out = False

        # May be set by abort() or by receiving the ABORT message. During
        # normal path, don't start the next step, but the first step of the
        # "abort path" instead. The "abort path" isn't affected.
        self._abort = False

        # A halted procedure step must stop a.s.a.p. and _not_ execute any new step(s) or path.
        self.__is_halted = False

        ## Which outcomes does this step have? To be overridden in subclasses.
        # Any time self.run() returns a value, it must be in this list.
        assert outcomes is None or isinstance(outcomes, list), "outcomes must be a list."
        self._possible_outcomes = list(OutcomeBase) + ([] if outcomes is None else outcomes)

        ## Which step is next depends on the transitions in this mapping.
        self._transition_conditions = OrderedDict()

    # Execute the actions to be taken in a ProcedureStep
    # May return an outcome as defined in self._possible_outcomes
    def run(self):
        raise NotImplementedError()

    # @brief Allows the override-able .run-method to be simpler. _run add timeouts, default outcomes and triggers stepFinished
    # @returns outcome of the .run() method but overrides this in case of errors and aborts.
    def _run(self):
        self._is_active = True
        try:
            # The timer, if desired, is triggered after the given amount of seconds, specified in the constructor.
            # This allows for custom logic (in _onStepTimeout) to be executed when a step takes too long, for whatever reason.
            if self._step_timeout:
                self._timeout_timer = Timer(self._step_timeout, self.__handleTimeout)
                self._timeout_timer.start()

            outcome = self.run()
            if outcome is None:
                outcome = OutcomeBase.Success

            if self.__is_halted:
                # Do not write to log from within HALTED state: this will only hide the actual cause of the halt.
                outcome = OutcomeBase.Error
            elif self._abort:
                log.info("Step to Aborted state: %s", self.__key)
                outcome = OutcomeBase.Aborted
            else:
                log.debug("Step to Finished state: %s", self.__key)

            if outcome and outcome not in self._possible_outcomes:
                self._is_active = False
                raise ValueError("Outcome {out} is not a valid outcome for {step} because it is not in the possible outcomes: {outcomes}".format(out=outcome, step=self, outcomes=str(self._possible_outcomes)))

        except:
            log.exception("Exception caught from step: %s", self.__key)
            outcome = OutcomeBase.Error
            self._onError()
        finally:
            if self._timeout_timer:
                self._timeout_timer.cancel()

        # Do not change state, trigger events or finish the procedure if it's halted.
        if not self.__is_halted:
            # The step is finished, either normal or with abort or error.
            self.stepFinished.emit()
            self._abort = False  # reset aborted flag for long lived steps
            self._is_active = False

        return outcome

    # Get the next procedure step in the chain.
    # If the next step is None the procedure is finished and will be marked as inactive.
    def getNextStep(self, outcome):
        log.debug("{self}.getNextStep({out}): {{{trans}}}".format(self=self.getKey(), trans=self._formatTransitions(), out=outcome))
        for condition, step in self._transition_conditions.items():
            if outcome == condition:
                log.debug("Outcome of {curr} is {out}, next: {nxt}".format(curr=self.getKey(), out=outcome, nxt=step.getKey() if step else step))
                return step
            elif callable(condition) and condition(self):
                log.debug("Condition {cond}({curr}) satisfied, next: {nxt}".format(curr=self.getKey(), cond=condition, nxt=step.getKey() if step else step))
                return step
        else:  # The loop finished without breaks/returns
            log.debug("{step} returned outcome {out} for which no transition is registered in {trans}"
                      .format(step=self, out=outcome, trans=self._transition_conditions))
            # This is allowable, because the Procedure could simply be done and there is no next step in that case/
            return None

    def abort(self):
        """ Abort the step. """
        if self.__is_halted:
            # Aborting will trigger execution of the abort steps, in case of a halt, this is not desired.
            # We also like to prevent any log messages appearing after a halt.
            return

        log.info("Aborting step: %s", str(self))
        self._abort = True

    # @brief Sets the procedure step to HALTED as well as ABORTED.
    def halt(self):
        # Do not write to log from within HALTED state: this will only hide the actual cause of the halt.
        self.__is_halted = True
        self._abort = True

    def receiveMessage(self, message):
        """ Default receive message handler.

        External services can send a message to active procedures. These
        messages are passed to the currently active procedure step at which
        point this default handler is called if the real step doesn't implement
        its own handler.
        """
        return

    # Set the next step which will be ran when this step is finished. Should only be set during setup of the procedure.
    # @param step the step to transition to by default.
    # @param confirm_loop_is_ok A step going back to itself may lead to an infinite loop. If you are sure this is OK, set this to True.
    def setNextDefaultStep(self, step, confirm_loop_is_ok=False):
        assert step is None or isinstance(step, ProcedureStep), "step is not of an accepted type"

        if step is self and not confirm_loop_is_ok:
            raise ValueError("Setting self as next step may lead to an infinite loop. Use confirm_loop_is_ok=True to override this check")

        self._transition_conditions[OutcomeBase.Success] = step
        return step

    ## @brief Set a step to be executed based on some condition or transition name
    # @param condition the condition on which to select. Must be wither a transition of this step OR a callable returning True when the condition is satisfied
    # @param step the step to take if the condition is True
    # @param confirm_loop_is_ok A step going back to itself may lead to an infinite loop. If you are sure this is OK, set this to True
    def setNextConditionalStep(self, condition, step, confirm_loop_is_ok=False):
        if step is self and not confirm_loop_is_ok:
            raise ValueError("Setting self as next step may lead to an infinite loop. Use confirm_loop_is_ok=True to override this check")

        if callable(condition):
            self._transition_conditions[condition] = step
        else:
            if condition in self._possible_outcomes:
                self._transition_conditions[condition] = step
            else:
                raise ValueError("{curr} has no possible outcome {outcome}".format(curr=self.getKey(), outcome=condition))
        log.debug("Add {curr}.{cond} --> {step}".format(curr=self.getKey(), cond=condition, step=step.getKey() if step else step))  # Print step key if it's an actual step.

    def getKey(self):
        return self.__key

    def getMetaData(self):
        return {"EMPTY_ORIGIN": "ProcedureStep"}

    # This function will be called when the derived step procedure throws an error.
    # Optional virtual function to be implemented in derived classes.
    def _onError(self):
        pass

    ## @brief What to do when the step is taking too long? Subclasses can override _onStepTimeout.
    # This private methods does some logging that something is taking longer than expected and sets the _step_timed_out flag.
    def __handleTimeout(self):
        log.warning("{cls}(\"{key}\") takes too long too run, over {timeout} seconds".format(cls=self.__class__.__name__, key=self.__key, timeout=self._step_timeout))
        self._step_timed_out = True
        self._onStepTimeout()

    ## @brief What to do when the step is taking too long? May be overridden in subclasses
    # The base implementation does nothing, subclasses can override
    def _onStepTimeout(self):
        pass

    def hasTimedOut(self):
        return self._step_timed_out

    # @return a boolean indicating whether the step is active or not.
    def isActive(self):
        return self._is_active

    ## @brief Check that self is an active descendant of a given procedure or ProcedureStep.
    # @param ancestor, Procedure or ProcedureStep that we want to know self is a descendant of or not.
    # @returns bool indicating that the ancestor is an ancestor of this step or not.
    def isActiveDescendant(self, ancestor):
        active_ancestor = ancestor.getActiveStep()
        if not active_ancestor:
            return False  # If there is no active step on the ancestor, I can't possibly be the active descendant.

        log.debug("isActiveDescendant({k}={anc})".format(anc=ancestor, k=ancestor.getKey()))
        log.debug("active_ancestor({k}={anc})".format(anc=active_ancestor, k=active_ancestor.getKey()))

        # This is for the case of an OtherProcedureStep.
        # We cannot check if we're an instance of that because importing that class would create a circular reference.
        # So this is the way around that.
        if hasattr(active_ancestor, "getProcedure"):
            active_ancestor = active_ancestor.getProcedure()
            log.debug("active_ancestor is an OtherProcedureStep, so replaced by actual procedure:({k}={anc})".format(anc=active_ancestor, k=active_ancestor.getKey()))

        if active_ancestor == self:
            return True
        else:
            return self.isActiveDescendant(active_ancestor)     # Recursive call

    def __repr__(self):
        return "{cls}(key='{key}')".format(cls=self.__class__.__name__, key=self.getKey())

    ## @brief Returns a string representation of the transitions between this step and possible next steps.
    # @param sep the separator str between transitions->step pairs.
    def _formatTransitions(self, sep=", "):
        return sep.join(["{cond}->{to}".format(cond=k, to=v) for k, v in self._transition_conditions.items()])

    ## Returns a number roughly proportional to the time this step will take, for the sake of progress estimation
    # @return a number. A weight of 1 corresponds to an estimated 1 second duration
    def getDurationEstimate(self):
        return 1