from griffin.printer.procedures.procedureStep import ProcedureStep
from griffin.printer.controller import Controller
import logging

log = logging.getLogger(__name__.split(".")[-1])


# Procedure step to move the printer head to a certain position, at a certain speed.
class GotoPositionStep(ProcedureStep):
    # @brief Create a new GotoPosition step. When this step is executed the printer head will move to a certain position in the printer.
    # Axis that are not specified will not be moved.
    #
    # @param x position in mm, or None if the head should not move in the X direction.
    # @param y position in mm, or None if the head should not move in the Y direction.
    # @param z position in mm, or None if the head should not move in the Z direction.
    # @param speed in mm/sec
    # @param e position in mm, or None if the extruder for the active hotend should nog be moved
    # @param relative_positions Indicates whether or not all positions are a relative offset
    # @param wait_for_queue_empty if true, this step waits till the move is executed and finished.
    #        (Warning: Should not be true during printing! As will cause blobs/delays in the movement)
    # @param calculate_position_function Callable function that returns the coordinates of the position to move to.
    #           This function should return a dictionary, with optionally "x","y","z","e" and "speed" as keys.
    #           This function is called when this step runs.
    def __init__(self, key, x=None, y=None, z=None, speed=None, *, wait_for_queue_empty=True, e=None, relative_positions=False, calculate_position_function=None):
        if isinstance(x, dict):
            self.__init__(key, **x) ## Hack to allow the initial parameter to be an dictionary, used in the jedi.json genericProcedures
            return
        super().__init__(key)

        # If we have a calculate_position_function, the x,y,z,e parameters should be None, because the function will overrule this.
        if calculate_position_function is not None:
            assert callable(calculate_position_function)
            assert x is None and y is None and z is None and e is None

        self._x = x
        self._y = y
        self._z = z
        self._e = e
        self._speed = speed
        self._wait_for_queue_empty = wait_for_queue_empty
        self._relative = relative_positions
        self.__calculate_position_function = calculate_position_function

    def run(self):
        if self.__calculate_position_function is not None:
            position = self.__calculate_position_function()
            assert isinstance(position, dict)
            self._x = position.get("x", None)
            self._y = position.get("y", None)
            self._z = position.get("z", None)
            self._e = position.get("e", None)
            self._speed = position.get("speed", self._speed)

        log.info("Moving to: x: %s y: %s z: %s e: %s speed: %s, relative: %s" % (self._x, self._y, self._z, self._e, self._speed, self._relative))
        c = Controller.getInstance()
        positions = c.getPosition()
        hotend_nr = c.getGCodeActiveHotend()

        gcode = "G1 "
        if self._x is not None:
            gcode += "X%f " % (self._x + (positions["X"] if self._relative else 0))
        if self._y is not None:
            gcode += "Y%f " % (self._y + (positions["Y"] if self._relative else 0))
        if self._z is not None:
            gcode += "Z%f " % (self._z + (positions["Z"] if self._relative else 0))
        if self._e is not None:
            gcode += "E%f " % (self._e + (positions["E{hotend_nr}".format(hotend_nr=hotend_nr)] if self._relative else 0))

        speed = self._speed
        if speed is None:
            speed = c.getPropertyValue("default_travel_speed")
        gcode += "F%f" % (speed * 60.0) # Convert mm/s to mm/min

        log.debug("Gcode to send to Marlin: %s", gcode)
        c.queueRawGCode(gcode)
        # should only be used in pre or post print situation !
        if self._wait_for_queue_empty:
            c.waitForQueueToBeEmpty()
