import re
import time
import logging

from griffin.printer.procedures.procedureStep import ProcedureStep
from griffin.printer.controller import Controller
from griffin.printer.translationXYNozzleOffset import TranslationXYNozzleOffset

log = logging.getLogger(__name__.split(".")[-1])

# @brief Sets the offset value for the specified axis.
# @param key is a (unique) identifier name for this procedure step.
# @param axis specifies for which axis the offset will be set.
# @param value = None causes a blocking wait for a DBUS message with parameter 'USER_INPUT <value>'
#      or when a value is specified, the value will be set immediately.
class AlignXyStep(ProcedureStep):
    def __init__(self, key, axis, value=None):
        super().__init__(key)
        self._axis = axis.upper()
        self._value = value
        self._finished = False

    def run(self):
        # set the value so the offset can be cleared in a calibration print
        if self._value is not None:
            self.setAxis(self._value)
            return
        while not self._finished and not self._abort:
            time.sleep(0.5)

    def receiveMessage(self, message):
        log.info("message: %s", message)
        # Match the text USER_INPUT from the start followed by space(s) and a signed floating point number with 1 decimal
        m = re.match(r"^USER_INPUT\s+([-+]?\d*\.\d+)$", message)
        if m:
            try:
                offset_index = str(float(m.group(1)))
            except ValueError:
                log.info("Unable to parse user input %r", m.group(1))
                offset_index = 0

            log.debug("offset index= %r", offset_index)
            self.setAxis(TranslationXYNozzleOffset.getOffset(**{self._axis: offset_index}))
            self._finished = True

    def setAxis(self, value):
        c = Controller.getInstance()

        c.getPrintHeadController().getHotendSlot(1).setPropertyValue(self._axis.lower() + "_offset", value)
        log.info("new %s value %s\n", self._axis, str(value))
