import logging

from griffin.printer.controller import Controller
from griffin.printer.procedures.dummyStep import DummyStep
from griffin.printer.procedures.otherProcedureStep import OtherProcedureStep
from griffin.printer.procedures.procedure import Procedure
from griffin.printer.procedures.wizard.waitMessageStep import WaitMessageStep

log = logging.getLogger(__name__.split(".")[-1])


## @brief Procedure to whether the cartridge is compatible with a (newly loaded) Cartridge. If not, the material is unloaded
class CheckCartridgeMaterialCompatibilityProcedure(Procedure):
    def __init__(self, key):
        super().__init__(key, Procedure.TYPE_MAINTENANCE)
        self._hotend_nr = 0

        self.__incompatible = None

    def setup(self, args):
        self._hotend_nr = int(args.get("hotend_nr", 0))

        if self.__isCartridgeIncompatibleWithMaterial():
            self.__incompatible = True

            skip_heat_up = bool(args.get("skip_heat_up", False))

            confirm_unload = WaitMessageStep(key="CONFIRM_MATERIAL_UNLOAD", message="UNLOAD")
            unload_step = OtherProcedureStep("UNLOAD_INCOMPATIBLE_MATERIAL", "UNLOAD_MATERIAL_WIZARD", args={"hotend_nr": self._hotend_nr, "skip_heat_up": skip_heat_up})

            self._addStepSequence([confirm_unload, unload_step])
        else:
            self.setFirstStep(DummyStep(key="MATERIAL_COMPATIBLE_DUMMYSTEP"))
            self.__incompatible = False

        return True

    ## @brief Returns whether the check detected that the material is *in*compatible or not.
    # Note that not incompatible does not means compatible per se, it may be unknown as well.
    def getIncompatible(self):
        return self.__incompatible

    ## @brief Check whether the Cartridge and material are INcompatible.
    # Not incompatible does not mean fully compatible per se, it may also be unknown
    # @returns bool indicating True is certainly incompatible and False otherwise
    def __isCartridgeIncompatibleWithMaterial(self):
        hcm = Controller.getInstance().getPrintHeadController().getHotendCartridgeManager()
        mm = Controller.getInstance().getMaterialManager()
        hotend_cartridge_id = hcm.getValue(self._hotend_nr, "hotend_cartridge_id")
        hotend = Controller.getInstance().getPrintHeadController().getHotendSlot(self._hotend_nr)
        material_guid = hotend.getPropertyValue("material_guid")

        incompatible = mm.isHotendCompatibleWithMaterial(hotend_cartridge_id, material_guid) == mm.INCOMPATIBLE
        log.info("The cartridge is slot {slot}:({ci}) is {comp}compatible with the material ({mat})"
                 .format(slot=self._hotend_nr, ci=hotend_cartridge_id, mat=material_guid,
                         comp="IN" if incompatible else ""))
        return incompatible
