import logging
import math

from griffin.printer.controller import Controller
from griffin.printer.procedures.procedureStep import ProcedureStep

log = logging.getLogger(__name__.split(".")[-1])


## @brief wait for the user to confirm that material is extruded from the nozzle
# If step_timeout is set and the user does not confirm this time
#   assume that the user walked off and material is in fact extruding and continue without confirmation
class MoveMaterialWaitMessageStep(ProcedureStep):
    ##< The default diameter of the material (see EM-952) in mm
    __DEFAULT_MATERIAL_DIAMETER = 2.85
    ##< The expected speed for the final step in mm3/s
    __DEFAULT_SPEED = 6.0

    def __init__(self, key, hotend = 0, direction = "forward", step_timeout=None):
        super().__init__(key, step_timeout=step_timeout)
        self._target_hotend = int(hotend)
        self._dir = direction
        # Movement speed converted from 6 mm3/s to mm/s: calculate the surface of the material using pi * r^2 formula.
        # The default diameter is 2.85 hence the half. Also, this hardcoded should not be changed in the future to use the actual material profile
        # TODO: EM-952
        self.__target_speed = self.__DEFAULT_SPEED / (math.pi * ((self.__DEFAULT_MATERIAL_DIAMETER / 2.0) ** 2))

    def run(self):
        self._done_received = False
        self._step_timed_out = False

        c = Controller.getInstance()
        original_active_hotend = c.getGCodeActiveHotend()
        c.queueRawGCode("M302 S0")
        if self._target_hotend != original_active_hotend:
            c.queueRawGCode("T%d" % self._target_hotend)
            c.setGCodeActiveHotend(self._target_hotend)

        # this is essentially priming, so add tracking of the primed state.
        c.getPrintHeadController().getHotendSlot(self._target_hotend).setPropertyValue("is_primed", True)
        while not self._done_received and not self._abort and not self.hasTimedOut():
            c.queueRawGCode("G92 E0")

            if self._dir == "reverse" or self._dir == "backwards":
                c.queueRawGCode("G1 E-0.1 F%f" % (self.__target_speed * 60))
            else:
                c.queueRawGCode("G1 E0.1 F%f" % (self.__target_speed * 60))

            c.waitForQueueToBeEmpty(max_planned_motions=3)

        # aborting the queue during the time when the 2nd nozzle is active corrupts the Z height information due to quickstop code in marlin.
        c.queueRawGCode("G92 E0")
        c.queueRawGCode("M302 S%d" % c.getPropertyValue("default_minimal_extrusion_temperature"))    # safety back on
        c.waitForQueueToBeEmpty(wait_for_motion_controller=True)

        # re-set extruded length after the user acknowledges material is coming out of the nozzle.
        # Use this step only to prime the nozzle after a (re)loading of material,
        # or to feed material into the feeder.
        c.setRetractedLength(0)
        if self._target_hotend != original_active_hotend:
            c.queueRawGCode("T%d" % original_active_hotend)
            c.setGCodeActiveHotend(original_active_hotend)


    def receiveMessage(self, message):
        """ Handle the "DONE" message. Pass rest to super() """

        log.info("Message: %s", message)

        if message == "DONE":
            self._done_received = True
        else:
            # Pass the message
            super().receiveMessage(message)

    ## @brief Log that we're continuing without user confirmation.
    def _onStepTimeout(self):
        log.warning("{key}: User has not confirmed material flowing out of machine after waiting {timeout}s. Continuing without user confirmation"
                    .format(key=self.getKey(), timeout=self._step_timeout))
