from griffin.printer.properties.printer.printerStateProperty import PrinterStateProperty
from .faultHandler import FaultHandler, FaultCode, FaultLevel
import time

## This class is responsible for triggering the message "a new firmware version is available!" on the display.
class UpdateNotificationHandler():
    _UPDATE_MESSAGE_DELAY = 10  ## seconds after a status change to idle to allow the showing of an update "popup" screen.

    def __init__(self, system, controller):
        self._controller = controller
        self._system = system
        system.connectSignal("onNewFirmwareVersionDetected", self._handleNewFirmwareVersionAvailable)
        self._trigger_new_firmware_version_update_message = None
        self._new_firmware_version = None
        state = self._controller.getPropertyContainer().get("state")
        state.onChange.connect(self._handleStateChange)

    ## Handles new firmware detected events from the system service and fires off a fault in case of the idle state being reached recently.
    # @param version the version of the newly detected firmware.
    def _handleNewFirmwareVersionAvailable(self, version):
        def trigger_firmware_message():
            new_version = '.'.join(version.split(".")[:-1])
            current_version = '.'.join(self._system.getVersion().split(".")[:-1])
            FaultHandler.info("new firmware version detected message is now shown on display! {firmware_version}".format(firmware_version=version),
                              FaultCode.NEW_FIRMWARE_VERSION_AVAILABLE, data={"new_version": new_version, "current_version": current_version} )

        self._trigger_new_firmware_version_update_message = trigger_firmware_message
        self._new_firmware_version = version
        state = self._controller.getPropertyContainer().get("state")
        if state.get() == PrinterStateProperty.IDLE and state.getLastChanged() + self._UPDATE_MESSAGE_DELAY > time.time():
            trigger_firmware_message()

    # Handles state changes of the printer state and fires off a fault in case of new firmware version detected and the idle state.
    # @param property the state property object
    # @param state the new printer state
    def _handleStateChange(self, property, state):
        if self._trigger_new_firmware_version_update_message is not None:
            if state == PrinterStateProperty.IDLE:
                self._trigger_new_firmware_version_update_message()
            else:
                FaultHandler.clear(FaultLevel.INFO, FaultCode.NEW_FIRMWARE_VERSION_AVAILABLE)

    ## Ignore the last update message, when the ignore button is used in the display on the new firmware detected message.
    def ignoreNewFirmwareVersion(self):
        self._trigger_new_firmware_version_update_message = None
        self._system.ignoreUpdateVersion(self._new_firmware_version)
        self._new_firmware_version = None
