import logging
import time
import subprocess
import platform
import re
import struct
import os
import uuid
import shutil
import glob
import gzip
import tarfile

from griffin.factoryReset import FactoryReset
from griffin import dbusif
from griffin.system import systemUpdate
from griffin.preferences.registryFile import RegistryFile
from griffin import thread

log = logging.getLogger(__name__.split(".")[-1])


# Basic system service, used to query some information about the system as a whole.
class SystemService(dbusif.ServiceObject):
    # Location of the system wide preferences file
    __PREFERENCES_FILE = "system_preferences.json"

    # Preference key: machine name
    __PREF_KEY_MACHINE_NAME = "machine_name"
    # Preference key: developer mode
    __PREF_KEY_DEVELOPER_MODE = "developer_mode"
    # Preference key: country code
    __PREF_KEY_COUNTRY_CODE = "country_code"
    # Preference key: system language
    __PREF_KEY_SYSTEM_LANGUAGE = "system_language"

    # Location of the EEPROM in the filesystem (eeprom is exposed as a file-like object by linux)
    __SYSTEM_EEPROM_FILE = "/sys/bus/i2c/devices/1-0050/eeprom"
    # Offsets for various parts in the EEPROM see docs/Board_EEPROM_layout.txt for details.
    __SYSTEM_EEPROM_BOM_OFFSET = 0x100
    __SYSTEM_EEPROM_REV_OFFSET = 0x104
    __SYSTEM_EEPROM_GUID_OFFSET = 0x108

    # These settings needs to be saved during a soft factory reset
    __SAVE_SETTINGS_ON_SOFT_RESET = []
    # These settings need to be set
    __ADD_SETTINGS_ON_SOFT_RESET = { }

    # The length of the MAC address set by the system as part of the hostname
    __MAC_ADDRESS_LENGTH = 12
    # The unique part length is the last part of the MAC which needs to be an int for indexing
    __MAC_UNIQUE_PART_LENGTH = int(__MAC_ADDRESS_LENGTH / 2)

    # The BOM id which identifies the machine as a programming station
    ___PROGRAMMING_STATION_BOM = 1797

    def __init__(self):
        super().__init__("system")

        self.__settings = RegistryFile(self.__PREFERENCES_FILE)
        if self.__settings.getAsString(self.__PREF_KEY_COUNTRY_CODE, None) is None:
            self.__settings.setAsString(self.__PREF_KEY_COUNTRY_CODE, "")
        if self.__settings.getAsString(self.__PREF_KEY_SYSTEM_LANGUAGE, None) is None:
            self.__settings.setAsString(self.__PREF_KEY_SYSTEM_LANGUAGE, "")

        self.__reset_callbacks = {}

        self._machine_bom_number = 0
        self._machine_revision_number = 0
        self._machine_guid = ""

        self._readEEPROM()
        self._initMachineName()

        self._system_update = systemUpdate.SystemUpdate(self._machine_bom_number)
        self._system_update.onUpdateStart.connect(self.onUpdateProcedureStart)
        self._system_update.onPreUpdateSuccessful.connect(self.reboot)
        self._system_update.onNewFirmwareVersionDetected.connect(self.onNewFirmwareVersionDetected)

        self.__firmware_check = ""
        self._system_update.onFirmwareVersionLoaded.connect(self._handleFirmwareUpdate)
        self._system_update.onFirmwareVersionLoadFailed.connect(self._handleFirmwareVersionLoadFail)

        log.info("Software version: %s", self.getVersion())

        if self.isDeveloperModeActive():
            log.info("Starting dropbear SSH server because developer mode is active.")
            p = subprocess.Popen(["sudo", "/bin/systemctl", "start", "dropbear"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()
            p.wait()

    # Load the machine BOM and machine revision from the board EEPROM
    #
    # Read the board EEPROM and read the machine BOM and REVISION from there. These are stored as unsigned 32bit integers at a certain offset in the EEPROM.
    # Writing the EEPROM needs to be done as root. Till we have the proper facilities for this, the following commandline can be used:
    # python3 -c 'bom_number = 1; revision_number = 0xFFFFFFFF; import struct; f = open("/sys/bus/i2c/devices/1-0050/eeprom", "r+b"); f.seek(0x100); f.write(struct.pack("!I", bom_number)); f.seek(0x104); f.write(struct.pack("!I", revision_number)); f.close()'
    def _readEEPROM(self):
        try:
            with open(self.__SYSTEM_EEPROM_FILE, "rb") as f:
                f.seek(self.__SYSTEM_EEPROM_BOM_OFFSET)
                bom_number = struct.unpack("!I", f.read(4))[0]
                f.seek(self.__SYSTEM_EEPROM_REV_OFFSET)
                rev_number = struct.unpack("!I", f.read(4))[0]
                f.seek(self.__SYSTEM_EEPROM_GUID_OFFSET)
                guid = str(uuid.UUID(bytes=f.read(16)))
        except OSError as e:
            log.error("Exception while reading machine BOM/Revision/GUID: %s\nUsing default values for BOM, GUID and revision.", e)
            bom_number = 0
            rev_number = 0
            guid = "00000000-0000-0000-0000-000000000000"
        if bom_number == 0xFFFFFFFF:
            bom_number = 0
        if rev_number == 0xFFFFFFFF:
            rev_number = 0
        self._machine_bom_number = bom_number
        self._machine_revision_number = rev_number
        self._machine_guid = guid
        log.info("Machine BOM: %d", bom_number)
        log.info("Machine Revision: %d", rev_number)
        log.info("Machine GUID: %s", guid)

    # Gets the machine name and defines a default when it's not set or 'Ultimaker Jedi'
    #  If a default name needs to be set, the settings are also saved
    #  @param settings The global settings
    #
    def _initMachineName(self):
        machine_name = self.__settings.getAsString("machine_name")
        if machine_name is None or machine_name == "Ultimaker Jedi":    # For backwards compatibility to early beta versions that used a static machine name
            # Create a user readable machine name.
            # Based on MAC address which should be the last 6 characters of the hostname, but after the - sign
            hostname = os.uname().nodename
            hostname_parts = hostname.split("-")
            unique_name = self._machine_guid
            if len(hostname_parts) == 2:
                unique_name = hostname_parts[1]
                # Only take the last half of the MAC address if possible
                if len(unique_name) >= SystemService.__MAC_ADDRESS_LENGTH:
                    unique_name = unique_name[-SystemService.__MAC_UNIQUE_PART_LENGTH:]
            else:
                log.error("The MAC address is not set for this printer, please see a service center!")

            self.setMachineName("Ultimaker-" + unique_name)

    # Determine if the machine name is valid (RFC 952 and RFC 1123)
    #  @param machine_name The machine name to be validated
    #  @return Returns True if the name is valid, otherwise False
    def _isValidMachineName(self, machine_name):
        allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
        return allowed.match(machine_name)

    ## @brief Get the version of the current install firmware
    #  @return Returns the current version
    @dbusif.method("", "s")
    def getVersion(self):
        return self._system_update.getVersion()

    ## @brief Check if a new firmware for the specified release type is available and emit signal for the result.
    #  @param release_type The firmware release type to check
    # As a result, it emits the onHasNewFirmware, onHasEqualFirmware or onFirmwareVersionLoadFailed signal for *each* release type
    @dbusif.method("s", "")
    def checkForFirmwareUpdates(self, release_type):
        log.debug("Check for new firmware version for '%s' initiated...", release_type)
        self.__firmware_check = release_type
        self._system_update.checkForFirmwareUpdates()

    ## @brief If a new version is found, send a signal
    #  @param release_type Specifies which release type of firmware
    def _handleFirmwareUpdate(self, release_type):
        if release_type == self.__firmware_check:
            if self._system_update.hasEqualFirmware(self.__firmware_check):
                self.onHasEqualFirmware(self.__firmware_check)
            else:
                self.onHasNewFirmware(self.__firmware_check, self._system_update.getFirmwareVersion(self.__firmware_check))

    ## @brief If we cannot load which version is newer, emit a signal
    #  @param release_type Specifies which release type of firmware
    def _handleFirmwareVersionLoadFail(self, release_type):
        if release_type == self.__firmware_check:
            self.onFirmwareVersionLoadFailed(release_type)

    ## @brief Signals that a new firmware version has been detected
    #  @param release_type Specifies which release type of firmware
    #  @param version The new version
    @dbusif.signal("ss")
    def onHasNewFirmware(self, release_type, version):
        log.debug("New firmware detected for '%s', version: %s", release_type, version)
        return True

    ## @brief Signals that an equal firmware version has been detected
    #  @param release_type Specifies which release type of firmware
    @dbusif.signal("s")
    def onHasEqualFirmware(self, release_type):
        log.debug("Equal firmware detected for '%s'", release_type)
        return True

    ## @brief Signals that it could not be detected whether there is a new firmware version
    #  @param release_type Specifies for which release type of firmware the new version could not be determined.
    @dbusif.signal("s")
    def onFirmwareVersionLoadFailed(self, release_type):
        log.info("Could not load/check whether there is a new firmware version for '%s'", release_type)
        return True

    ## Signals that there is a new firmware version available, that is not ignored by the user, etc.
    @dbusif.signal("s")
    def onNewFirmwareVersionDetected(self, version):
        log.debug("detected new valid firmware version '%s'", version)
        return True

    ## Set a specific version as the version to be ignored.
    # @param version The version to be ignored
    @dbusif.method("s", "")
    def ignoreUpdateVersion(self, version):
        self._system_update.ignoreVersion(version)

    @dbusif.method("", "s")
    def getPlatform(self):
        return str(platform.platform())

    @dbusif.method("", "s")
    def getHostName(self):
        return str(platform.node())

    @dbusif.method("", "b")
    def isDeveloperModeActive(self):
        active = self.__settings.getAsBoolean(self.__PREF_KEY_DEVELOPER_MODE)
        if active is None:
            active = False
            self.__settings.setAsBoolean(self.__PREF_KEY_DEVELOPER_MODE, active)
        return active

    # Set the developer mode option on/off.
    # Changing the developer mode from on to off or the other way around will reboot the machine.
    # This is to make sure all systems are properly configured for developer mode or not. And is a lot easier to do than
    # re-configuring all services. This means users of the developermode flag do not need to worry about dynamic behaviour of this flag.
    @dbusif.method("b", "")
    def setDeveloperMode(self, active):
        if self.isDeveloperModeActive() != active:
            self.__settings.setAsBoolean(self.__PREF_KEY_DEVELOPER_MODE, active)
            self.reboot()

    # Set the current country code, according to https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
    # Currently this is only used for WiFi regulatory domain settings. But in the future this could be expanded for other reasons as well.
    @dbusif.method("s", "b")
    def setCountry(self, country_code):
        # Country codes need to be 2 letters. And should be in https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
        # For forwards compatbility, we just check 2 upper case characters.
        # Also allow an empty string, for "no country set", which is the default startup state.
        if not re.match("^[A-Z]{2}$", country_code) and country_code != "":
            return False
        if self.getCountry() != country_code:
            self.__settings.setAsString(self.__PREF_KEY_COUNTRY_CODE, country_code)
            self.countryChanged(country_code)  # Send the dbus signal that the country has changed.
        return True

    # DBus interface to get the currently active selected country code.
    # Will return an empty string for no country set, or a 2 upper case letter code according to https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
    @dbusif.method("", "s")
    def getCountry(self):
        return self.__settings.getAsString(self.__PREF_KEY_COUNTRY_CODE, "")

    # Dbus signal, calling this function emits the dbus signal that the country has changed.
    @dbusif.signal("s")
    def countryChanged(self, country):
        log.info("Country changed: %s", country)

    @dbusif.method("s", "")
    def setLanguage(self, system_language):
        self.__settings.setAsString(self.__PREF_KEY_SYSTEM_LANGUAGE, system_language)
        self.languageChanged(system_language)

    @dbusif.method("", "s")
    def getLanguage(self):
        return self.__settings.getAsString(self.__PREF_KEY_SYSTEM_LANGUAGE, "")

    # Dbus signal, calling this function emits the dbus signal that the system language has changed.
    @dbusif.signal("s")
    def languageChanged(self, system_language):
        log.info("Language changed: %s", system_language)

    # If the specified machine name differs, it's set and saved into the settings file
    # @param string The new machine name
    # #return Returns True if machinename is valid and set, otherwise False
    @dbusif.method("s", "b")
    def setMachineName(self, machine_name):
        if self._isValidMachineName(machine_name):
            self.__settings.setAsString(self.__PREF_KEY_MACHINE_NAME, machine_name)
            self.machineNameChanged(machine_name)
            return True
        log.warning("Machine name '%s' is not a valid name", machine_name)
        return False

    # Returns the machine name from the settings file
    # @return string The machine
    @dbusif.method("", "s")
    def getMachineName(self):
        return self.__settings.getAsString(self.__PREF_KEY_MACHINE_NAME)

    # Dbus signal, calling this function emits the dbus signal that the machine name has changed.
    @dbusif.signal("s")
    def machineNameChanged(self, machine_name):
        log.info("Machine name changed: %s", machine_name)

    # Returns the current machine BOM and Revision number.
    # The BOM number is the number of the hardware assembly that is used to build this machine.
    # The Revision number is an optional revision.
    #
    # @return tuple of two integers. First one being the BOM number, second one being the revision number.
    #       The BOM is 0 if there is no BOM set yet.
    #       The revision is 0 if there is no revision set, or the revision is unknown. (default)
    @dbusif.method("", "(uu)")
    def getMachineBOM(self):
        return (self._machine_bom_number, self._machine_revision_number)

    # Return the machine GUID
    #
    # @return single string representing the GUID of the machine
    @dbusif.method("", "s")
    def getMachineGUID(self):
        return self._machine_guid

    ## @brief Initiate a factory reset. This will wipe all settings, and reboot the machine.
    #  @param reset_type Has a string parameter, which needs to contain the value "DESTROY" to make sure this function is not called by mistake.
    @dbusif.method("s", "")
    def factoryReset(self, reset_type):
        if reset_type not in (FactoryReset.FACTORY_RESET_KEYWORD_DESTROY, FactoryReset.FACTORY_RESET_KEYWORD_SOFT_DESTROY):
            return
        log.info("Factory Reset executing...")

        log.info("Executing factory resets for registered callbacks")
        for dbus_object_name, callback_function in self.__reset_callbacks.items():
            log.info("Calling reset callback function '%s' for dbus service '%s'", callback_function, dbus_object_name)
            ## Dirty hack because of naming issues
            if dbus_object_name == "auth_api":
                service = dbusif.RemoteObject("nl.ultimaker.interface.http", "interface/http/auth")
            else:
                service = dbusif.RemoteObject(dbus_object_name)
            try:
                method = getattr(service, callback_function)
                method(reset_type)
            except:
                # Ignore non existing function or dbus timeouts
                log.exception("Callable function '%s' does not exist for dbus service '%s'", callback_function, dbus_object_name)
                pass
        log.info("Callbacks finished")

        log.info("Cleaning up before reboot")
        if reset_type == FactoryReset.FACTORY_RESET_KEYWORD_SOFT_DESTROY:
            self.__settings.backupAndSetup(self.__SAVE_SETTINGS_ON_SOFT_RESET, self.__ADD_SETTINGS_ON_SOFT_RESET)
        else:
            self.__settings.erase()

        log.info("rebooting")
        self.reboot()

    ## @brief Reboot the machine, note that this a forceful reboot, so everything will be shut down, even if prints and stuff are still busy.
    #  The onReboot signal will be emitted to indicate that we are rebooting.
    @dbusif.method("", "")
    def rebootMachine(self):
        self.reboot()

    ## @brief This function registers a callback function to the dictionary. Each callback will be
    #  executed (in random order) when a factory reset is called for
    #  @param dbus_service_name The name of the dbus service object
    #  @param function_name The name of the function to be called (implemented in the specified dbus service object)
    @dbusif.method("ss", "")
    def addFactoryResetCallback(self, dbus_service_name, function_name):
        if dbus_service_name in self.__reset_callbacks.keys():
            log.warning("Duplicate reset callback: %s:%s" % (dbus_service_name, function_name))
        log.info("Added dbus service '%s' to the reset callbacks with function '%s'", dbus_service_name, function_name)
        self.__reset_callbacks[dbus_service_name] = function_name

    @dbusif.method("", "at")
    def getMemoryUsage(self):
        return self._getMemory()

    @dbusif.method("", "as")
    def getLog(self):
        p = subprocess.Popen(["/bin/journalctl", "-n250", "-o", "cat", "-u", "griffin.*"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        p.wait()
        return out.decode("utf-8", "replace").split("\n")

    @dbusif.signal("")
    def onUpdateProcedureStart(self):
        pass

    # onReboot is fired when the system is going to reboot.
    @dbusif.signal("")
    def onReboot(self):
        pass

    @dbusif.method("", "sd")
    def getUpdateState(self):
        return self._system_update.getState(), self._system_update.getDownloadProgress()

    ## Initiate a system update. This will cause a download of the new image, and this image will be checked with the signature keys in the current image.
    #  If the signature check is valid, then the update will be installed.
    #  As this is async process, feedback from this process is given with the getUpdateState function and onUpdateStart signal.
    #  @param release_type: One of the keys in the SystemUpdate.__SYSTEM_UPDATE_URLS dict. To indicate which type of firmware to install.
    @dbusif.method("s", "")
    def startUpdate(self, release_type):
        self._system_update.downloadUpdate(release_type)

    ## Initiate a system update. This will copy the given files to tmpfs, and this image will be checked with the signature keys in the current image.
    #  If the signature check is valid, then the update will be installed.
    #  As this is async process, feedback from this process is given with the getUpdateState function and onUpdateStart signal.
    #  @param update_filename: tar.xz filename to install, needs to exist in the system.
    @dbusif.method("s", "")
    def startLocalUpdate(self, update_filename):
        self._system_update.installLocalUpdate(update_filename)

    ## DBus method to get the possible update target types (given to "release_type") which their possible version number.
    #  @return a dictionary with the release_type as key, and the version string as value. The version string is empty when the version string is not yet known.
    @dbusif.method("", "a{ss}")
    def getUpdateTargetVersions(self):
        return self._system_update.getUpdateTargetVersions()

    def _getMemory(self):
        info = {}
        info["MemTotal"] = 0
        info["MemFree"] = 0
        info["Buffers"] = 0
        info["Cached"] = 0
        with open("/proc/meminfo", "r") as f:
            for line in f:
                key, value = map(str.strip, line.split(":", 2))
                try:
                    if value.endswith(" kB"):
                        info[key] = int(value[:-3]) * 1024
                    else:
                        info[key] = int(value)
                except ValueError:
                    pass
        total_memory = info["MemTotal"]
        total_free = info["MemFree"] + info["Buffers"] + info["Cached"]
        return [total_memory - total_free, total_memory]

    ## Initiate a reboot of the machine
    #  This is pushed to a background thread to give the caller some time to finish up what it was doing
    #  And give the system time to display towards the user that it is rebooting.
    def reboot(self):
        t = thread.Thread("reboot", self._rebootThread)
        t.start()

    def _rebootThread(self):
        log.info("Reboot starting")
        self.onReboot()
        self.__settings.forceSave()
        # Give the system 1 second to react on the reboot signal before we go and shutdown
        time.sleep(1.0)

        # the linux sync command to make sure all files are synced to disc before rebooting.
        p = subprocess.Popen(["/bin/sync"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.communicate()
        p.wait()

        p = subprocess.Popen(["/usr/bin/sudo", "/bin/systemctl", "reboot"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        log.info("rebooting %r\n%r", p.communicate())
        p.wait()

    ## @brief Service that runs a task to copy the journal and logfiles to the specified path (usb)
    #  @param path The destination path
    @dbusif.method("s", "")
    def dumpLogs(self, path):
        if path == "":
            path = self._getUsbDevicePath()

        if path is not None:
            thread.Thread("Dump Log Files", self.__copyLogFiles, args=(path,)).start()

    # Find usb partition
    #  @return Returns last found USB path. If no path was found, return None
    def _getUsbDevicePath(self):
        mounts = []
        usb_paths = []
        with open('/proc/mounts', 'r') as f:
            mounts = [line.split()[1] for line in f.readlines()]

        for path in mounts:
            path_parts = path.split("/")
            if len(path_parts) > 2:
                if "usb" in path_parts[-1]:
                    usb_paths.append(path)

        if len(usb_paths) > 0:
            return usb_paths[-1]
        else:
            return None

    ## @brief Copy the journal and log files, signalling when done
    #  @param path The destination path
    def __copyLogFiles(self, path):
        base_filename = path + "/" + self.getHostName() + "." + self.getVersion()
        boot_nr = 0
        last_journal = False
        # Assume (!) the worst
        # Note: removing the USB during copying might still continue without error.
        # Depending on when the writing takes place, this may be using the mountpoint on the filesystem
        # We're not handling this case - if you are removing the USB stick while the display is mentioning that copying is in progress,
        # it will be user error when nothing is put on the disk (check the local FS, it's likely to be there instead).
        result = False

        try:
            self.__mountUsb(path, "rw")

            log.info("Start copying the journal(s)")
            while not last_journal:
                target_filename = base_filename + ".boot" + str(boot_nr) + ".log.gz"
                log.info("Copying journal %r", target_filename)

                journal_process = subprocess.Popen(["/bin/journalctl", "--no-pager", "-b", str(boot_nr), "-o", "short"], stdout=subprocess.PIPE)
                # TODO: EM-1396 Stupid solution because piping to gzip seems to be nearly impossible to do this right / correctly
                # With (too) large journalctl output, this will eventually cause a memory issue
                journal_data = journal_process.stdout.read()
                journal_result = journal_process.wait()

                if journal_result != 0:
                    last_journal = True
                else:
                    gzip_handle = gzip.open(target_filename, "wb")
                    gzip_handle.write(journal_data)
                    gzip_handle.close()

                boot_nr -= 1
            log.info("Finished copying journal(s), start copying the log file(s)")

            for filename in glob.glob(os.path.join("/tmp", "*.log")):
                log.info("Copying logfile %r", filename)
                shutil.copy(filename, path)
            log.info("Finished copying log file(s), start copying configuration files")
            targz = tarfile.open("%s.config.tar.gz" % (base_filename), "w:gz")
            targz.add("/var/lib/griffin", "configuration", recursive=True, filter=self.__filterConfigurationPack)
            targz.close()
            log.info("Finished copying configuration files")

            result = True
        except:
            log.exception("Error while copying logfiles")

        # Even after failure, try to sync and remount, but in case of errors, silently ignore them
        try:
            os.sync()
            self.__mountUsb(path, "ro")
        except:
            pass

        self.onLogFilesCopied(result)

    ## Filter the files that will be packed into the targz for diagnostics.
    #  We want to exclude certain fails if they contain sensitive information, like the http API credentials
    #  @param tar_info: A TarInfo object from the tarfile module
    #  @return None if the file should be excluded.
    #  @return the given tar_info object if the file should be included.
    def __filterConfigurationPack(self, tar_info):
        if tar_info.name.endswith("http_api_users.json"):
            return None
        return tar_info

    ## @brief Mounts a path with remount
    #  @param path The path that needs to be remounted
    #  @param read_write Use "ro" for read-only, "rw" for read+write
    def __mountUsb(self, path, read_write):
        assert read_write in ["ro", "rw"]
        # Use try/except here. If the USB drive was removed, this could cause a kernel panic(!)
        try:
            p = subprocess.Popen(["sudo", "/bin/mount", "-o", "remount," + read_write, path])
            p.communicate()
            p.wait()
        except:
            log.error("Failing to remount path %s as %s", path, read_write)

    ## @brief The signal that is raised to indicate that the copying of the log files has been completed
    #  @param state When copying succeeded, this will be True, otherwise False
    @dbusif.signal("b")
    def onLogFilesCopied(self, state):
        return True

    @dbusif.method("", "b")
    def wasUpdateCompleted(self):
        return self._system_update.wasUpdateCompleted()

    ## @brief Acknowledge that the system was updated so that the user is not notified again
    @dbusif.method("", "")
    def acknowledgeUpdateCompleted(self):
        self._system_update.acknowledgeUpdateCompleted()

    @dbusif.method("", "i")
    def getUptime(self):
        return time.clock_gettime(time.CLOCK_MONOTONIC_RAW)

    # Get the system time as a timestamp in UTC time
    @dbusif.method("", "d")
    def getUTCSystemTime(self):
        return time.clock_gettime(time.CLOCK_REALTIME)

    # Set the system time as a timestamp in UTC time
    @dbusif.method("d", "")
    def setUTCSystemTime(self, new_time):
        time.clock_settime(time.CLOCK_REALTIME, new_time)

    # @brief Returns whether or not this machine is a programming station
    @dbusif.method("", "b")
    def isProgrammingStation(self):
        return self.getMachineBOM()[0] == SystemService.___PROGRAMMING_STATION_BOM

